#ifdef WIN32
# if defined(__MINGW32__) || defined(GNUWINCE)
#  include <sys/stat.h>
# endif
# ifndef GNUWINCE
#  include <windows.h>
# endif
# if defined(__MINGW32_MAJOR_VERSION) && defined(__MINGW32_MINOR_VERSION) \
		&& (__MINGW32_MAJOR_VERSION > (3) \
		|| (__MINGW32_MAJOR_VERSION == (3) && __MINGW32_MINOR_VERSION >= (10)))
#  include <sys/time.h> // gettimeofday exists in mingw-runtime >= 3.10
# else
#  define WIN_GETTIME
# endif
#else
# include <sys/stat.h>
# include <sys/time.h>
# include <unistd.h>
#endif

#ifdef GNUWINCE
# include <time.h>
# include <sys/time.h>
# undef WIN_GETTIME
#endif

#ifndef GNUWINCE
#define BUILTIN_SYSTEM
#endif

#if !defined(MINGW32)
#define BUILTIN_CONNECTTCPSTREAM
#ifdef GNUWINCE // work around a bug in GNUWINCE
# include <bits/types.h>
#endif /* GNUWINCE */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#endif

#ifdef MINGW32
#define BUILTIN_CONNECTTCPSTREAM
#include <winsock2.h>
#endif

#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <time.h>
#include <errno.h>

#include "stutter.h"

#if defined(__CYGWIN__) || defined(GNUWINCE)
# define timezone _timezone
#endif

#ifdef WIN_GETTIME
struct timezone;

/**
 * windows replacement for gettimeofday
 */
int gettimeofday(struct timeval *tv, struct timezone *tz)
{
	union {
	__int64 ns100; /*time since 1 Jan 1601 in 100ns units */
	FILETIME ft;
	} _now;


	/* substract time from 1 Jan 1601 to 1 Jan 1970 in 100ns units
	 * and convert to microseconds
	 */
	GetSystemTimeAsFileTime(&(_now.ft));
	_now.ns100 -= 116444736000000000LL;
	_now.ns100 /= 10LL;
	tv->tv_sec = (long)(_now.ns100 / 1000000LL);
	tv->tv_usec = (long)(_now.ns100 % 1000000LL);

	return 0;
}
#endif /* #ifdef WIN_GETTIME */

#ifdef REFPATH
const char *_REFPATH = "." SEPSTR REFPATH;
#else
const char *_REFPATH = ".";
#endif

/* get value of next parameter (returns 1 on success, 0 if there are no more) */
inline int nextparm(s_obj **obj, s_obj **parm)
{
	if((*parm)->type != ST_CONS) return 0;
	*obj = (*parm)->d.cons.car;
	*parm = (*parm)->d.cons.cdr;
	return 1;
}

/* get and evaluate next parameter (returns 1 on success, 0 if there are no more) */
inline int nextarg(varctx *ctx, s_obj **obj, s_obj **parm)
{
	if(!nextparm(obj, parm)) return 0;
	*obj = eval(ctx, *obj);
	return 1;
}

/* get and evaluate next parameter (returns 1 on success, 0 if there are no more).
 * if it's the last parameter and tail is true, calls eval
 * with tail flag set to 1! */
inline int nextargtail(varctx *ctx, s_obj **obj, s_obj **parm, const int tail)
{
	if(!nextparm(obj, parm)) return 0;
	*obj = eval_(ctx, *obj, tail && (!(*parm) || (*parm)->type != ST_CONS));
	return 1;
}

/* get and evaluate next parameter, ensure it's from a type
 * returns 1 on success */
int next_t(varctx *ctx, s_obj **obj, s_obj **parm, const int type)
{
	if(!nextarg(ctx, obj, parm)) return 0;
	if((*obj)->type != type) return 0;
	return 1;
}

/* builtin: set */
BUILTIN(sb_set)
{
	s_obj *name, *value = NULL;

	while(parm->type == ST_CONS) {
		nextarg(ctx, &name, &parm);
		if(name->type == ST_ERROR) return name;
		if(name->type != ST_SYMBOL) return create_error(SE_PARM, "Cannot set non-symbol");

		if(!nextarg(ctx, &value, &parm))
			return create_error(SE_PARM, "Insufficient parameters for set");
		if(value->type == ST_ERROR) return value;

		if(value->type == ST_FUNC && !value->d.func.symbol)
			value->d.func.symbol = get_symbolname(name->d.symbol.sdbm, name->d.symbol.id);

		varctx_set_s(ctx, name->d.symbol.sdbm, name->d.symbol.id, value);
	}

	return value ? value : create_nil();
}

/* builtin: setg */
BUILTIN(sb_setg)
{
	s_obj *name, *value = NULL;
	s_var *var;

	while(parm->type == ST_CONS) {
		nextarg(ctx, &name, &parm);
		if(name->type == ST_ERROR) return name;
		if(name->type != ST_SYMBOL) return create_error(SE_PARM, "Cannot set non-symbol");

		if(!nextarg(ctx, &value, &parm))
			return create_error(SE_PARM, "Invalid parameters for setg");
		if(value->type == ST_ERROR) return value;

		while(ctx->parent) {
			var = varctx_retr_s(ctx, name->d.symbol.sdbm, name->d.symbol.id);
			if(var) {
				var->value = value;
				return value;
			}
			ctx = ctx->parent;
		}

		varctx_set_s(ctx, name->d.symbol.sdbm, name->d.symbol.id, value);
	}

	return value ? value : create_nil();
}

/* builtin: scar */
BUILTIN(sb_scar)
{
	ARG_CONS(name);
	ARG_OBJ(value);

	if(data == 0)
		name->d.cons.car = value;
	else
		name->d.cons.cdr = value;

	return value;
}

/* print cons-dump */
s_obj * consdump(varctx *ctx, s_obj *obj)
{
	s_obj *o;

	if(obj->type != ST_CONS) {
		o = stream_printf(ctx, "not a CONS: ");
		if(o->type == ST_ERROR) return o;
		o = printobj(ctx, obj);
		if(o->type == ST_ERROR) return o;
		return stream_stdout_write(ctx, '\n');
	}

	o = stream_stdout_write(ctx, '(');
	if(o->type == ST_ERROR) return o;
	if(obj->d.cons.car && obj->d.cons.car->type == ST_CONS) o = consdump(ctx, obj->d.cons.car);
		else o = printobj(ctx, obj->d.cons.car);
	if(o->type == ST_ERROR) return o;
	o = stream_printf(ctx, " . ");
	if(o->type == ST_ERROR) return o;
	if(obj->d.cons.cdr && obj->d.cons.cdr->type == ST_CONS) o = consdump(ctx, obj->d.cons.cdr);
		else o = printobj(ctx, obj->d.cons.cdr);
	if(o->type == ST_ERROR) return o;
	return stream_stdout_write(ctx, ')');
}

/* builtin: consdump */
BUILTIN(sb_consdump)
{
	ARG_OBJ(obj);

	obj = consdump(ctx, obj);
	if(obj->type == ST_ERROR) return obj;
	return stream_stdout_write(ctx, '\n');
}

/* builtin: cons */
BUILTIN(sb_cons)
{
	ARG_OBJ(car);
	ARG_OBJ(cdr);

	return create_cons(car, cdr);
}

/* builtin: list */
BUILTIN(sb_list)
{
	s_obj *obj, *first, *cur;
	first = cur = NULL;

	while(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_ERROR) return obj;

		if(!cur) first = cur = create_cons(obj, create_nil());
		else {
			cur->d.cons.cdr = create_cons(obj, create_nil());
			cur = cur->d.cons.cdr;
		}
	}

	if(!first) return create_nil();
	return first;
}

/* builtin: append */
BUILTIN(sb_append)
{
	s_obj *obj, *i, *first, *cur;
	first = cur = NULL;

	while(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_ERROR) return obj;

		if(obj->type == ST_CONS) {
			for(i = obj; i && i->type == ST_CONS; i = i->d.cons.cdr) {
				if(!cur) first = cur = create_cons(i->d.cons.car, create_nil());
				else {
					cur->d.cons.cdr = create_cons(i->d.cons.car, create_nil());
					cur = cur->d.cons.cdr;
				}
			}
		} else if(obj->type != ST_NIL) {
			if(!cur) return obj;
			cur->d.cons.cdr = obj;
		}
	}

	if(!first) return create_nil();
	return first;
}

/* builtin: car */
BUILTIN(sb_car)
{
	ARG_OBJ(obj);

	if(obj->type == ST_NIL) return obj; /* car nil is nil */
	if(obj->type != ST_CONS)
		return create_error(SE_PARM, "Non-CONS parameter for car");

	return obj->d.cons.car;
}

/* builtin: cdr */
BUILTIN(sb_cdr)
{
	ARG_OBJ(obj);

	if(obj->type == ST_NIL) return obj; /* cdr nil is nil */
	if(obj->type != ST_CONS)
		return create_error(SE_PARM, "Non-CONS parameter for cdr");

	return obj->d.cons.cdr;
}

/* builtin: atom */
BUILTIN(sb_atom)
{
	ARG_OBJ(obj);

	if(obj->type == ST_CONS)
		return create_nil();
	return create_t();
}

/* builtin: typep */
BUILTIN(sb_typep)
{
	ARG_OBJ(obj);

	if(obj->type == (s_type) data)
		return create_t();
	return create_nil();
}

/* builtin: functionp */
BUILTIN(sb_functionp)
{
	ARG_OBJ(obj);

	if(obj->type == ST_FUNC || obj->type == ST_BUILTIN)
		return create_t();
	return create_nil();
}

/* builtin: listp */
BUILTIN(sb_listp)
{
	ARG_OBJ(obj);

	if(obj->type == ST_CONS) {
		while(obj->d.cons.cdr->type == ST_CONS)
			obj = obj->d.cons.cdr;
		return (obj->d.cons.cdr->type == ST_NIL)
			? create_t() : create_nil();
	}

	if(obj->type == ST_NIL) return create_t();
	return create_nil();
}

/* builtin: length */
BUILTIN(sb_length)
{
	int l = 0;

	ARG_OBJ(obj);

	if(obj->type == ST_STRING)
		return create_int(strlen(obj->d.string.value));

	for(; obj->type == ST_CONS; obj = obj->d.cons.cdr)
		l++;

	return create_int(l);
}

/* builtin: maplist */
BUILTIN(sb_maplist)
{
	s_obj *lists[FUNC_MAXPARMS];
	s_obj *obj, *call, *res = NULL, *calltail, *restail = NULL;
	s_obj *quote;
	s_mark mark;
	int filter = (ptrint)data & 1;
	int car = (ptrint)data & 2;
	int i, lnum;

	ARG_FUNCTION(func);

	for(lnum = 0; lnum < FUNC_MAXPARMS && parm->type == ST_CONS; lnum++) {
		nextarg(ctx, &lists[lnum], &parm);
		if(lists[lnum]->type == ST_ERROR) return lists[lnum];
	}
	if(!lnum)
		return create_error(SE_PNUM, "Insufficient parameters for maplist/filter");

	quote = create_builtin(sb_quote);

	while(1) {
		for(i=0; i<lnum; i++)
			if(lists[i]->type != ST_CONS) break;
		if(i < lnum) break;

		mark = gc_prot_mark();
		calltail = call = create_cons(func, create_nil());
		if(car)
			for(i=0; i<lnum; i++) {
				calltail->d.cons.cdr = create_cons(create_cons(quote, create_cons(lists[i]->d.cons.car, create_nil())), create_nil());
				calltail = calltail->d.cons.cdr;
			}
		else
			for(i=0; i<lnum; i++) {
				calltail->d.cons.cdr = create_cons(create_cons(quote, create_cons(lists[i], create_nil())), create_nil());
				calltail = calltail->d.cons.cdr;
			}

		obj = eval(ctx, call);
		if(obj->type == ST_ERROR) return obj;

		if(filter) {
			if(obj->type != ST_NIL) {
				if(lnum > 1) {
					calltail = call = NULL;
					for(i=0; i<lnum; i++) {
						if(!calltail)
							calltail = call = create_cons(lists[i]->d.cons.car, create_nil());
						else {
							calltail->d.cons.cdr = create_cons(lists[i]->d.cons.car, create_nil());
							calltail = calltail->d.cons.cdr;
						}
					}
				} else call = lists[0]->d.cons.car;

				if(!restail)
					restail = res = create_cons(call, create_nil());
				else {
					restail->d.cons.cdr = create_cons(call, create_nil());
					restail = restail->d.cons.cdr;
				}
			}
		} else {
			if(!restail)
				restail = res = create_cons(obj, create_nil());
			else {
				restail->d.cons.cdr = create_cons(obj, create_nil());
				restail = restail->d.cons.cdr;
			}
		}

		for(i=0; i<lnum; i++)
			lists[i] = lists[i]->d.cons.cdr;

		gc_prot_free(mark, res);
	}

	return res ? res : create_nil();
}

/* builtin: ->num */
s_obj *sb_2num(varctx *ctx, s_obj *parm, void *data, const int tail)
{
	s_obj *obj;
	if(!nextarg(ctx, &obj, &parm))
		return create_error(SE_PNUM, "->num needs a parameter");

	switch(obj->type) {
		case ST_INT:
			return create_num(obj->d.integer.value);
		case ST_NUM:
		case ST_ERROR:
			return obj;
		case ST_STRING:
			return create_num(atof(obj->d.string.value));
		default:
			return create_error(SE_TYPE, "Type can't be converted to number");
	}
}

/* builtin: ->int */
s_obj *sb_2int(varctx *ctx, s_obj *parm, void *data, const int tail)
{
	s_obj *obj;
	if(!nextarg(ctx, &obj, &parm))
		return create_error(SE_PNUM, "->int needs a parameter");

	switch(obj->type) {
		case ST_NUM:
			return create_int((int) obj->d.num.value);
		case ST_INT:
		case ST_ERROR:
			return obj;
		case ST_STRING:
			return create_int(atoi(obj->d.string.value));
		default:
			return create_error(SE_TYPE, "Type can't be converted to integer");
	}
}

/* builtin: ->str */
s_obj *sb_2str(varctx *ctx, s_obj *parm, void *data, const int tail)
{
	s_obj *obj;
	char buf[32], *p;

	if(!nextarg(ctx, &obj, &parm))
		return create_error(SE_PNUM, "->str needs a parameter");

	switch(obj->type) {
		case ST_STRING:
		case ST_ERROR:
			return obj;
		case ST_SYMBOL:
			if(obj->d.symbol.colon) {
				p = (char *) xmalloc(strlen(obj->d.symbol.name) + 1);
				strcpy(p, ":");
				strcat(p, obj->d.symbol.name);
				return create_string_allocd(p);
			} else
				return create_string(obj->d.symbol.name);
		case ST_INT:
			sprintf(buf, "%d", obj->d.integer.value);
			return create_string(buf);
		case ST_CHAR:
			buf[0] = obj->d.ch.ch;
			buf[1] = 0;
			return create_string(buf);
		case ST_NUM:
			sprintf(buf, "%f", obj->d.num.value);
			return create_string(buf);
		case ST_NIL:
			return create_string("nil");
		case ST_T:
			return create_string("T");
		default:
			return create_error_with_obj(SE_TYPE, "Type can't be converted to string", obj);
	}
}

/* builtin: ->symbol */
s_obj *sb_2symbol(varctx *ctx, s_obj *parm, void *data, const int tail)
{
	s_obj *obj;

	if(!nextarg(ctx, &obj, &parm))
		return create_error(SE_PNUM, "->symbol needs a parameter");

	switch(obj->type) {
		case ST_SYMBOL:
		case ST_ERROR:
			return obj;
		case ST_STRING:
			if(obj->d.string.value[0] == ':') {
				obj = create_symbol(obj->d.string.value + 1);
				obj->d.symbol.colon = 1;
				return obj;
			} else
				return create_symbol(obj->d.string.value);
		default:
			return create_error(SE_TYPE, "Type can't be converted to symbol");
	}
}

/* builtin: sb_random */
BUILTIN(sb_random)
{
	s_obj *obj, *o;
	int i;
	double r;

	if(data == 0) {
		/* random */
		if(!nextarg(ctx, &obj, &parm))
			return create_error(SE_PNUM, "random needs a parameter");
		switch(obj->type) {
			case ST_ERROR: return obj;
			case ST_INT:
				if(obj->d.integer.value == 0)
					return create_error(SE_PARM, "Integer parameter for random can't be zero");
				if(obj->d.integer.value < 0)
					return create_int(- rand() % obj->d.integer.value);
				else
					return create_int(rand() % obj->d.integer.value);
			case ST_CONS:
				/* be clever: return random element of list */
				for(o=obj,i=0; o->type == ST_CONS; o=o->d.cons.cdr, i++);
				i = rand() % i;
				for(o=obj; i>0; o=o->d.cons.cdr, i--);
				return o->d.cons.car;
			default: return create_error(SE_TYPE, "Parameter for random must be an integer (or a CONS)");
		}
	} else {
		r = (double) rand() / RAND_MAX;
		if(nextarg(ctx, &obj, &parm)) {
			if(obj->type == ST_ERROR) return obj;
			if(obj->type != ST_NUM)
				return create_error(SE_TYPE, "Parameter for rand must be a float");
			return create_num(r * obj->d.num.value);
		}
		return create_num(r);
	}
}

/* math helper: performs primitive algebraic functions with number polymorphism
 */
s_obj * alg_perform(s_obj *obj, s_obj *arg, int func)
{
	int dtype;
	int obj_int;
	double obj_num;
	int arg_int;
	double arg_num;

	dtype = obj->type;
	switch(dtype) {
		case ST_INT:
			switch(arg->type) {
				case ST_NUM:
					obj_num = obj->d.integer.value;
					arg_num = arg->d.num.value;
					dtype = ST_NUM;
					break;
				case ST_INT:
					obj_int = obj->d.integer.value;
					arg_int = arg->d.integer.value;
					break;
				default:
					return create_error(SE_TYPE, "Type mismatch (algebra arg 2 after integer)");

			}
			break;
		case ST_NUM:
			obj_num = obj->d.num.value;
			switch(arg->type) {
				case ST_INT:
					arg_num = arg->d.integer.value;
					break;
				case ST_NUM:
					arg_num = arg->d.num.value;
					break;
				default:
					return create_error(SE_TYPE, "Type mismatch (algebra arg 2 after float)");
			}
			break;
		default:
			return create_error(SE_TYPE, "Type mismatch (algebra arg 1)");
	}

	switch(dtype) {
		case ST_INT: /* perform on integers */
			switch(func) {
				case '+':
					return create_int(obj_int + arg_int);
				case '-':
					return create_int(obj_int - arg_int);
				case '*':
					return create_int(obj_int * arg_int);
				case '/':
					if(arg_int == 0) return create_error(SE_MATH, "Division by zero");
					return create_int(obj_int / arg_int);
				case '%':
					if(arg_int == 0) return create_error(SE_MATH, "Division by zero");
					return create_int(obj_int % arg_int);

				case '<': /* predicate support: return rvalue on true, NULL on false */
					if(obj_int < arg_int)
						return arg;
					else
						return NULL;

				case '>':
					if(obj_int > arg_int)
						return arg;
					else
						return NULL;

				case '{':
					if(obj_int <= arg_int)
						return arg;
					else
						return NULL;

				case '}':
					if(obj_int >= arg_int)
						return arg;
					else
						return NULL;

				case '=':
					if(obj_int == arg_int)
						return arg;
					else
						return NULL;

				case '&': /* ..let's support bitwise logical operations while we're at it */
					return create_int(obj_int & arg_int);
				case '|':
					return create_int(obj_int | arg_int);
				case '^':
					return create_int(obj_int ^ arg_int);
				default:
					return create_error(SE_TYPE, "Function not supported on integers");
			}

			break;
		case ST_NUM: /* perform on floats */
			switch(func) {
				case '+':
					return create_num(obj_num + arg_num);
				case '-':
					return create_num(obj_num - arg_num);
				case '*':
					return create_num(obj_num * arg_num);
				case '/':
/*					if(arg_num == 0) return create_error(SE_MATH, "Division by zero"); */
					return create_num(obj_num / arg_num);
				case '%':
/*					if(arg_num == 0) return create_error(SE_MATH, "Division by zero"); */
					return create_num(fmod(obj_num, arg_num));

				case '<': /* predicate support: return rvalue on true, NULL on false */
					if(obj_num < arg_num) {
						if(arg->type == ST_INT)
							return create_num(arg_num);
						else
							return arg;
					} else {
						return NULL;
					}

				case '>':
					if(obj_num > arg_num) {
						if(arg->type == ST_INT)
							return create_num(arg_num);
						else
							return arg;
					} else {
						return NULL;
					}

				case '{':
					if(obj_num <= arg_num) {
						if(arg->type == ST_INT)
							return create_num(arg_num);
						else
							return arg;
					} else {
						return NULL;
					}

				case '}':
					if(obj_num >= arg_num) {
						if(arg->type == ST_INT)
							return create_num(arg_num);
						else
							return arg;
					} else {
						return NULL;
					}

				case '=':
					if(obj_num == arg_num) {
						if(arg->type == ST_INT)
							return create_num(arg_num);
						else
							return arg;
					} else {
						return NULL;
					}

				default:
					return create_error(SE_TYPE, "Function not supported on floats");
			}

		default:
			return create_error(SE_TYPE, "Algebra: Type mismatch");
	}
}

/* builtin: sb_lognot */
BUILTIN(sb_lognot)
{
	ARG_INT(obj);

	return create_int(~obj->d.integer.value);
}

/* builtin: sb_logic */
BUILTIN(sb_logic)
{
	s_obj *value, *arg;
	int op;
	switch((ptrint) data) {
		case 0:
		default:
			value = create_int(0);
			op = '|';
			break;
		case 1:
			value = create_int(-1);
			op = '&';
			break;
		case 2:
			value = create_int(0);
			op = '^';
			break;
	}

	while(nextarg(ctx, &arg, &parm)) {
		if(arg->type == ST_ERROR) return arg;
		value = alg_perform(value, arg, op);
		if(value->type == ST_ERROR) return value;
	}

	return value;
}

/* builtin: plus */
BUILTIN(sb_plus)
{
	s_obj *value = create_int(0);
	s_obj *obj;
	char *str = NULL;
	int strl = 0;
	int stra = 0;
	int le;
	int type = -1;

	while(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_ERROR) return obj;
		if(type == -1) {
			type = obj->type;
			if(type == ST_CHAR) {
				strl = 1;
				stra = 3;
				str = (char *) xmalloc(stra);
				str[0] = obj->d.ch.ch;
				str[1] = 0;
				type = ST_STRING;
				continue;
			}

			if(type != ST_STRING && type != ST_NUM && type != ST_INT)
				return create_error(SE_TYPE, "+: Type mismatch");
		}

		if(obj->type == ST_STRING || obj->type == ST_CHAR) {
			if(type != ST_STRING) return create_error(SE_TYPE, "+: Type mismatch, adding to string");

			if(obj->type == ST_CHAR) le = 1; else le = strlen(obj->d.string.value);
			if(!str) {
				strl = le; stra = le * 2 + 1;
				str = (char *) xmalloc(stra);
				strcpy(str, obj->d.string.value);
			} else {
				if(stra < strl + le + 1) {
					do stra *= 2; while(stra < strl + le + 1);
					str = (char *) xrealloc(str, stra);
				}

				if(obj->type == ST_STRING) {
					strcat(str, obj->d.string.value);
				} else {
					str[strl] = obj->d.ch.ch;
					str[strl+1] = 0;
				}

				strl += le;
			}

			continue;
		}

		if(type == ST_STRING) return create_error(SE_TYPE, "+: Type mismatch, adding string to number");

		value = alg_perform(value, obj, '+');
		if(value->type == ST_ERROR) return value;
	}

	if(type == ST_STRING) {
		if(!str) return create_string(""); else {
			obj = create_string_allocd(str);
			return obj;
		}
	} else return value;
}

/* builtin: minus */
BUILTIN(sb_minus)
{
	s_obj *value;
	s_obj *obj;
	int cnt = 0;

	if(!nextarg(ctx, &obj, &parm))
		return create_error(SE_PNUM, "Subtraction requires at least one parameter");
	if(obj->type == ST_ERROR) return obj;

	switch(obj->type) {
		case ST_NUM:
		case ST_INT:
			value = obj;
			break;
		default:
			return create_error(SE_TYPE, "-: Type mismatch");
	}

	while(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_ERROR) return obj;

		value = alg_perform(value, obj, '-');
		if(value->type == ST_ERROR) return value;

		cnt = 1;
	}

	if(!cnt) value = alg_perform(create_int(0), value, '-');
	return value;
}

/* builtin: mul */
BUILTIN(sb_mul)
{
	s_obj *value = create_int(1);
	int ival, j, l;
	s_obj *obj, *nobj;
	char *str, *c, *d;

	while(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_ERROR) return obj;

		if(obj->type == ST_STRING) {
				/* special case: string by integer multiplication? :) */
				if(!nextarg(ctx, &nobj, &parm))
					return create_error(SE_TYPE, "*: Type mismatch (string by int)");
				if(nobj->type == ST_ERROR) return nobj;
				if(nobj->type != ST_INT)
					return create_error(SE_TYPE, "*: Type mismatch (string by int)");

				l = strlen(obj->d.string.value);
				str = (char *) malloc(l * abs(nobj->d.integer.value) + 1);
				if(!str)
					return create_error(SE_MEM, "Not enough memory for new string?");

				c = str;
				if(nobj->d.integer.value < 0) {
					/* reverse string */

					for(ival=nobj->d.integer.value;ival<0;ival++) {
						d = obj->d.string.value + l - 1;
						for(j=0;j<l;j++)
							*c++ = *d--;
					}
					*c++ = 0;

				} else {
					for(ival=0;ival<nobj->d.integer.value;ival++) {
						d = obj->d.string.value;
						for(j=0;j<l;j++)
							*c++ = *d++;
					}
					*c++ = 0;
				}

				obj = create_string_allocd(str);
				return obj;
		}

		if(obj->type != ST_INT && obj->type != ST_NUM) return create_error(SE_TYPE, "*: Type mismatch (arg not number)");

		value = alg_perform(value, obj, '*');
		if(value->type == ST_ERROR) return value;
	}

	return value;
}

/* builtin: div */
BUILTIN(sb_div)
{
	s_obj *value;
	s_obj *obj;
	int cnt = 0;

	if(!nextarg(ctx, &obj, &parm))
		return create_error(SE_PNUM, "Division requires at least one parameter");
	if(obj->type == ST_ERROR) return obj;

	switch(obj->type) {
		case ST_NUM:
		case ST_INT:
			value = obj;
			break;
		default:
			return create_error(SE_TYPE, "/: Type mismatch (arg not number)");
	}

	while(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_ERROR) return obj;

		value = alg_perform(value, obj, '/');
		if(value->type == ST_ERROR) return value;

		cnt = 1;
	}

	if(!cnt) value = alg_perform(create_num(1), value, '/');
	return value;
}

/* builtin: mod */
BUILTIN(sb_mod)
{
	s_obj *value;
	s_obj *obj;
	int cnt = 0;

	if(!nextarg(ctx, &obj, &parm))
		return create_error(SE_PNUM, "Modulo requires at least one parameter");
	if(obj->type == ST_ERROR) return obj;

	switch(obj->type) {
		case ST_NUM:
		case ST_INT:
			value = obj;
			break;
		default:
			return create_error(SE_TYPE, "mod: Type mismatch (arg not number)");
	}

	while(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_ERROR) return obj;

		value = alg_perform(value, obj, '%');
		if(value->type == ST_ERROR) return value;

		cnt = 1;
	}

	if(!cnt) value = alg_perform(create_int(1), value, '%');
	return value;
}

/* builtin: defuse */
BUILTIN(sb_defuse)
{
	s_obj *obj;
	s_obj *b;

	if(!nextarg(ctx, &obj, &parm))
		return create_error(SE_PNUM, "Defuse requires a parameter");
	if(obj->type != ST_BUILTIN) return obj;

	b = create_builtin(obj->d.builtin.handler);
	b->d.builtin.data = obj->d.builtin.data;
	b->d.builtin.nosubst = 1;
	return b;
}

/* builtin: weak-car/weak-cdr */
BUILTIN(sb_weak_car)
{
	ARG_CONS(obj);

	weak_check(((ptrint) data) ? obj->d.cons.cdr : obj->d.cons.car);
	weak_reference(obj, obj->serial, ((ptrint) data) ? &obj->d.cons.cdr : &obj->d.cons.car);
	return obj;
}

/* builtin: weakptr-clean */
BUILTIN(sb_weakptr_clean)
{
	s_obj *obj;

	if(!nextarg(ctx, &obj, &parm))
		return create_error(SE_PNUM, "Weakptr-clean requires a parameter");
	if(obj->type == ST_ERROR) return obj;

	return create_int(weak_check(obj));
}

/* builtin: begin */
BUILTIN(sb_begin)
{
	s_obj *obj = NULL;
	while(nextargtail(ctx, &obj, &parm, tail)) {
		if(obj->type == ST_ERROR) return obj;
	}

	return obj ? obj : create_nil();
}

/* builtin: funcall */
BUILTIN(sb_funcall)
{
	/* get function */
	ARG_OBJ(func);

	switch(func->type) {
	case ST_FUNC:
	case ST_GENERIC:
		return(funcexec(ctx, parm, func, parm, tail, (ptrint)data == 2));
	case ST_BUILTIN:
		return func->d.builtin.handler(ctx, parm, func->d.builtin.data, tail);
	default:
		return create_error(SE_WFUNC, "Invalid function given to funcall.");
	}
}

/* builtin: apply/elev */
BUILTIN(sb_apply)
{
	varctx *tctx;
	s_obj *i;

	tctx = ctx;
	if((ptrint)data == 1 && tctx->parent) tctx = tctx->parent; /* elev runs the function under the parent context */

	/* get function */
	ARG_OBJ(func);

	/* get parameter list */
	if(!nextarg(ctx, &i, &parm))
		return create_error(SE_PNUM, "Insufficient parameters for apply");
	if(i->type == ST_ERROR) return i;
	if(i->type != ST_CONS && i->type != ST_NIL)
		return create_error(SE_PARM, "Second parameter of apply has to be a list");
	parm = i;

	switch(func->type) {
	case ST_FUNC:
	case ST_GENERIC:
		return(funcexec(tctx, parm, func, parm, tail, (ptrint)data == 2));
	case ST_BUILTIN:
		return func->d.builtin.handler(tctx, parm, func->d.builtin.data, tail);
	default:
		return create_error(SE_WFUNC, "Invalid function given to apply.");
	}
}

/* builtin: eval */
BUILTIN(sb_eval)
{
	s_obj *obj, *j = NULL;

	while(nextarg(ctx, &obj, &parm)) {
		j = eval(ctx, compile(ctx, obj)); /* evaluate twice */
		if(j->type == ST_ERROR)
			return j;
	}

	return j ? j : create_nil();
}

/* builtin: trace */
BUILTIN(sb_trace)
{
	s_obj *obj;
	if(!nextarg(ctx, &obj, &parm))
		return create_error(SE_PNUM, "Insufficient parameters for trace");
	if(obj->type == ST_ERROR) return obj;
	if(obj->type != ST_FUNC)
		return create_error(SE_TYPE, "trace requires a function argument");

	obj->d.func.trace = (ptrint) data;
	return obj;
}

/* builtin: not */
BUILTIN(sb_not)
{
	s_obj *obj;

	if(!nextarg(ctx, &obj, &parm))
		return create_error(SE_PNUM, "Not requires at least one parameter");
	if(obj->type == ST_ERROR) return obj;
	if(obj->type == ST_NIL)
		return create_t();
	else
		return create_nil();
}

/* builtin: eq */
BUILTIN(sb_eq)
{
	s_obj *obj, *objo;

	if(!nextparm(&obj, &parm))
		return create_error(SE_PNUM, "eq requires at least one parameter");
	obj = eval(ctx, obj);
	if(obj->type == ST_ERROR) return obj;
	objo = obj;

	while(nextparm(&obj, &parm)) {
		obj = eval(ctx, obj);
		if(obj->type == ST_ERROR) return obj;

		if(obj->type == ST_SYMBOL) {
			if(objo->type != ST_SYMBOL) return create_nil();
			if(objo->d.symbol.id != obj->d.symbol.id) return create_nil();
		} else {
			if(obj != objo) return create_nil();
		}
		objo = obj;
	}

	return create_t();
}

/* builtin: toupper */
BUILTIN(sb_toupper)
{
	char *c;
	int (* conv) (int) = data;

	ARG_STRING(obj);

	obj = create_string(obj->d.string.value); /* this strdups */
	for(c = obj->d.string.value; *c; c++) {
		*c = conv(*c);
	}

	return obj;
}

/* builtin: substr */
BUILTIN(sb_substr)
{
	char *str, *nstr;
	int start = 0, len = -1, le;

	ARG_STRING(obj);

	str = obj->d.string.value;
	le = strlen(str);

	if(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_ERROR) return obj;
		if(obj->type != ST_INT)
			return create_error(SE_TYPE, "substr: Type mismatch (start not int)");
		start = obj->d.integer.value;

		if(nextarg(ctx, &obj, &parm)) {
			if(obj->type == ST_ERROR) return obj;
			if(obj->type != ST_INT)
				return create_error(SE_TYPE, "substr: Type mismatch (len not int)");
			len = obj->d.integer.value;
		}
	}

	if(!len) return create_string("");

	le = strlen(str);
	if(!le) return create_string("");

	if(start < 0) start = le - (0 - start) % le;
	if(start >= le) return create_string("");

	if(len == -1) len = le;
	if(len > (le - start)) len = le - start;
	if(len < 0) len = le + len - start;
	if(len < 0) return create_string("");

	nstr = (char *) xmalloc(len + 1);
	strncpy(nstr, str + start, len);
	*(nstr + len) = 0;

	return create_string_allocd(nstr);
}

/* builtin: explode */
BUILTIN(sb_explode)
{
	char *str, *needle, *p;
	s_obj *list = NULL, *ltail = NULL;
	int limit = -1;
	int lneedle;

	ARG_STRING(obj);

	needle = obj->d.string.value;
	lneedle = strlen(needle);
	if(!lneedle) return create_nil();

	if(!nextarg(ctx, &obj, &parm))
		return create_error(SE_PNUM, "Insufficient parameters for explode");
	if(obj->type == ST_ERROR) return obj;
	if(obj->type != ST_STRING)
		return create_error(SE_TYPE, "explode: Type mismatch (str not string)");
	str = obj->d.string.value;

	if(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_ERROR) return obj;
		if(obj->type != ST_INT)
			return create_error(SE_TYPE, "explode: Type mismatch (limit not int)");
		limit = obj->d.integer.value;
	}

	for(;;) {
		if(!*str) {
			if(!ltail) list = ltail = create_cons(create_string(""), create_nil());
			else {
				ltail->d.cons.cdr = create_cons(create_string(""), create_nil());
				ltail = ltail->d.cons.cdr;
			}
			return list;
		}

		p = strstr(str, needle);

		limit--;
		if(limit == 0 || !p) {
			if(!ltail) list = ltail = create_cons(create_string(str), create_nil());
			else ltail->d.cons.cdr = create_cons(create_string(str), create_nil());
			return list;
		}

		if(!ltail) list = ltail = create_cons(create_string_counted(str, p - str), create_nil());
		else {
			ltail->d.cons.cdr = create_cons(create_string_counted(str, p - str), create_nil());
			ltail = ltail->d.cons.cdr;
		}

		str = p + lneedle;
	}
}

/* builtin: trim */
BUILTIN(sb_trim)
{
	ARG_STRING(str);
	int l = strlen(str->d.string.value);
	char *bf = (char *) xmalloc(l + 1);

	int start,end;
	for(start=0; start<l; start++) {
		if(!isspace(str->d.string.value[start])) break;
	}

	for(end=l-1; end>=0; end--) {
		if(!isspace(str->d.string.value[end])) break;
	}

	if(end == -1) {
		start = 0;
	}

	int i, j = 0;
	for(i = start; i<=end; i++) {
		bf[j++] = str->d.string.value[i];
	}
	bf[j] = 0;

	return create_string_allocd(bf);
}

/* builtin: char->code */
BUILTIN(sb_char_code)
{
	s_obj *obj;
	if(!nextarg(ctx, &obj, &parm))
		return create_error(SE_PNUM, "Insufficient parameters for char-code");
	if(obj->type == ST_ERROR) return obj;
	switch(obj->type) {
		case ST_STRING:
			return create_int((unsigned char) obj->d.string.value[0]);
		case ST_CHAR:
			return create_int((unsigned char) obj->d.ch.ch);
		default:
			return create_error(SE_TYPE, "char->code: Type mismatch (not string)");
	}
}

/* builtin: code->char */
BUILTIN(sb_code_char)
{
	ARG_INT(obj);
	return create_char(obj->d.integer.value);
}

/* builtin: nthchar */
BUILTIN(sb_nthchar)
{
	int i, l;

	ARG_INT(obj);
	ARG_STRING(str);

	l = strlen(str->d.string.value);
	if(!l) return create_error(SE_PARM, "Empty string given to nthchar");
	i = obj->d.integer.value;

	if(i > 0) i = i % l;
	else if(i < 0) i = l - ((0-i) % l);

	return create_char(str->d.string.value[i]);
}

/* builtin: = */
BUILTIN(sb_equal)
{
	s_obj *value = NULL;
	char *str = NULL;
	int res = 0;
	int type;

	ARG_OBJ(obj);

	switch(obj->type) {
		case ST_INT:
			type = obj->type;
			value = obj;
			break;
		case ST_NUM:
			type = obj->type;
			value = obj;
			break;
		case ST_STRING:
			type = ST_STRING;
			str = obj->d.string.value;
			break;
		case ST_SYMBOL:
			type = ST_SYMBOL;
			str = obj->d.symbol.name;
			break;
		case ST_CHAR:
			type = ST_CHAR;
			value = obj;
			break;
		default:
			return create_error(SE_TYPE, "=: Type mismatch (unknown type)");
	}

	while(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_ERROR) return obj;

		switch(obj->type) {
			case ST_NUM:
			case ST_INT:
				if(type != ST_NUM && type != ST_INT)
					return create_error(SE_TYPE, "=: Type mismatch (= num notnum)");

				value = alg_perform(value, obj, (ptrint)data);
				if(value && value->type == ST_ERROR) return value;
				res = (value != NULL);
				break;
			case ST_CHAR:
				if(type != ST_CHAR)
					return create_error(SE_TYPE, "=: Type mismatch (= char notchar)");
				switch((ptrint) data) {
					case '=': res = (value->d.ch.ch == obj->d.ch.ch); break;
					case '<': res = (value->d.ch.ch < obj->d.ch.ch); break;
					case '>': res = (value->d.ch.ch > obj->d.ch.ch); break;
					case '{': res = (value->d.ch.ch <= obj->d.ch.ch); break;
					case '}': res = (value->d.ch.ch >= obj->d.ch.ch); break;
				}
				break;
			case ST_STRING:
				if(type != ST_STRING)
					return create_error(SE_TYPE, "=: Type mismatch (= str notstr)");

				switch((ptrint)data) {
					case '=': res = !strcmp(obj->d.string.value, str); break;
					case '<': res = (strcmp(obj->d.string.value, str) > 0); break;
					case '>': res = (strcmp(obj->d.string.value, str) < 0); break;
					case '{': res = (strcmp(obj->d.string.value, str) >= 0); break;
					case '}': res = (strcmp(obj->d.string.value, str) <= 0); break;
				}
				str = obj->d.string.value;
				break;
			case ST_SYMBOL:
				if(type != ST_SYMBOL)
					return create_error(SE_TYPE, "=: Type mismatch (= sym notsym)");

				switch((ptrint)data) {
					case '=': res = !strcmp(obj->d.symbol.name, str); break;
					case '<': res = (strcmp(obj->d.symbol.name, str) > 0); break;
					case '>': res = (strcmp(obj->d.symbol.name, str) < 0); break;
					case '{': res = (strcmp(obj->d.symbol.name, str) >= 0); break;
					case '}': res = (strcmp(obj->d.symbol.name, str) <= 0); break;
				}
				str = obj->d.symbol.name;
				break;
			default:
				;
		}

		if(!res)
			return create_nil();
	}

	return create_t();
}

/* builtin: eql */
BUILTIN(sb_eql)
{
	s_obj *obj, *o = NULL;

	if(!nextarg(ctx, &o, &parm))
		create_error(SE_PNUM, "Insufficient parameters for eql");
	if(o->type == ST_ERROR) return o;

	while(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_ERROR) return obj;
		if(obj->type != o->type) return create_nil();

		switch(o->type) {
			case ST_INT:
			case ST_NUM:
			case ST_STRING:
			case ST_SYMBOL:
			case ST_CHAR:
				if(!isoform(o, obj)) return create_nil();
				break;
			default:
				if(o != obj) return create_nil();
		}
	}

	return create_t();
}

/* builtin: equal */
BUILTIN(sb_equalop)
{
	s_obj *obj, *o = NULL;

	if(!nextarg(ctx, &o, &parm))
		create_error(SE_PNUM, "Insufficient parameters for equal");
	if(o->type == ST_ERROR) return o;

	while(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_ERROR) return obj;

		if(!isoform(o, obj)) return create_nil();
	}

	return create_t();
}

/* builtin: isset */
BUILTIN(sb_isset)
{
	s_obj *name;
	if(!nextarg(ctx, &name, &parm))
		return create_error(SE_PNUM, "Insufficient parameters for isset");
	if(name->type != ST_SYMBOL)
		return create_error(SE_TYPE, "isset: Type mismatch (not symbol)");

	if(data == 0)
		return varctx_get_s(ctx, name->d.symbol.sdbm, name->d.symbol.id) ? create_t() : create_nil();
	else
		return varctx_retr_s(ctx, name->d.symbol.sdbm, name->d.symbol.id) ? create_t() : create_nil();
}

/* builtin: gc */
BUILTIN(sb_gc)
{
	/* make (gc) do a full sweep */
	gc_gen2_count = 1;
	gc_gen3_count = 1;

	return create_int(gc_sweep(ctx));
}

/* builtin: and */
BUILTIN(sb_and)
{
	s_obj *obj = NULL;

	while(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_NIL || obj->type == ST_ERROR) return obj;
	}

	return obj ? obj : create_nil();
}

/* builtin: or */
BUILTIN(sb_or)
{
	s_obj *obj = NULL;

	while(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_ERROR) return obj;
		if(obj->type != ST_NIL) return obj;
	}

	return obj ? obj : create_nil();
}

/* builtin: dumpctx */
BUILTIN(sb_dumpctx)
{
	return varctx_dump(ctx);
}

/* builtin: limit */
BUILTIN(sb_limit)
{
	s_obj *obj;
	int lim;

	if(!next_t(ctx, &obj, &parm, ST_INT))
		return create_error(SE_PARM, "Invalid parameters for limit");
	lim = recursion_limit;
	recursion_limit = obj->d.integer.value;
	while(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_ERROR) {
			recursion_limit = lim;
			return obj;
		}
	}

	recursion_limit = lim;
	return obj;
}

/* builtin: quit */
BUILTIN(sb_quit)
{
	s_obj *obj = NULL;

	if(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_ERROR) return obj;
		if(obj->type != ST_INT)
			return create_error(SE_TYPE, "quit: Type mismatch (not int)");

		exit(obj->d.integer.value);
	}

	exit(0);
}

/* builtin: load */
BUILTIN(sb_load)
{
	s_obj *obj = NULL;
	char str[512];
	char buf[1024];
	char *st, *e;
	struct stat stb;
	int l;

	if(!((ptrint) data & 1))
		while(ctx->parent) ctx = ctx->parent; /* load globally */

	while(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_ERROR) return obj;
		if(obj->type != ST_STRING && obj->type != ST_SYMBOL)
			return create_error(SE_TYPE, "load: Type mismatch (not string or symbol)");

		/* check len of string to avoid buffer overflows.. */
		st = (obj->type == ST_STRING ? obj->d.string.value : obj->d.symbol.name);
		l = strlen(st);
		if(l>500) return create_error(SE_TYPE, "Too long include >_<");

		if((ptrint) data & 2) { /* called as 'require' */
			strcpy(str, st);
			strcat(str, ".stt");

			st = (char*)_REFPATH;
			while(*st) {
				e = strchr(st, SEPCHAR);
				if(!e) {
					strcpy(buf, st);
				} else {
					strncpy(buf, st, (e - st));
					buf[e - st] = 0;
				}
				strcat(buf, PATHSEP);
				strcat(buf, str);

				if(!stat(buf, &stb)) {
					obj = evalfile(ctx, buf);
					return obj;
				}

				if(e) st = e + 1; else break;
			}

			return create_error(SE_NFND, "Not found");
		} else {
			obj = evalfile(ctx, st); /* simple 'load' */
			if(obj->type == ST_ERROR)
				return obj;
		}
	}

	return obj ? obj : create_nil();
}

/* builtin: cond */
BUILTIN(sb_cond)
{
	s_obj *clause, *truth, *ret = NULL;
	while(nextparm(&clause, &parm)) {
		if(clause->type != ST_CONS)
			return create_error(SE_TYPE, "Cond clause has to be a list");
		if(!nextarg(ctx, &truth, &clause))
			return create_error(SE_PARM, "Missing test for cond clause");
		if(truth->type == ST_ERROR) return truth;
		if(truth->type == ST_NIL) continue;

		ret = truth;
		while(nextargtail(ctx, &ret, &clause, tail))
			if(ret->type == ST_ERROR) return ret;
		return ret;
	}

	return create_nil();
}

/* builtin: while */
BUILTIN(sb_while)
{
	s_obj *truth;
	s_obj *val, *obj = NULL;
	s_mark mark;

	if(!nextparm(&truth, &parm))
		return create_error(SE_PNUM, "Insufficient parameters for while");

	while(1) {
		mark = gc_prot_mark();
		val = eval(ctx, truth);
		if(val->type == ST_NIL) {
			gc_prot_free(mark, val);
			return val;
		}

		if(val->type == ST_ERROR) {
			gc_prot_free(mark, val);
			if(val->d.error.code == SE_RET) return val->d.error.obj ? val->d.error.obj : create_nil();
			return val;
		}

		val = parm;
		while(val->type == ST_CONS) {
			obj = eval(ctx, val->d.cons.car);
			if(obj->type == ST_ERROR) {
				gc_prot_free(mark, obj);
				if(obj->d.error.code == SE_RET) return obj->d.error.obj ? obj->d.error.obj : create_nil();
				return obj;
			}
			val = val->d.cons.cdr;
		}

		gc_prot_free(mark, NULL);
	}
}

/* builtin: throw */
BUILTIN(sb_throw)
{
	s_obj *eobj = NULL;
	int code;
	char *msg = NULL;

	ARG_INT(obj);

	code = obj->d.integer.value;

	if(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_ERROR) return obj;
		if(obj->type != ST_STRING)
			return create_error(SE_TYPE, "Error message needs to be a string");
		msg = obj->d.string.value;

		if(nextarg(ctx, &obj, &parm)) {
			if(obj->type == ST_ERROR) return obj;
			eobj = obj;
		}
	}

	obj = create_error(code, msg);
	obj->d.error.obj = eobj;

	if((ptrint)data == 1) {
		obj->d.error.file = _stutter_file;
		obj->d.error.line = _stutter_parsed_line;
	}

	return obj;
}

/* builtin: return */
BUILTIN(sb_return)
{
	s_obj *obj, *e;
	if(!nextarg(ctx, &obj, &parm))
		obj = create_nil();
	if(obj->type == ST_ERROR) return obj;
	e = create_error(SE_RET, NULL);
	e->d.error.obj = obj;
	return e;
}

/* builtin: print */
BUILTIN(sb_print)
{
	s_obj *obj = NULL;
	s_obj *o;
	int f = 0;
	while(nextarg(ctx, &obj, &parm)) {
		if(f) {
			o = stream_stdout_write(ctx, ' ');
			if(o->type == ST_ERROR) return obj;
		}

		if(obj->type == ST_ERROR) return obj;

		if((ptrint) data == 1) { /* princ */
			if(obj->type == ST_STRING) {
				o = stream_printf(ctx, "%s", obj->d.string.value);
			} else if(obj->type == ST_CHAR) {
				o = stream_stdout_write(ctx, obj->d.ch.ch);
			} else {
				o = printobj(ctx, obj);
			}
		} else { /* print */
			o = printobj(ctx, obj);
		}

		if(o->type == ST_ERROR) return o;

		f = 1;
	}

	return obj ? obj : create_nil();
}

/* builtin: println */
BUILTIN(sb_println)
{
	s_obj *obj;
	s_obj *o;
	obj = sb_print(ctx, parm, data, tail);
	if(obj->type == ST_ERROR) return obj;
	o = stream_stdout_write(ctx, '\n');
	if(o->type == ST_ERROR) return o;
	return obj;
}

/* builtin: read-char */
BUILTIN(sb_read_char)
{
	s_obj *str;

	if(nextarg(ctx, &str, &parm)) {
		if(str->type == ST_ERROR) return str;
		if(str->type != ST_STREAM) return create_error(SE_TYPE, "read-char: Type mismatch (not stream)");
		return stream_read_char(ctx, str);
	}

	return stream_stdin_read(ctx);
}

/* builtin: unread-char */
BUILTIN(sb_unread_char)
{
	ARG_OBJ(str);
	ARG_CHAR(chr);

	return stream_unread_char(str, chr->d.ch.ch);
}

/* builtin: read-line */
BUILTIN(sb_read_line)
{
	s_obj *str;
	s_obj *stream = NULL;
	s_obj *o;
	int le = 1, l = 0;
	str = create_string("");

	if(nextarg(ctx, &stream, &parm)) {
		if(stream->type == ST_ERROR) return stream;
		if(stream->type != ST_STREAM) return create_error(SE_TYPE, "read-line takes a stream parameter");
	}

	while(1) {
		if(stream) o = stream_read_char(ctx, stream);
		else o = stream_stdin_read(ctx);

		if(o->type == ST_ERROR)
			return o;
		if(o->type == ST_EOF) {
			if(l > 0) {
				str->d.string.value[l] = 0;
				break;
			} else return o;
		}

		if(le == l + 1) {
			le = le * 2 + 1;
			str->d.string.value = realloc(str->d.string.value, le);
			if(!str->d.string.value) return create_error(SE_MEM, "Out of memory");
		}

		str->d.string.value[l++] = o->d.ch.ch;
		if(o->d.ch.ch == '\n') {
			str->d.string.value[l-1] = 0;
			break;
		}
	}

	return str;
}

/* builtin: read */
BUILTIN(sb_read)
{
	int ol;
	read_state *rs;
	s_obj *o;
	s_obj *stream = NULL;
	int c;

	if(nextarg(ctx, &stream, &parm)) {
		if(stream->type == ST_ERROR) return stream;
		if(stream->type != ST_STREAM) return create_error(SE_TYPE, "read-line takes a stream parameter");
	}


	ol = _stutter_parsed_line;
	rs = create_read_state();

	for(;;) {
		if(stream) o = stream_read_char(ctx, stream);
		else o = stream_stdin_read(ctx);

		if(o->type == ST_ERROR) {
			free_read_state(rs);
			_stutter_parsed_line = ol;
			return o;
		}
		if(o->type == ST_EOF)
			c = EOF;
		else
			c = o->d.ch.ch;

		o = read_ch(rs, c);
		if(o && o->type == ST_ERROR) {
			free_read_state(rs);
			_stutter_parsed_line = ol;
			return o;
		}

		if(o || c == EOF) break;
	}

	free_read_state(rs);
	_stutter_parsed_line = ol;

	if(c == EOF)
		return create_eof();

	return o ? o : create_nil();
}

/* builtin: make-stream */
BUILTIN(sb_make_stream)
{
	s_obj *obj;
	s_obj *str;

	if(!nextarg(ctx, &obj, &parm))
		return create_error(SE_PNUM, "make-stream requires a parameter");
	if(obj->type == ST_ERROR) return obj;
	if(obj->type != ST_FUNC || obj->d.func.macro > 0)
		return create_error(SE_TYPE, "make-stream requires a lambda parameter");

	str = create_stream(STREAM_FUNC);
	str->d.stream.s.func = obj;
	return str;
}

/* builtin: open-file */
BUILTIN(sb_open_file)
{
	s_obj *mode;
	FILE *fp;

	ARG_STRING(obj);

	if(!nextarg(ctx, &mode, &parm)) {
		fp = fopen(obj->d.string.value, "r");
	} else {
		if(mode->type == ST_ERROR) return mode;
		if(mode->type != ST_STRING)
			return create_error(SE_TYPE, "open-file file mode has to be a string");

		fp = fopen(obj->d.string.value, mode->d.string.value);
	}

	if(!fp)
		return create_error_with_obj(SE_IO, "Cannot open file", create_string(strerror(errno)));
	obj = create_stream(STREAM_FP);
	obj->d.stream.s.fp = fp;
	return obj;
}

/* builtin: close-file */
BUILTIN(sb_close_file)
{
	ARG_STREAM(obj);

	if(obj->d.stream.type != STREAM_FP)
		return create_error(SE_TYPE, "Invalid stream (not opened with open-file?)");
	if(!obj->d.stream.s.fp)
		return create_error(SE_IO, "Stream already closed");
	if(fclose(obj->d.stream.s.fp))
		return create_error_with_obj(SE_IO, "Stream close failed", create_string(strerror(errno)));
	obj->d.stream.s.fp = NULL;

	return create_t();
}

/* stream writer for string buckets */
int bucket_writer(int ch, void **data, const int type)
{
	int buf, len;
	if(type != 0) return -1;

	if(!*data) {
		*data = xcalloc(1, 8 + 2 * sizeof(int));
		buf = *((int *) *data) = 8;
		len = *(((int *) *data) + 1) = 1;
	} else {
		buf = *((int *) *data);
		len = *(((int *) *data) + 1);
	}

	if(len == buf) {
		buf *= 2;
		*data = xrealloc(*data, buf + 2 * sizeof(int));
		*((int *) *data) = buf;
	}

	((char *) (*data)) [(2 * sizeof(int)) + len] = 0;
	((char *) (*data)) [(2 * sizeof(int)) + ((len++) - 1)] = ch;

	*(((int *) *data) + 1) = len;

	return 0;
}

/* builtin: string-bucket */
BUILTIN(sb_string_bucket)
{
	s_obj *stream = create_stream(STREAM_C);
	stream->d.stream.s.c.type = 0;
	stream->d.stream.s.c.data = NULL;
	stream->d.stream.s.c.reader = NULL;
	stream->d.stream.s.c.writer = bucket_writer;
	return stream;
}

/* builtin: empty-bucket */
BUILTIN(sb_empty_bucket)
{
	s_obj *str;

	ARG_STREAM(obj);

	if(obj->d.stream.type != STREAM_C || obj->d.stream.s.c.type != 0)
		return create_error(SE_TYPE, "Invalid type of stream");
	if(!obj->d.stream.s.c.data)
		return create_string("");
	str = create_string((char *) (obj->d.stream.s.c.data + 2*sizeof(int)));
	free(obj->d.stream.s.c.data);
	obj->d.stream.s.c.data = NULL;
	return str;
}

#ifdef BUILTIN_CONNECTTCPSTREAM

/* stream reader for tcp streams */
int tcpstream_reader(void **data, const int type)
{
	char ch;

	if(type != 1) return -1;
	if(!*data) return -1;

#ifdef MINGW32
	ssize_t rd;

	rd = recv(*((int *) *data), &ch, 1, 0);
	if(ch == -1) return -1;
	return ch;
#else
//	rd = read(*((int *) *data), &ch, 1);

    if(feof((FILE *) *data)) return -1;
    ch = fgetc((FILE *) *data);
	if(ch == EOF) return -1;
	return ch;
#endif
}

/* stream writer for tcp streams */
int tcpstream_writer(int ch, void **data, const int type)
{
	char c = ch;
	if(type != 1) return -1;
	if(!*data) return -1;

#ifdef MINGW32
	int retry = 5;
	ssize_t wr;

	do {
		retry --;
		if(retry < 0) return -1;

		wr = send(*((int *) *data), &c, 1, 0);
		if(wr < 0) return -1;
	} while(wr == 0);
#else

	return fputc(c, (FILE *) *data) == EOF ? -1 : 0;

#endif

	return 0;
}

/* builtin: connect-tcpstream */
BUILTIN(sb_connect_tcpstream)
{
	char *host;
	int port;
	int sock;
	struct hostent *server;
	struct sockaddr_in serv_addr;
#ifndef MINGW32
	struct in_addr inp;
#endif

	ARG_STRING(obj);

	host = obj->d.string.value;

	if(!nextarg(ctx, &obj, &parm))
		return create_error(SE_PNUM, "Insufficient parameters for connect-tcpstream");
	if(obj->type == ST_ERROR) return obj;
	if(obj->type != ST_INT)
		return create_error(SE_TYPE, "connect-tcpstream: Type mismatch (not int)");
	port = obj->d.integer.value;

	sock = socket(AF_INET, SOCK_STREAM, 0);
#ifdef MINGW32
	if(sock == INVALID_SOCKET)
#else
	if(sock < 0)
		return create_error_with_obj(SE_IO, "Error opening socket", create_string(strerror(errno)));
#endif

	memset((void *) &serv_addr, 0, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_port = htons(port);

#ifndef MINGW32
	if(inet_aton(host, &inp)) {
		serv_addr.sin_addr = inp;
	} else {
#endif
		server = gethostbyname(host);
		if(server == NULL)
			return create_error_with_obj(SE_IO, "DNS lookup error", create_string(strerror(errno)));
		memcpy((void *) &serv_addr.sin_addr.s_addr, (void *) server->h_addr, server->h_length);
#ifndef MINGW32
	}
#endif

	if(connect(sock, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
		return create_error_with_obj(SE_IO, "Couldn't connect", create_string(strerror(errno)));

	obj = create_stream(STREAM_C);
	obj->d.stream.s.c.type = 1;

#ifdef MINGW32
	obj->d.stream.s.c.data = xmalloc(sizeof(int));
	*((int *) obj->d.stream.s.c.data) = sock;
#else
	obj->d.stream.s.c.data = fdopen(sock, "r+");
#endif

	obj->d.stream.s.c.reader = tcpstream_reader;
	obj->d.stream.s.c.writer = tcpstream_writer;
	return obj;
}

/* builtin: close-tcpstream */
BUILTIN(sb_close_tcpstream)
{
	ARG_STREAM(stream);

	if(stream->d.stream.type != STREAM_C || stream->d.stream.s.c.type != 1)
		return create_error(SE_TYPE, "Invalid type of stream");
	if(!stream->d.stream.s.c.data)
		return create_error(SE_IO, "Stream already closed");

#ifdef MINGW32
	int sock = *((int *) stream->d.stream.s.c.data);
	free(stream->d.stream.s.c.data);
	stream->d.stream.s.c.data = NULL;

	if(closesocket(sock) != 0)
		return create_error_with_obj(SE_IO, "Error closing stream", create_string(strerror(errno)));
#else
	int res = fclose((FILE *) stream->d.stream.s.c.data);
	stream->d.stream.s.c.data = NULL;
	if(res == EOF)
		return create_error_with_obj(SE_IO, "Error closing stream", create_string(strerror(errno)));
#endif

	return create_nil();
}
#endif /* #ifdef BUILTIN_CONNECTTCPSTREAM */


#ifdef BUILTIN_SYSTEM

/* builtin: system */
BUILTIN(sb_system)
{
	ARG_STRING(command);

	return create_int(system(command->d.string.value));
}

#endif

/* array loader (make-array helper) */
s_obj *arrayloader(s_obj *ar, s_obj **addr, int dim, s_obj *list) {
	int i;
	s_obj *val;
	int ds;

	ds = 1; /* compute offset */
	for(i=0; i<dim; i++)
		ds *= ar->d.array.size[i];

	if(dim == ar->d.array.dim-1) {
		/* final level: load elements */

		for(i=0; i<ar->d.array.size[dim]; i++) {
			if(!nextparm(&val, &list))
				return create_error(SE_PARM, "Initializer comes short");

			if(val->type == ST_ERROR) return val;
			*addr = val;

			addr += ds;
		}
	} else {
		/* there are more dimensions: recursion */

		for(i=0; i<ar->d.array.size[dim]; i++) {
			if(!nextparm(&val, &list))
				return create_error(SE_PARM, "Initializer comes short");
			if(val->type != ST_CONS)
				return create_error(SE_TYPE, "Initializer must be a list");
			val = arrayloader(ar, addr, dim + 1, val);
			if(val->type == ST_ERROR) return val;

			addr += ds;
		}
	}

	return ar;
}

/* builtin: make-array */
BUILTIN(sb_make_array)
{
	s_obj *pl, *ar;
	int size[ARRAY_MAXDIMS];
	int dim = 0;

	ARG_CONS(obj);

	pl = obj;
	while(nextparm(&obj, &pl)) {
		if(dim == ARRAY_MAXDIMS)
			return create_error(SE_PARM, "Array has too many dimensions");

		if(obj->type != ST_INT)
			return create_error(SE_TYPE, "Array dimensions must be integers");
		if(obj->d.integer.value < 1)
			return create_error(SE_PARM, "Array dimension must be >= 1");

		size[dim++] = obj->d.integer.value;
	}

	if(!dim)
		return create_error(SE_PARM, "Can't create zero-dimensional array");

	ar = create_array(dim, size);
	if(ar->type == ST_ERROR) return ar; /* create_array might throw SE_MEM */

	if(nextarg(ctx, &obj, &parm)) {
		/* initializer present */

		if(obj->type == ST_ERROR) return obj;
		if(obj->type != ST_CONS)
			return create_error(SE_TYPE, "Initializer must be a list");

		obj = arrayloader(ar, ar->d.array.data, 0, obj);
		if(obj->type == ST_ERROR) return obj;
	}

	return ar;
}

/* builtin: adim */
BUILTIN(sb_adim)
{
	s_obj *l = NULL, *lt = NULL;
	int i;

	ARG_ARRAY(ar);

	for(i=0; i<ar->d.array.dim; i++) {
		if(!lt) l = lt = create_cons(create_int(ar->d.array.size[i]), create_nil());
		else lt = lt->d.cons.cdr = create_cons(create_int(ar->d.array.size[i]), create_nil());
	}

	return l;
}

/* builtin: aset */
BUILTIN(sb_aset)
{
	s_obj *obj, *val;
	s_obj **addr;
	int ds = 1, i;

	ARG_ARRAY(ar);

	addr = ar->d.array.data;
	for(i=0; i<ar->d.array.dim; i++) {
		if(!nextarg(ctx, &obj, &parm))
			return create_error(SE_PNUM, "Insufficient array indices for aset");
		if(obj->type == ST_ERROR) return obj;
		if(obj->type != ST_INT)
			return create_error(SE_TYPE, "aset index must be an integer");
		if(obj->d.integer.value < 0 || obj->d.integer.value >= ar->d.array.size[i])
			return create_error(SE_PNUM, "aset index out of bounds");
		addr += (ds * obj->d.integer.value);
		ds *= ar->d.array.size[i];
	}

	if(!nextarg(ctx, &val, &parm))
		return create_error(SE_PNUM, "Insufficient parameters for aset");
	if(val->type == ST_ERROR) return val;

	*addr = val;

	if((ptrint)data == 1) {
		/* weak-aset */
		weak_check(val);
		weak_reference(ar, ar->serial, addr);
	}

	return val;
}

/* builtin: aref */
BUILTIN(sb_aref)
{
	ARG_ARRAY(ar);
	s_obj *obj;
	s_obj **addr;
	int ds = 1, i;

	addr = ar->d.array.data;
	for(i=0; i<ar->d.array.dim; i++) {
		if(!nextarg(ctx, &obj, &parm))
			return create_error(SE_PNUM, "Insufficient array indices for aref");
		if(obj->type == ST_ERROR) return obj;
		if(obj->type != ST_INT)
			return create_error(SE_TYPE, "aref index must be an integer");
		if(obj->d.integer.value < 0 || obj->d.integer.value >= ar->d.array.size[i])
			return create_error(SE_PNUM, "aref index out of bounds");
		addr += (ds * obj->d.integer.value);
		ds *= ar->d.array.size[i];
	}

	if(!*addr) return create_nil();
	return *addr;
}

/* builtin: make-dictionary */
BUILTIN(sb_make_dictionary)
{
	return create_dict();
}

/* builtin: dictionary-keys */
BUILTIN(sb_dictionary_keys)
{
	ARG_DICT(dict);
	s_obj *list;
	int i;

	list = create_nil();
	for(i=0; i<dict->d.dict.size; i++) {
		if(dict->d.dict.data[i].name && (ptrint)dict->d.dict.data[i].name != 1)
			list = create_cons(dict->d.dict.data[i].name, list);
	}

	return list;
}

/* builtin: dref */
BUILTIN(sb_dref)
{
	ARG_DICT(dict);
	ARG_OBJ(key);

	s_obj *value;

	value = dictget(dict, key);
	return value ? value : create_nil();
}

/* builtin: ddel */
BUILTIN(sb_ddel)
{
	ARG_DICT(dict);
	ARG_OBJ(key);

	return dictdel(dict, key) ? create_t() : create_nil();
}

/* builtin: dput */
BUILTIN(sb_dput)
{
	ARG_DICT(dict);
	ARG_OBJ(key);
	ARG_OBJ(value);

	dictput(dict, hashfunc(key), key, value);
	return create_t();
}

/* builtin: get-time */
BUILTIN(sb_get_time)
{
	return create_int(time(NULL));
}

/* builtin: get-ftime */
BUILTIN(sb_get_ftime)
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return create_num(tv.tv_sec + (double) tv.tv_usec / 1000000);
}

/* builtin: get-decoded-time */
BUILTIN(sb_get_decoded_time)
{
	s_obj *obj;
	time_t t;
	struct tm *gm;

	if(nextarg(ctx, &obj, &parm)) {
		if(obj->type == ST_ERROR) return obj;
		if(obj->type != ST_INT)
			return create_error(SE_TYPE, "Parameter of get-decoded-time must be an integer");
		t = obj->d.integer.value;
	} else t = time(NULL);

	gm = localtime(&t);

	return create_cons(create_int(gm->tm_sec),
			create_cons(create_int(gm->tm_min),
			create_cons(create_int(gm->tm_hour),
			create_cons(create_int(gm->tm_mday),
			create_cons(create_int(gm->tm_mon + 1),
			create_cons(create_int(gm->tm_year + 1900),
			create_cons(create_int(gm->tm_wday == 0 ? 6 : gm->tm_wday - 1),
			create_cons(gm->tm_isdst ? create_t() : create_nil(),
			create_cons(create_int(-timezone / 3600), create_nil())))))))));
}

/* builtin: sleep */
BUILTIN(sb_sleep)
{
	ARG_INT(tim);

	#ifndef WIN32
	sleep(tim->d.integer.value);
	#else
	Sleep(tim->d.integer.value * 1000);
	#endif

	return tim;
}

/*************************** CLOS BUILTINS ***/

/* builtin: defclass */
BUILTIN(sb_defclass)
{
	s_obj *name;
	s_obj *parentlist;
	s_obj *slotlist;
	s_obj *cl;
	s_var *var;
	s_obj *iter, *iter2;
	s_obj *o, *sl;

	s_obj *crsr;
	slotasg *asg;
	int i;

	unsigned long _ACCESSOR = symbolid("ACCESSOR");
	unsigned long _WRITER = symbolid("WRITER");
	unsigned long _READER = symbolid("READER");
	unsigned long _ALLOCATION = symbolid("ALLOCATION");
	unsigned long _INITFORM = symbolid("INITFORM");
	unsigned long _INITARG = symbolid("INITARG");
	unsigned long _CLASS = symbolid("CLASS");
	unsigned long _INSTANCE = symbolid("INSTANCE");

	/* retrieve arguments */

	if(!nextparm(&name, &parm))
		return create_error(SE_PNUM, "Defclass needs parameter: NAME");
	if(name->type != ST_SYMBOL)
		return create_error(SE_TYPE, "Defclass: NAME must be a symbol");

	if(nextparm(&parentlist, &parm)) {

		if(parentlist->type != ST_NIL && parentlist->type != ST_CONS)
			return create_error(SE_TYPE, "Defclass: PARENTLIST must be a list");

		if(nextparm(&slotlist, &parm)) {

			if(slotlist->type != ST_NIL && slotlist->type != ST_CONS)
				return create_error(SE_TYPE, "Defclass: SLOTLIST must be a list");

		} else {
			slotlist = create_nil();
		}

	} else {
		parentlist = create_nil();
		slotlist = create_nil();
	}

	/* check whether class of this name already exists */

	var = varctx_retr_s(class_hash, name->d.symbol.sdbm, name->d.symbol.id);
	if(var) {
		return create_error(SE_PARM, "Defclass: class redefinition isn't supported");
	}

	/* create class object */

	cl = create_class();
	cl->d.clos_class.typeid = gettypeid();
	cl->d.clos_class.name = name;
	typedb[cl->d.clos_class.typeid].cl = cl;

	varctx_set_s(class_hash, name->d.symbol.sdbm, name->d.symbol.id, cl);

	/* look up each element in parent list and add to class object */

	crsr = NULL;
	iter = parentlist;
	while(nextparm(&o, &iter)) {
		if(o->type != ST_SYMBOL)
			return create_error(SE_TYPE, "Defclass: parent name has to be a SYMBOL");

		var = varctx_retr_s(class_hash, o->d.symbol.sdbm, o->d.symbol.id);
		if(!var)
			return create_error(SE_PARM, "Defclass: parent class undefined");

		if(crsr) {
			crsr->d.cons.cdr = create_cons(var->value, create_nil());
			crsr = crsr->d.cons.cdr;
		} else {
			crsr = cl->d.clos_class.parents = create_cons(var->value, create_nil());
		}
	}

	/* parse slot definitions, add them to class object */

	crsr = NULL;
	iter = slotlist;
	while(nextparm(&o, &iter)) {
		sl = create_slot();
		sl->d.clos_slot.alloc = CL_ALLOC_INSTANCE;

		if(o->type == ST_SYMBOL) {
			/* plain slot definition */
			sl->d.clos_slot.name = o;
		} else if (o->type == ST_CONS) {
			/* complicated slot definition */

			iter2 = o;
			nextparm(&o, &iter2);

			if(o->type != ST_SYMBOL)
				return create_error(SE_TYPE, "Defclass: slot name has to be a symbol");
			sl->d.clos_slot.name = o;

			/* read slot option list */

			while(nextparm(&o, &iter2)) {
				if(o->type != ST_SYMBOL)
					return create_error(SE_TYPE, "Defclass: invalid slot definition (slot option)");

				if(o->d.symbol.id == _ACCESSOR) {

					if(!nextparm(&o, &iter2))
						return create_error(SE_PNUM, "Defclass: unexpected end of slot option list (after :ACCESSOR)");
					if(o->type != ST_SYMBOL)
						return create_error(SE_TYPE, "Defclass: accessor name has to be a symbol");
					sl->d.clos_slot.accessor = o;
					continue;

				} else if(o->d.symbol.id == _WRITER) {

					if(!nextparm(&o, &iter2))
						return create_error(SE_PNUM, "Defclass: unexpected end of slot option list (after :WRITER)");
					if(o->type != ST_SYMBOL)
						return create_error(SE_TYPE, "Defclass: writer name has to be a symbol");
					sl->d.clos_slot.writer = o;
					continue;

				} else if(o->d.symbol.id == _READER) {

					if(!nextparm(&o, &iter2))
						return create_error(SE_PNUM, "Defclass: unexpected end of slot option list (after :READER)");
					if(o->type != ST_SYMBOL)
						return create_error(SE_TYPE, "Defclass: reader name has to be a symbol");
					sl->d.clos_slot.reader = o;
					continue;

				} else if(o->d.symbol.id == _INITARG) {

					if(!nextparm(&o, &iter2))
						return create_error(SE_PNUM, "Defclass: unexpected end of slot option list (after :INITARG)");
					if(o->type != ST_SYMBOL)
						return create_error(SE_TYPE, "Defclass: initarg name has to be a symbol");
					sl->d.clos_slot.initarg = o;
					continue;

				} else if(o->d.symbol.id == _ALLOCATION) {

					if(!nextparm(&o, &iter2))
						return create_error(SE_PNUM, "Defclass: unexpected end of slot option list (after :ALLOCATION)");
					if(o->type != ST_SYMBOL)
						return create_error(SE_TYPE, "Defclass: allocation type has to be a symbol");

					if(o->d.symbol.id == _CLASS)
						sl->d.clos_slot.alloc = CL_ALLOC_CLASS;
					else if(o->d.symbol.id == _INSTANCE)
						sl->d.clos_slot.alloc = CL_ALLOC_INSTANCE;
					else
						return create_error(SE_PARM, "Defclass: invalid allocation type");
					continue;

				} else if(o->d.symbol.id == _INITFORM) {

					if(!nextparm(&o, &iter2))
						return create_error(SE_PNUM, "Defclass: unexpected end of slot option list (after :INITFORM)");

					sl->d.clos_slot.initform = o;
					continue;

				} else {
					return create_error(SE_PARM, "Defclass: invalid slot option");
				}

			} /* slot option list */


		} else {
			return create_error(SE_PARM, "Defclass: invalid slot definition");
		}

		/* add ready slot object to class slot list */

		if(crsr) {
			crsr->d.cons.cdr = create_cons(sl, create_nil());
			crsr = crsr->d.cons.cdr;
		} else {
			crsr = cl->d.clos_class.slots = create_cons(sl, create_nil());
		}

	} /* slot definition list */

	/* update internal data */
	o = class_calculate(cl);
	if(o->type == ST_ERROR) return o;

	/* evaluate initforms for static slots */

	for(i=0, asg = cl->d.clos_class.allslots; i < cl->d.clos_class.allslotsnum; i++, asg++) {
		if(asg->stor) {
			if(!cl->d.clos_class.data[asg->num]) {
				o = eval(ctx, asg->slot->d.clos_slot.initform);
				if(o->type == ST_ERROR) return o;
				cl->d.clos_class.data[asg->num] = o;
			}
		}
	}

	return name;
}

/* builtin: class-precedence */
BUILTIN(sb_class_precedence)
{
	s_obj *first = NULL, *ltail = NULL;
	s_obj *o, *iter;

	iter = class_precedence;
	while(nextparm(&o, &iter)) {
		if(ltail) {
			ltail->d.cons.cdr = create_cons(o->d.clos_class.name, create_nil());
			ltail = ltail->d.cons.cdr;
		} else {
			first = ltail = create_cons(o->d.clos_class.name, create_nil());
		}
	}

	return first ? first : create_nil();
}

/* builtin: class-ancestry */
BUILTIN(sb_class_ancestry)
{
	s_var *var;
	s_obj *iter;
	s_obj *first = NULL, *ltail = NULL;

	ARG_SYMBOL(obj);

	var = varctx_retr_s(class_hash, obj->d.symbol.sdbm, obj->d.symbol.id);
	if(!var)
		return create_error(SE_UNDEF, "Class-ancestry: undefined class");

	iter = var->value->d.clos_class.ancestry;
	while(nextparm(&obj, &iter)) {
		if(ltail) {
			ltail->d.cons.cdr = create_cons(obj->d.clos_class.name, create_nil());
			ltail = ltail->d.cons.cdr;
		} else {
			first = ltail = create_cons(obj->d.clos_class.name, create_nil());
		}
	}

	return first ? first : create_nil();
}

/* builtin: class-slotinfo */
BUILTIN(sb_class_slotinfo)
{
	s_var *var;
	int i;
	s_obj *first = NULL, *ltail = NULL;
	s_obj *info;
	slotasg *asg;

	ARG_SYMBOL(obj);

	var = varctx_retr_s(class_hash, obj->d.symbol.sdbm, obj->d.symbol.id);
	if(!var)
		return create_error(SE_UNDEF, "Class-slotinfo: undefined class");

	obj=var->value;

	for(i=0; i<obj->d.clos_class.allslotsnum; i++) {
		asg = obj->d.clos_class.allslots + i;
		info = create_nil();
		if(asg->slot->d.clos_slot.accessor)
			info = create_cons(create_cons(create_symbol("ACCESSOR"), asg->slot->d.clos_slot.accessor), info);
		if(asg->slot->d.clos_slot.writer)
			info = create_cons(create_cons(create_symbol("WRITER"), asg->slot->d.clos_slot.writer), info);
		if(asg->slot->d.clos_slot.reader)
			info = create_cons(create_cons(create_symbol("READER"), asg->slot->d.clos_slot.reader), info);

		info = create_cons(asg->slot->d.clos_slot.name, info);

		if(ltail) {
			ltail->d.cons.cdr = create_cons(info, create_nil());
			ltail = ltail->d.cons.cdr;
		} else {
			first = ltail = create_cons(info, create_nil());
		}
	}

	return first ? first : create_nil();
}

/* builtin: slot-value */
BUILTIN(sb_slot_value)
{
	s_obj *type;
	s_var *var;
	slotasg *asg;

	ARG_OBJECT(obj);
	ARG_SYMBOL(slot);

	type = obj->d.clos_object.type;
	var = varctx_retr_s(type->d.clos_class.slotname, slot->d.symbol.sdbm, slot->d.symbol.id);
	if(!var)
		return create_error(SE_UNDEF, "Slot-value: undefined slot");

	asg = type->d.clos_class.allslots + var->value->d.integer.value;

	if(asg->stor) {
		return asg->stor->d.clos_class.data[asg->num];
	} else {
		return obj->d.clos_object.data[asg->num];
	}
}

/* builtin: slot-value-set */
BUILTIN(sb_slot_value_set)
{
	s_obj *type;
	s_var *var;
	slotasg *asg;

	ARG_OBJECT(obj);
	ARG_SYMBOL(slot);
	ARG_OBJ(value);

	type = obj->d.clos_object.type;
	var = varctx_retr_s(type->d.clos_class.slotname, slot->d.symbol.sdbm, slot->d.symbol.id);
	if(!var)
		return create_error(SE_UNDEF, "Slot-value-set: undefined slot");

	asg = type->d.clos_class.allslots + var->value->d.integer.value;

	if(asg->stor) {
		asg->stor->d.clos_class.data[asg->num] = value;
	} else {
		obj->d.clos_object.data[asg->num] = value;
	}

	return value;
}

/* builtin: slot-value-set-weak */
BUILTIN(sb_slot_value_set_weak)
{
	s_obj *type;
	s_var *var;
	slotasg *asg;
	s_obj **pntr;

	ARG_OBJECT(obj);
	ARG_SYMBOL(slot);
	ARG_OBJ(value);

	type = obj->d.clos_object.type;
	var = varctx_retr_s(type->d.clos_class.slotname, slot->d.symbol.sdbm, slot->d.symbol.id);
	if(!var)
		return create_error(SE_UNDEF, "Slot-value-set-weak: undefined slot");

	asg = type->d.clos_class.allslots + var->value->d.integer.value;

	if(asg->stor) {
		pntr = asg->stor->d.clos_class.data + asg->num;
		*pntr = value;
		weak_check(value);
		weak_reference(asg->stor, asg->stor->serial, pntr);
	} else {
		pntr = obj->d.clos_object.data + asg->num;
		*pntr = value;
		weak_check(value);
		weak_reference(obj, obj->serial, pntr);
	}

	return value;
}

/* builtin: make-instance */
BUILTIN(sb_make_instance)
{
	s_obj *type;
	s_obj *inst;
	s_var *var;
	s_obj *name, *value;
	slotasg *asg;
	int i;

	if(!nextarg(ctx, &type, &parm))
		return create_error(SE_PNUM, "Insufficient parameters for make-instance");
	if(type->type == ST_ERROR) return type;
	if(type->type != ST_SYMBOL)
		return create_error(SE_TYPE, "make-instance: type mismatch (not symbol)");

	var = varctx_retr_s(class_hash, type->d.symbol.sdbm, type->d.symbol.id);
	if(!var)
		return create_error(SE_UNDEF, "make-instance: undefined class");
	type = var->value;

	inst = create_object(type);

	/* parse initargs */

	while(nextarg(ctx, &name, &parm)) {
		if(name->type == ST_ERROR) return name;
		if(name->type != ST_SYMBOL)
			return create_error(SE_TYPE, "make-instance: initarg name must be a symbol");

		if(!nextarg(ctx, &value, &parm))
			return create_error(SE_PNUM, "make-instance: missing value after initarg name");
		if(value->type == ST_ERROR) return value;

		var = varctx_retr_s(type->d.clos_class.initarg, name->d.symbol.sdbm, name->d.symbol.id);
		if(!var)
			return create_error(SE_UNDEF, "make-instance: initarg undefined");

		asg = type->d.clos_class.allslots + var->value->d.integer.value;

		if(asg->stor) {
			asg->stor->d.clos_class.data[asg->num] = value;
		} else {
			inst->d.clos_object.data[asg->num] = value;
		}
	}

	/* write defaults/NILs to unset slots */

	for(i=0, asg = type->d.clos_class.allslots; i < type->d.clos_class.allslotsnum; i++, asg++) {
		if(!asg->stor) {
			if(!inst->d.clos_object.data[asg->num]) {
				value = eval(ctx, asg->slot->d.clos_slot.initform);
				if(value->type == ST_ERROR) return value;
				inst->d.clos_object.data[asg->num] = value;
			}
		}
	}

	return inst;
}

/* builtin: is-generalisation */
BUILTIN(sb_is_generalisation)
{
	ARG_SYMBOL(class_a);
	ARG_SYMBOL(class_b);

	s_var *vara = varctx_retr_s(class_hash, class_a->d.symbol.sdbm, class_a->d.symbol.id);
	if(!vara) return create_error(SE_PARM, "Class A undefined");
	s_var *varb = varctx_retr_s(class_hash, class_b->d.symbol.sdbm, class_b->d.symbol.id);
	if(!varb) return create_error(SE_PARM, "Class B undefined");

	return is_generalisation(vara->value, varb->value) ? create_t() : create_nil();
}

/* builtin: call-next-method */
BUILTIN(sb_call_next_method)
{
	s_var * var = varctx_retr_s(ctx, _stutter_dispatch_info_sdbm, _stutter_dispatch_info_id);
	if(!var) return create_error(SE_UNDEF, "call-next-method called outside a dispatch");
	s_obj *dispatch = var->value;
	if(dispatch->type != ST_DISPATCH) return create_error(SE_TYPE, "_DISPATCH-TYPE not a dispatch.");

	s_obj *item;
	int i = 0;
	while(nextarg(ctx, &item, &parm)) {
		if(item->type == ST_ERROR) return item;
		if(i == 0) {
			/* if there are any arguments, create new dispatch info */
			s_obj *nd;
			nd = create_dispatch();
			nd->d.clos_dispatch.method = dispatch->d.clos_dispatch.method;
			nd->d.clos_dispatch.generic = dispatch->d.clos_dispatch.generic;
			dispatch = nd;
		}

		dispatch->d.clos_dispatch.argeval[i++] = item;
	}
	if(i > 0)
		dispatch->d.clos_dispatch.argeval[i] = NULL;

	s_obj *oldm = dispatch->d.clos_dispatch.method;

	s_obj *rtrn = funcexec(ctx, oldm, dispatch, create_nil(), 0, 0);

	dispatch->d.clos_dispatch.method = oldm;

	return rtrn;
}

/* builtin: next-method-p */
BUILTIN(sb_next_method_p)
{
	s_var * var = varctx_retr_s(ctx, _stutter_dispatch_info_sdbm, _stutter_dispatch_info_id);
	if(!var) return create_error(SE_UNDEF, "next-method-p called outside a dispatch");
	s_obj *dispatch = var->value;
	if(dispatch->type != ST_DISPATCH) return create_error(SE_TYPE, "_DISPATCH-TYPE not a dispatch.");

	s_obj * method = dispatch->d.clos_dispatch.method;

	for(;;) {
		method = method->d.clos_method.next;

		if(!method) return create_nil();
		if(is_applicable(method, dispatch->d.clos_dispatch.argeval)) return create_t();
	}
}

/* register builtins in variable context */
void register_builtins(varctx *ctx)
{
	s_mark mark;
	s_obj *o;

	mark = gc_prot_mark();

	/* constants */
	varctx_set(ctx, "NIL", create_nil());
	varctx_set(ctx, "T", create_t());
	varctx_set(ctx, "EOF", create_eof());

	varctx_set(ctx, "NEW-NIL", create_builtin((builtin_handler) create_new_nil));
	varctx_set(ctx, "NEW-T", create_builtin((builtin_handler) create_new_t));

	/* register error constants */
#define REG_ERR(CODE) varctx_set(ctx, #CODE, create_int(CODE))
	REG_ERR(SE_UNDEF);
	REG_ERR(SE_WTF);
	REG_ERR(SE_PARM);
	REG_ERR(SE_WFUNC);
	REG_ERR(SE_TYPE);
	REG_ERR(SE_IPLIST);
	REG_ERR(SE_PNUM);
	REG_ERR(SE_IO);
	REG_ERR(SE_PARSE);
	REG_ERR(SE_NFND);
	REG_ERR(SE_LIMIT);
	REG_ERR(SE_MEM);
	REG_ERR(SE_MATH);
	REG_ERR(SE_RET);
	REG_ERR(SE_COMMA);
	REG_ERR(SE_COMMA_SPLICE);
	REG_ERR(SE_STOPREAD);

#ifdef WITH_SQLITE
	REG_ERR(SE_SQLITE);
#endif

	REG_ERR(SE_LAST);

	register_sforms(ctx);

	/* basic builtin functions */
	varctx_set(ctx, "SET", create_builtin(sb_set));
	varctx_set(ctx, "SETG", create_builtin(sb_setg));

	varctx_set(ctx, "APPLY", create_builtin_with_data(sb_apply, (void *) 0));
	varctx_set(ctx, "FUNCALL", create_builtin_with_data(sb_funcall, (void *) 0));
	varctx_set(ctx, "ELEV", create_builtin_with_data(sb_apply, (void *) 1));
	varctx_set(ctx, "EXPAND", create_builtin_with_data(sb_funcall, (void *) 2));
	varctx_set(ctx, "APPLY-EXPAND", create_builtin_with_data(sb_apply, (void *) 2));
	varctx_set(ctx, "BEGIN", create_builtin(sb_begin));

	varctx_set(ctx, "NOT", create_builtin(sb_not));
	varctx_set(ctx, "EQ", create_builtin(sb_eq));
	varctx_set(ctx, "EQL", create_builtin(sb_eql));
	varctx_set(ctx, "EQUAL", create_builtin(sb_equalop));
	varctx_set(ctx, "WHILE", create_builtin(sb_while));
	varctx_set(ctx, "COND", create_builtin(sb_cond));
	varctx_set(ctx, "RETURN", create_builtin(sb_return));

	varctx_set(ctx, "OR", create_builtin(sb_or));
	varctx_set(ctx, "AND", create_builtin(sb_and));

	varctx_set(ctx, "QUIT", create_builtin(sb_quit));

	varctx_set(ctx, "DEFUSE", create_builtin(sb_defuse));
	varctx_set(ctx, "WEAKPTR-CLEAN", create_builtin(sb_weakptr_clean));

	/* lists */
	varctx_set(ctx, "CONSDUMP", create_builtin(sb_consdump));
	varctx_set(ctx, "CONS", create_builtin(sb_cons));
	varctx_set(ctx, "LIST", create_builtin(sb_list));
	varctx_set(ctx, "APPEND", create_builtin(sb_append));
	varctx_set(ctx, "CAR", create_builtin(sb_car));
	varctx_set(ctx, "CDR", create_builtin(sb_cdr));
	varctx_set(ctx, "LENGTH", create_builtin(sb_length));
	varctx_set(ctx, "SCAR", create_builtin_with_data(sb_scar, (void *) 0));
	varctx_set(ctx, "SCDR", create_builtin_with_data(sb_scar, (void *) 1));
	varctx_set(ctx, "MAPLIST", create_builtin_with_data(sb_maplist, (void *) 0));
	varctx_set(ctx, "FILTERLIST", create_builtin_with_data(sb_maplist, (void *) 1));
	varctx_set(ctx, "MAPCAR", create_builtin_with_data(sb_maplist, (void *) 2));
	varctx_set(ctx, "FILTER", create_builtin_with_data(sb_maplist, (void *) 3));

	/* types */
	varctx_set(ctx, "ATOM", create_builtin(sb_atom));
	varctx_set(ctx, "LISTP", create_builtin(sb_listp));
	varctx_set(ctx, "FUNCTIONP", create_builtin(sb_functionp));
	varctx_set(ctx, "INTEGERP", create_builtin_with_data(sb_typep, (void *) ST_INT));
	varctx_set(ctx, "STRINGP", create_builtin_with_data(sb_typep, (void *) ST_STRING));
	varctx_set(ctx, "FUNCP", create_builtin_with_data(sb_typep, (void *) ST_FUNC));
	varctx_set(ctx, "BUILTINP", create_builtin_with_data(sb_typep, (void *) ST_BUILTIN));
	varctx_set(ctx, "NUMP", create_builtin_with_data(sb_typep, (void *) ST_NUM));
	varctx_set(ctx, "CONSP", create_builtin_with_data(sb_typep, (void *) ST_CONS));
	varctx_set(ctx, "TP", create_builtin_with_data(sb_typep, (void *) ST_T));
	varctx_set(ctx, "SYMBOLP", create_builtin_with_data(sb_typep, (void *) ST_SYMBOL));
	varctx_set(ctx, "ARRAYP", create_builtin_with_data(sb_typep, (void *) ST_ARRAY));
	varctx_set(ctx, "VARCTXP", create_builtin_with_data(sb_typep, (void *) ST_VARCTX));
	varctx_set(ctx, "DICTP", create_builtin_with_data(sb_typep, (void *) ST_DICT));
	varctx_set(ctx, "COOKIEP", create_builtin_with_data(sb_typep, (void *) ST_COOKIE));
	varctx_set(ctx, "CHARP", create_builtin_with_data(sb_typep, (void *) ST_CHAR));
	varctx_set(ctx, "EOFP", create_builtin_with_data(sb_typep, (void *) ST_EOF));
	varctx_set(ctx, "READTABLEP", create_builtin_with_data(sb_typep, (void *) ST_READTABLE));

	varctx_set(ctx, "WEAK-CAR", create_builtin_with_data(sb_weak_car, (void *) 0));
	varctx_set(ctx, "WEAK-CDR", create_builtin_with_data(sb_weak_car, (void *) 1));

	/* arrays */
	varctx_set(ctx, "MAKE-ARRAY", create_builtin_with_data(sb_make_array, (void *) 0));
	varctx_set(ctx, "AREF", create_builtin(sb_aref));
	varctx_set(ctx, "ASET", create_builtin(sb_aset));
	varctx_set(ctx, "WEAK-ASET", create_builtin_with_data(sb_aset, (void *) 1));
	varctx_set(ctx, "ADIM", create_builtin(sb_adim));

	/* dictionary */
	varctx_set(ctx, "MAKE-DICTIONARY", create_builtin(sb_make_dictionary));
	varctx_set(ctx, "DREF", create_builtin(sb_dref));
	varctx_set(ctx, "DPUT", create_builtin(sb_dput));
	varctx_set(ctx, "DDEL", create_builtin(sb_ddel));
	varctx_set(ctx, "DICTIONARY-KEYS", create_builtin(sb_dictionary_keys));

	/* algebra */
	varctx_set(ctx, "+", create_builtin(sb_plus));
	varctx_set(ctx, "-", create_builtin(sb_minus));
	varctx_set(ctx, "*", create_builtin(sb_mul));
	varctx_set(ctx, "/", create_builtin(sb_div));
	varctx_set(ctx, "MOD", create_builtin(sb_mod));

	varctx_set(ctx, "LOGIOR", create_builtin_with_data(sb_logic, (void *) 0));
	varctx_set(ctx, "LOGAND", create_builtin_with_data(sb_logic, (void *) 1));
	varctx_set(ctx, "LOGXOR", create_builtin_with_data(sb_logic, (void *) 2));
	varctx_set(ctx, "LOGNOT", create_builtin(sb_lognot));

	varctx_set(ctx, "->NUM", create_builtin(sb_2num));
	varctx_set(ctx, "->STR", create_builtin(sb_2str));
	varctx_set(ctx, "->SYMBOL", create_builtin(sb_2symbol));
	varctx_set(ctx, "->INT", create_builtin(sb_2int));

	varctx_set(ctx, "RANDOM", create_builtin_with_data(sb_random, (void *) 0));
	varctx_set(ctx, "RAND", create_builtin_with_data(sb_random, (void *) 1));

	/* interpreter control */
	varctx_set(ctx, "TRACE", create_builtin_with_data(sb_trace, (void *) 1));
	varctx_set(ctx, "UNTRACE", create_builtin_with_data(sb_trace, (void *) 0));
	varctx_set(ctx, "GC", create_builtin(sb_gc));
	varctx_set(ctx, "LIMIT", create_builtin(sb_limit));
	varctx_set(ctx, "EVAL", create_builtin(sb_eval));

	varctx_set(ctx, "LOAD", create_builtin_with_data(sb_load, (void *) 0));
	varctx_set(ctx, "LOADHERE", create_builtin_with_data(sb_load, (void *) 1));
	varctx_set(ctx, "REQUIRE", create_builtin_with_data(sb_load, (void *) 2));
	varctx_set(ctx, "REQUIREHERE", create_builtin_with_data(sb_load, (void *) 3));

	varctx_set(ctx, "THROW", create_builtin(sb_throw));
	varctx_set(ctx, "THROW-PARSE", create_builtin_with_data(sb_throw, (void *) 1));
	varctx_set(ctx, "DUMPCTX", create_builtin(sb_dumpctx));

	/* string functions */
	varctx_set(ctx, "SUBSTR", create_builtin(sb_substr));
	varctx_set(ctx, "EXPLODE", create_builtin(sb_explode));
	varctx_set(ctx, "CHAR->CODE", create_builtin(sb_char_code));
	varctx_set(ctx, "CODE->CHAR", create_builtin(sb_code_char));
	varctx_set(ctx, "NTHCHAR", create_builtin(sb_nthchar));

	varctx_set(ctx, "TOUPPER", create_builtin_with_data(sb_toupper, (void *) toupper));
	varctx_set(ctx, "TOLOWER", create_builtin_with_data(sb_toupper, (void *) tolower));

	varctx_set(ctx, "TRIM", create_builtin(sb_trim));

	/* some predicates */
	varctx_set(ctx, "=", create_builtin_with_data(sb_equal, (void *) '='));
	varctx_set(ctx, "<", create_builtin_with_data(sb_equal, (void *) '<'));
	varctx_set(ctx, ">", create_builtin_with_data(sb_equal, (void *) '>'));
	varctx_set(ctx, "<=", create_builtin_with_data(sb_equal, (void *) '{'));
	varctx_set(ctx, ">=", create_builtin_with_data(sb_equal, (void *) '}'));

	varctx_set(ctx, "ISSET", create_builtin_with_data(sb_isset, (void *) 0));
	varctx_set(ctx, "ISLOCAL", create_builtin_with_data(sb_isset, (void *) 1));

	/* I/O */
	varctx_set(ctx, "PRINT", create_builtin_with_data(sb_print, (void *) 0));
	varctx_set(ctx, "PRINTLN", create_builtin_with_data(sb_println, (void *) 0));
	varctx_set(ctx, "PRINC", create_builtin_with_data(sb_print, (void *) 1));
	varctx_set(ctx, "PRINCLN", create_builtin_with_data(sb_println, (void *) 1));
	varctx_set(ctx, "READ-LINE", create_builtin(sb_read_line));
	varctx_set(ctx, "READ-CHAR", create_builtin(sb_read_char));
	varctx_set(ctx, "UNREAD-CHAR", create_builtin(sb_unread_char));
	varctx_set(ctx, "READ", create_builtin(sb_read));

	varctx_set(ctx, "MAKE-STREAM", create_builtin(sb_make_stream));
	varctx_set(ctx, "OPEN-FILE", create_builtin(sb_open_file));
	varctx_set(ctx, "CLOSE-FILE", create_builtin(sb_close_file));

	varctx_set(ctx, "STRING-BUCKET", create_builtin(sb_string_bucket));
	varctx_set(ctx, "EMPTY-BUCKET", create_builtin(sb_empty_bucket));

#ifdef BUILTIN_SYSTEM
	varctx_set(ctx, "SYSTEM", create_builtin(sb_system));
#endif

#ifdef BUILTIN_CONNECTTCPSTREAM
	varctx_set(ctx, "CONNECT-TCPSTREAM", create_builtin(sb_connect_tcpstream));
	varctx_set(ctx, "CLOSE-TCPSTREAM", create_builtin(sb_close_tcpstream));
#endif

	/* date/time */

	varctx_set(ctx, "GET-TIME", create_builtin(sb_get_time));
	varctx_set(ctx, "GET-FTIME", create_builtin(sb_get_ftime));
	varctx_set(ctx, "GET-DECODED-TIME", create_builtin(sb_get_decoded_time));
    varctx_set(ctx, "SLEEP", create_builtin(sb_sleep));

	/* CLOS */

	varctx_set(ctx, "DEFCLASS-INT", create_builtin(sb_defclass));
	varctx_set(ctx, "CLASS-PRECEDENCE", create_builtin(sb_class_precedence));
	varctx_set(ctx, "CLASS-ANCESTRY", create_builtin(sb_class_ancestry));
	varctx_set(ctx, "CLASS-SLOTINFO", create_builtin(sb_class_slotinfo));
	varctx_set(ctx, "SLOT-VALUE", create_builtin(sb_slot_value));
	varctx_set(ctx, "SLOT-VALUE-SET", create_builtin(sb_slot_value_set));
	varctx_set(ctx, "SLOT-VALUE-SET-WEAK", create_builtin(sb_slot_value_set_weak));
	varctx_set(ctx, "MAKE-INSTANCE", create_builtin(sb_make_instance));
	varctx_set(ctx, "IS-GENERALISATION", create_builtin(sb_is_generalisation));
	varctx_set(ctx, "CALL-NEXT-METHOD", create_builtin(sb_call_next_method));
	varctx_set(ctx, "NEXT-METHOD-P", create_builtin(sb_next_method_p));

	/* math */

	register_math_builtins(ctx);

	/* streams */

	o = create_stream(STREAM_FP);
	o->d.stream.s.fp = stdin;
	varctx_set(ctx, "*INPUT-STREAM*", o);
	varctx_set(ctx, "*STDIN*", o);

	o = create_stream(STREAM_FP);
	o->d.stream.s.fp = stdout;
	varctx_set(ctx, "*OUTPUT-STREAM*", o);
	varctx_set(ctx, "*STDOUT*", o);

	o = create_stream(STREAM_FP);
	o->d.stream.s.fp = stderr;
	varctx_set(ctx, "*STDERR*", o);

	/* initialize modules */

#ifdef WITH_SQLITE
	modinit_sqlite(ctx);
#endif

#ifdef WITH_XML
	modinit_xml(ctx);
#endif

	gc_prot_free(mark, NULL);
}

/* load SYNTAX.stt */
s_obj * register_syntax(varctx *ctx)
{
	return eval(ctx, create_cons(create_symbol("REQUIRE"), create_cons(create_string("SYNTAX"), create_nil())));
}

