#ifndef _BUILTINS_H
#define _BUILTINS_H

#include "types.h"
#include "memory.h"

#ifndef PATHSEP
#define PATHSEP "/"
#endif

#ifdef WIN32
#define SEPCHAR ';'
#define SEPSTR ";"
#else
#define SEPCHAR ':'
#define SEPSTR ":"
#endif

#ifdef __EVC4__
struct timeval {
	time_t tv_sec;
	unsigned int tv_usec;
};

#endif


/* handy macros for builtin definition
 * ARG_X(name)
 *  defines s_obj *name
 *  retrieves next argument from argument list
 *  checks for error and type
 * GRAB_X(name)
 *  as ARG_X, but doesn't define variable (reuses existing)
 */

#define BUILTIN(name) s_obj * name (varctx *ctx, s_obj *parm, void *data, const int tail)

#define GRAB_OBJ(name) \
	if(!nextarg(ctx, & name, &parm)) \
		return create_error(SE_PNUM, "Insufficient arguments, missing " #name); \
	if(name->type == ST_ERROR) return name;

#define ARG_OBJ(name) \
	s_obj * name; \
	GRAB_OBJ(name);

#define ARG_OBJ_T(name, dtype) \
	ARG_OBJ(name); \
	if(name->type != dtype) return create_error(SE_TYPE, "Type mismatch: " #name " expected to be " #dtype);

#define GRAB_OBJ_T(name, dtype) \
	GRAB_OBJ(name); \
	if(name->type != dtype) return create_error(SE_TYPE, "Type mismatch: " #name " expected to be " #dtype);

#define ARG_FUNCTION(name) \
	ARG_OBJ(name); \
	if(name->type != ST_BUILTIN && name->type != ST_FUNC) \
		return create_error(SE_TYPE, "Type mismatch: " #name " expected to be a function");

#define ARG_LIST(name) \
	ARG_OBJ(name); \
	if(name->type != ST_NIL && name->type != ST_CONS) \
		return create_error(SE_TYPE, "Type mismatch: " #name " expected to be a list");

#define GRAB_FUNCTION(name) \
	GRAB_OBJ(name); \
	if(name->type != ST_BUILTIN && name->type != ST_FUNC) \
		return create_error(SE_TYPE, "Type mismatch: " #name " expected to be a function");

#define GRAB_LIST(name) \
	GRAB_OBJ(name); \
	if(name->type != ST_NIL && name->type != ST_CONS) \
		return create_error(SE_TYPE, "Type mismatch: " #name " expected to be a list");

#define ARG_INT(name) ARG_OBJ_T(name, ST_INT);
#define ARG_NUM(name) ARG_OBJ_T(name, ST_NUM);
#define ARG_CHAR(name) ARG_OBJ_T(name, ST_CHAR);
#define ARG_STRING(name) ARG_OBJ_T(name, ST_STRING);
#define ARG_ARRAY(name) ARG_OBJ_T(name, ST_ARRAY);
#define ARG_DICT(name) ARG_OBJ_T(name, ST_DICT);
#define ARG_STREAM(name) ARG_OBJ_T(name, ST_STREAM);
#define ARG_CLASS(name) ARG_OBJ_T(name, ST_CLASS);
#define ARG_SLOT(name) ARG_OBJ_T(name, ST_SLOT);
#define ARG_OBJECT(name) ARG_OBJ_T(name, ST_OBJECT);
#define ARG_GENERIC(name) ARG_OBJ_T(name, ST_GENERIC);
#define ARG_METHOD(name) ARG_OBJ_T(name, ST_METHOD);
#define ARG_SYMBOL(name) ARG_OBJ_T(name, ST_SYMBOL);
#define ARG_CONS(name) ARG_OBJ_T(name, ST_CONS);

#define GRAB_INT(name) GRAB_OBJ_T(name, ST_INT);
#define GRAB_NUM(name) GRAB_OBJ_T(name, ST_NUM);
#define GRAB_CHAR(name) GRAB_OBJ_T(name, ST_CHAR);
#define GRAB_STRING(name) GRAB_OBJ_T(name, ST_STRING);
#define GRAB_ARRAY(name) GRAB_OBJ_T(name, ST_ARRAY);
#define GRAB_DICT(name) GRAB_OBJ_T(name, ST_DICT);
#define GRAB_STREAM(name) GRAB_OBJ_T(name, ST_STREAM);
#define GRAB_CLASS(name) GRAB_OBJ_T(name, ST_CLASS);
#define GRAB_SLOT(name) GRAB_OBJ_T(name, ST_SLOT);
#define GRAB_OBJECT(name) GRAB_OBJ_T(name, ST_OBJECT);
#define GRAB_GENERIC(name) GRAB_OBJ_T(name, ST_GENERIC);
#define GRAB_METHOD(name) GRAB_OBJ_T(name, ST_METHOD);
#define GRAB_SYMBOL(name) GRAB_OBJ_T(name, ST_SYMBOL);
#define GRAB_CONS(name) GRAB_OBJ_T(name, ST_CONS);




/* print cons-dump */
s_obj * consdump(varctx *ctx, s_obj *obj);

/* get value of next parameter (returns 1 on success, 0 if there are no more) */
inline int nextparm(s_obj **obj, s_obj **parm);

/* get and evaluate next parameter (returns 1 on success, 0 if there are no more) */
inline int nextarg(varctx *ctx, s_obj **obj, s_obj **parm);

/* get and evaluate next parameter (returns 1 on success, 0 if there are no more).
 * if it's the last parameter and tail is true, calls eval
 * with tail flag set to 1! */
inline int nextargtail(varctx *ctx, s_obj **obj, s_obj **parm, const int tail);

/* get and evaluate next parameter, ensure it's from a type
 * returns 1 on success */
int next_t(varctx *ctx, s_obj **obj, s_obj **parm, const int type);

/* register builtins in variable context */
void register_builtins(varctx *ctx);

/* register builtins in variable context */
void register_sforms(varctx *ctx);

/* load SYNTAX.stt */
s_obj * register_syntax(varctx *ctx);

/* register math builtins in variable context */
void register_math_builtins(varctx *ctx);

/* builtin: quote */
BUILTIN(sb_quote);

/* quote an object for passing as parameter.
 * returns a list: (quote OBJ)
 */
s_obj *quoteobj(s_obj *obj);

#endif
