#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include "eval.h"
#include "types.h"
#include "memory.h"
#include "builtins.h"
#include "read.h"

char *stutter_version = VERSION;

int _stutter_trace = 0;
int _stutter_depth = 0;

char *_stutter_file = "stdin";
char *_stutter_lastcall_file = NULL;
int _stutter_lastcall_line = 0;
int _stutter_parsed_line = 0;

int recursion_limit = 1000;

unsigned long _stutter__rest_sdbm = -1;
unsigned long _stutter__rest_id = -1;
unsigned long _stutter_dispatch_info_sdbm = -1;
unsigned long _stutter_dispatch_info_id = -1;
unsigned long _stutter_lambda_sdbm = -1;
unsigned long _stutter_lambda_id = -1;


char *errornames[] = {
	"SE_UNDEF",
	"SE_WTF",
	"SE_PARM",
	"SE_WFUNC",
	"SE_TYPE",
	"SE_IPLIST",
	"SE_PNUM",
	"SE_IO",
	"SE_PARSE",
	"SE_NFND",
	"SE_LIMIT",
	"SE_MEM",
	"SE_MATH",
	"SE_RET",
	"SE_COMMA",
	"SE_COMMA_SPLICE",
	"SE_STOPREAD",
	"SE_COMPILE",

#ifdef WITH_SQLITE
	"SE_SQLITE",
#endif

	"SE_LAST"
};

/* used to bind variables after evaluation */
typedef struct {
	s_obj *def;
	s_obj *var;
	s_obj *value;
} varset;

/* user function executor 
 * this is capable of executing multiple functions in place
 * (tail recursion optimisation)
 *
 * ctx:  hosting varctx
 * obj:  code of the call - the list interpreted as a function invocation.
 * func: CAR of the list evaluated as a function
 * args: the CDR (list of arguments)
 * tail: whether the call was made in a tail position
 *
 * inhibit_macro_expansion: if true, return values of macros aren't evaluated again.
 */
s_obj * funcexec(varctx *ctx, s_obj *obj, s_obj *func, s_obj *args, const int tail, int inhibit_macro_expansion)
{
	varctx *fctx = NULL;	/* ctx for function code evaluation */
	varctx *ectx = ctx;		/* ctx for function argument evaluation */

	varctx *octx = ctx, *cctx, *ictx;
	varset bindings[FUNC_MAXPARMS];

	int nbind, k;

	s_mark mark;

	s_obj *parm, *vars;
	s_obj *i, *j, *first, *cur;
	s_obj *o;
	s_obj *ofunc = func;
	s_obj *oobj = obj;

	s_obj *dispatch = NULL;
	s_obj **argeval = NULL;

	s_obj *rtrn = NULL; /* ultimate return value for methods */

	int ii, tailcall = 0;

	int trace = 0, closure = 0, noafter = 0;

	mark = gc_prot_mark();

handlecall:

	/* repeats in case of a tail call - also for next methods,
	 * before/after method execution */

	switch(func->type) {

	case ST_FUNC:
		/* just a function call. */
		trace = func->d.func.trace;
		closure = func->d.func.closure ? 1 : 0;
		break;

	case ST_GENERIC:
		/* generic function */

		dispatch = create_dispatch();
		mark = gc_prot_mark();
		argeval = dispatch->d.clos_dispatch.argeval;

		dispatch->d.clos_dispatch.generic = func;

		/* evaluate all arguments */
		parm = args; ii = 0;
		while(nextparm(&i, &parm)) {
			o = eval(ctx, i);
			if(o->type == ST_ERROR) {
				if(fctx && !fctx->ref) varctx_free(fctx);
				gc_prot_free(mark, o);
				return o;
			}

			gc_protlist_add(o);

			argeval[ii++] = o;

			if(ii == FUNC_MAXPARMS) {
				if(fctx && !fctx->ref) varctx_free(fctx);
				gc_prot_free(mark, NULL);
				return create_error(SE_PNUM, "Maximum number of arguments for generic exceeded");
			}
		}

		argeval[ii] = NULL;

		/* try dispatch to a method */

		dispatch->d.clos_dispatch.method = method_dispatch(dispatch->d.clos_dispatch.generic, dispatch->d.clos_dispatch.argeval, 0);
		if(!dispatch->d.clos_dispatch.method) {
			if(fctx && !fctx->ref) varctx_free(fctx);
			gc_prot_free(mark, NULL);
			return create_error(SE_PARM, "Method dispatch failed");
		}

		if(dispatch->d.clos_dispatch.method->d.clos_method.place == CL_METHOD_PRIMARY) {
			/* try to run before methods before primary method. */

			if(dispatch->d.clos_dispatch.generic->d.clos_generic.m_before->type == ST_CONS) {
				func = dispatch->d.clos_dispatch.generic->d.clos_generic.m_before->d.cons.car;
				goto handlecall;
			}
		}

		func = dispatch->d.clos_dispatch.method;
		goto handlecall;

		break;

	case ST_METHOD:
		/* don't check for the function being a macro later on..
		 * it's not even a function */
		inhibit_macro_expansion = 1;

		switch(func->d.clos_method.place) {
		case CL_METHOD_BEFORE:
			/* executing a :before method. */

			if(!is_applicable(func, argeval))  {
				/* not applicable, look at next method */
				func = func->d.clos_method.next;

				if(func == NULL) {
					/* out of :before methods, jump to primary method. */
					func = dispatch->d.clos_dispatch.method;
				}

				goto handlecall;
			}

			break;

		case CL_METHOD_AFTER:
			/* executing an :after method. */

			if(!is_applicable(func, argeval)) {
				/* not applicable. look at next method */
				func = func->d.clos_method.next;

				if(func == NULL) {
					/* out of :after methods. return */

					return rtrn;
				}

				goto handlecall;
			}

			break;

		case CL_METHOD_AROUND:
			break;

		case CL_METHOD_PRIMARY:
			break;
		}

		break;
	
	case ST_DISPATCH:

		/* call-next-method */

		dispatch = func;
		argeval = dispatch->d.clos_dispatch.argeval;

		/* we can set dispatch.method here (it's saved and restored
		 * in the call-next-method builtin...) */

		/* the old method gets passed in obj. sorry for that */

		func = dispatch->d.clos_dispatch.method = obj->d.clos_method.next;
		if(!func) {
			if(fctx && !fctx->ref) varctx_free(fctx);
			gc_prot_free(mark, NULL);
			return create_error(SE_WFUNC, "call-next-method called and there's no method available");
		}

		noafter = 1;

		if(obj->d.clos_method.place == CL_METHOD_AROUND && func->d.clos_method.place == CL_METHOD_PRIMARY) {
			/* get first applicable primary method */

			noafter = 0;

			func = dispatch->d.clos_dispatch.method = method_dispatch(dispatch->d.clos_dispatch.generic, argeval, 1);
			if(!func) {
				if(fctx && !fctx->ref) varctx_free(fctx);
				gc_prot_free(mark, NULL);
				return create_error(SE_PARM, "call-next-method called from :around with no primary method available");
			}

			/* try to run before methods before primary method. */

			if(dispatch->d.clos_dispatch.generic->d.clos_generic.m_before->type == ST_CONS) {
				func = dispatch->d.clos_dispatch.generic->d.clos_generic.m_before->d.cons.car;
				goto handlecall;
			}
		}

		goto handlecall;

	default:
		assert(0); /* doesn't happen. */
	}


	/* in case of tracing enabled, draw trace output */
	if(trace) {
		for(ii=0; ii<=_stutter_depth; ii++) {
			o = stream_stdout_write(ctx, ' ');
			if(o->type == ST_ERROR) return o;
		}
		o = stream_printf(ctx, "%c> ", tailcall ? '*' : '=');
		if(o->type == ST_ERROR) return o;
		o = printobj(ctx, obj ? obj->d.cons.car : NULL);
		if(o->type == ST_ERROR) return o;
		o = stream_printf(ctx, " (");
		if(o->type == ST_ERROR) return o;
	}

	if(tailcall) {
		/* ectx is the context in which function arguments are evaluated,
		 * while in fctx the function code gets executed.
		 * in case of a tail call, the current fctx becomes ectx
		 */
		ectx = fctx;

		if(closure) {
			/* if it's not done already, make the current fctx into
			 * a first class object - this way the GC can reclaim it
			 * when necessary. */
			if(!fctx->ref) fctx->ref=create_obj_varctx(fctx);
		}
	} else {
		if(!closure) {
			fctx = varctx_create(ctx, FUNC_LOCALCTX_SIZE);
		}

		if(trace)
			_stutter_depth ++;
	}

	if(closure) {
		fctx = varctx_create(NULL, FUNC_LOCALCTX_SIZE);
		cctx = NULL;

		/* link all varctxes described in the closure objects as parents
		 * to the new one created for this function invocation. */

		for(ictx = func->d.func.closure->d.varctx.ctx; ictx->parent; ictx = ictx->parent) {
			if(cctx) {
				cctx->parent = varctx_create(ctx, 0);
				cctx = cctx->parent;
			} else {
				cctx = varctx_create(ctx, 0);
				fctx->parent = cctx;
			}

			cctx->hash = ictx->hash;
			cctx->ns = ictx->ns;
			cctx->nsm = ictx->nsm;
			cctx->defer = 1;
			cctx->ref = create_obj_varctx(cctx);
		}
	}

	nbind = 0;

	if(func->type == ST_METHOD) {
		/* bind method arguments */

		for(ii=0; ii<func->d.clos_method.argn; ii++) {
			if(!argeval[ii]) {
				if(!fctx->ref) varctx_free(fctx);
				i = create_error(SE_PARM, "Out of specified method arguments (bad call-next-method?)");
				gc_prot_free(mark, i);
				if(trace) _stutter_depth --;
				return i;
			}

			varctx_set_s(fctx, func->d.clos_method.args[ii].name->d.symbol.sdbm, func->d.clos_method.args[ii].name->d.symbol.id, argeval[ii]);
		}

		/* fill in bindings[] from method miscargs */
		vars = func->d.clos_method.miscargs;

		/* set _DISPATCH-INFO */
		varctx_set_s(fctx, _stutter_dispatch_info_sdbm, _stutter_dispatch_info_id, dispatch);
	} else {
		/* bindings[] source is the standard function argument list */
		vars = func->d.func.parm;
	}

	/* map variables */
	parm = args;
	first = cur = NULL;

	/* fill in bindings array with function argument list */
	ii = 0;
	while(nextparm(&i, &vars)) {
		if(ii == FUNC_MAXPARMS) {
			if(!fctx->ref) varctx_free(fctx);
			i = create_error(SE_PARM, "Function has too many arguments");
			gc_prot_free(mark, i);
			assert(i != NULL);
			if(trace) _stutter_depth --;
			return i;
		}

		if(i->type == ST_CONS) {
			if(i->d.cons.cdr->type == ST_CONS)
				bindings[ii].def = i->d.cons.cdr->d.cons.car;
			else
				bindings[ii].def = i->d.cons.cdr;
			bindings[ii].var = i->d.cons.car;
		} else {
			bindings[ii].def = NULL;
			bindings[ii].var = i;
		}

		bindings[ii].value = NULL;

		ii++;
	}
	nbind = ii;

	/* traverse call arguments */

	if(func->type == ST_METHOD) {
		/* copy values from argeval to bindings */

		int l = 0;
		for(ii = func->d.clos_method.argn; argeval[ii]; ii++) {
			if(argeval[ii]->type == ST_SYMBOL && argeval[ii]->d.symbol.colon) {
				/* key-value pair */

				i = argeval[ii];

				/* get value */
				ii ++;
				if(!argeval[ii]) {
					if(!fctx->ref) varctx_free(fctx);
					i = create_error(SE_PARM, "Can't have a key/value pair without a value");
					gc_prot_free(mark, i);
					if(trace) _stutter_depth --;
					return i;
				}

				j = argeval[ii];

				/* find variable slot for key */
				for(k=0; k<nbind; k++)
					if(bindings[k].var->d.symbol.id == i->d.symbol.id)
						break;

				if(k == nbind) {
					/* invalid key */
					if(!fctx->ref) varctx_free(fctx);
					i = create_error(SE_PARM, "Invalid parameter key");
					gc_prot_free(mark, i);
					if(trace) _stutter_depth --;
					return i;
				}

				bindings[k].value = j;
			} else {
				/* by-position argument */

				i = argeval[ii];

				if(ii == nbind) {
					/* argument list exhausted, append this to the _rest list */
					if(!cur) {
						first = cur = create_cons(i, create_nil());
					} else {
						cur->d.cons.cdr = create_cons(i, create_nil());
						cur = cur->d.cons.cdr;
					}
				} else {
					bindings[l].value = i;
					l++;
				}
			}
		}

	} else {

		/* evaluate function argument list, as usual */

		ii = 0;
		while(nextparm(&i, &parm)) {
			if(i->type == ST_SYMBOL && i->d.symbol.colon) {
				/* key-value pair */

				/* already got key, get value */
				if(!nextparm(&j, &parm)) {
					if(!fctx->ref) varctx_free(fctx);
					i = create_error(SE_PARM, "Can't have a key/value pair without a value");
					gc_prot_free(mark, i);
					if(trace) _stutter_depth --;
					return i;
				}

				/* evaluate argument */
				if(!func->d.func.macro) { /* macro functions don't get their parameters evaluated */
					j = eval(ectx, j);
					if(j->type == ST_ERROR) {
						if(!fctx->ref) varctx_free(fctx);
						gc_prot_free(mark, j);
						if(trace) _stutter_depth --;
						return j;
					}
					gc_protlist_add(j);
				}

				/* find variable slot for key */
				for(k=0; k<nbind; k++)
					if(bindings[k].var->d.symbol.id == i->d.symbol.id)
						break;

				if(k == nbind) {
					/* invalid key */

					if(!fctx->ref) varctx_free(fctx);
					i = create_error(SE_PARM, "Invalid parameter key");
					gc_prot_free(mark, i);
					assert(i != NULL);
					if(trace) _stutter_depth --;
					return i;
				}

				if(trace) {
					/* if tracing is enabled, draw trace output. */
					o = stream_stdout_write(ctx, ' ');
					if(o->type == ST_ERROR) return o;
					o = printobj(ctx, i);
					if(o->type == ST_ERROR) return o;
					o = stream_stdout_write(ctx, ' ');
					if(o->type == ST_ERROR) return o;
					o = printobj(ctx, j);
				}

				bindings[k].value = j;

			} else {
				/* by-position argument */

				/* evaluate argument */
				if(!func->d.func.macro) { /* macro functions don't get their parameters evaluated */
					i = eval(ectx, i);
					if(i->type == ST_ERROR) {
						if(!fctx->ref) varctx_free(fctx);
						gc_prot_free(mark, i);
						assert(i != NULL);
						if(trace) _stutter_depth --;
						return i;
					}
					gc_protlist_add(i);
				}

				if(trace) {
					/* if tracing is enabled, draw trace output. */
					o = stream_stdout_write(ctx, ' ');
					if(o->type == ST_ERROR) return o;
					o = printobj(ctx, i);
					if(o->type == ST_ERROR) return o;
				}

				if(ii == nbind) {
					/* argument list exhausted, append this to the _rest list */
					if(!cur) {
						first = cur = create_cons(i, create_nil());
					} else {
						cur->d.cons.cdr = create_cons(i, create_nil());
						cur = cur->d.cons.cdr;
					}
				} else {
					bindings[ii].value = i;
					ii++;
				}
			}
		}
	}

	if(tailcall && fctx->ref) {
		/* fctx is referenced by a closure, we can't reuse it */
		fctx = varctx_create(ctx, FUNC_LOCALCTX_SIZE);
	}

	/* set existent bindings in the varctx */
	for(k=0; k<nbind; k++)
		if(bindings[k].value)
			varctx_set_s(fctx, bindings[k].var->d.symbol.sdbm, bindings[k].var->d.symbol.id, bindings[k].value);

	/* check for unset slots, evaluate default values */
	for(k=0; k<nbind; k++) {
		if(!bindings[k].value) {
			if(!bindings[k].def) {
				/* no default value for parameter */
				if(!fctx->ref) varctx_free(fctx);
				i = create_error(SE_PNUM, "Insufficient arguments when calling user function");
				gc_prot_free(mark, i);
				assert(i != NULL);
				if(trace) _stutter_depth --;
				return i;
			} else {
				/* evaluate and bind default parameter */

				i = eval(fctx, bindings[k].def);
				if(i->type == ST_ERROR) {
					if(!fctx->ref) varctx_free(fctx);
					gc_prot_free(mark, i);
					assert(i != NULL);
					if(trace) _stutter_depth --;
					return i;
				}

				varctx_set_s(fctx, bindings[k].var->d.symbol.sdbm, bindings[k].var->d.symbol.id, i);
			}
		}
	}

	if(trace) {
		o = stream_printf(ctx, " ) at %s:%d\n", obj ? obj->file : NULL, obj ? obj->line : 0);
		if(o->type == ST_ERROR) return o;
	}

	/* bind _rest in the varctx */
	varctx_set_s(fctx, _stutter__rest_sdbm, _stutter__rest_id, first ? first : create_nil());

	/* execute function body */

	if(func->type == ST_METHOD) {
		/* executing a method */


		parm = func->d.clos_method.code;
		j = NULL;

		while(nextparm(&i, &parm)) {
			j = eval(fctx, i);

			if(j->type == ST_ERROR) {
				/* append location to trace list */
				if(!j->d.error.trace) {
					j->d.error.trace = create_cons(o = create_error(j->d.error.code, NULL), create_nil());
				} else {
					for(o = j->d.error.trace; o->d.cons.cdr->type == ST_CONS; o = o->d.cons.cdr);
					o->d.cons.cdr = create_cons(create_error(j->d.error.code, NULL), create_nil());
					o = o->d.cons.cdr->d.cons.car;
				}
				o->d.error.obj = func;
				o->d.error.file = i->file;
				o->d.error.line = i->line;

				if(!fctx->ref) varctx_free(fctx);
				if(trace) _stutter_depth --;
				gc_prot_free(mark, j);
				return j;
			}
		}
		if(!j) j = create_nil(); /* empty functions exist.. */

	} else {

		/* executing an ordinary function */

		parm = func->d.func.exp;
		j = NULL;
		while(nextparm(&i, &parm)) {
			j = eval_(fctx, i, !parm || parm->type != ST_CONS);

			if(j->type == ST_DEFER) break; /* we got a tailcall :) */

			if(j->type == ST_ERROR) {
				/* append location to trace list */
				if(!j->d.error.trace) {
					j->d.error.trace = create_cons(o = create_error(j->d.error.code, NULL), create_nil());
				} else {
					for(o = j->d.error.trace; o->d.cons.cdr->type == ST_CONS; o = o->d.cons.cdr);
					o->d.cons.cdr = create_cons(create_error(j->d.error.code, NULL), create_nil());
					o = o->d.cons.cdr->d.cons.car;
				}
				o->d.error.obj = func;
				o->d.error.file = i->file;
				o->d.error.line = i->line;

				if(!fctx->ref) varctx_free(fctx);
				if(trace) _stutter_depth --;
				gc_prot_free(mark, j);
				return j;
			}
		}
		if(!j) j = create_nil(); /* empty functions exist.. */

		/* handle a tailcall */
		if(j->type == ST_DEFER) {
			func = j->d.defer.func;
			args = j->d.defer.args;
			obj = j->d.defer.obj;
			tailcall = 1;

			gc_prot_free(mark, NULL);

			goto handlecall;
		}
	}

	/* finish function execution. */

	if(trace) {
		/* if tracing is enabled, draw trace output */
		for(ii=0; ii<_stutter_depth; ii++) {
			o = stream_stdout_write(ctx, ' ');
			if(o->type == ST_ERROR) return o;
		}
		o = stream_printf(ctx, "<= ");
		if(o->type == ST_ERROR) return o;
		o = printobj(ctx, j);
		if(o->type == ST_ERROR) return o;
		o = stream_stdout_write(ctx, '\n');
		if(o->type == ST_ERROR) return o;
		_stutter_depth --;
	}

	if(!fctx->ref) varctx_free(fctx);

	if(!inhibit_macro_expansion) {
		if(ofunc->d.func.macro == 2) {
			/* this is an r-macro, we get to evaluate the results AGAIN! yay! */
			j = eval_(octx, j, tail);
		} else if(ofunc->d.func.macro & 1) {
			/* this is a macro. we replace the original s-expression with the
			 * results. */
#ifdef REPORT_EXPANSION
			printobj(octx, ofunc);
			printf(" <- macro %s:%d\n", oobj->file, oobj->line);
#endif
			freedata(oobj);
			deepcopy(oobj, j);
			j = eval_(octx, oobj, tail);
		}
	}

	gc_prot_free(mark, j);
	if(rtrn) gc_protlist_add(rtrn);

	if(func->type == ST_METHOD) {
		/* progress with method execution */

		switch(func->d.clos_method.place) {
		case CL_METHOD_BEFORE:
			func = func->d.clos_method.next;
			if(!func) func = dispatch->d.clos_dispatch.method;
			goto handlecall;

		case CL_METHOD_PRIMARY:
			rtrn = j;

			if(!noafter && dispatch->d.clos_dispatch.generic->d.clos_generic.m_after->type == ST_CONS) {
				func = dispatch->d.clos_dispatch.generic->d.clos_generic.m_after->d.cons.car;
				goto handlecall;
			}
			break;

		case CL_METHOD_AFTER:
			if((func = func->d.clos_method.next)) {
				goto handlecall;
			}

			j = rtrn;

			break;

		default: ;
		}
	}

	return j;
}

/* evaluate expression */
s_obj *eval_(varctx *ctx, s_obj *obj, const int tail)
{
	s_var *var;
	s_obj *func;
	s_obj *i, *j;
	s_mark mark;

	assert(obj != NULL);

	if(_stutter_gc_allocd > GC_ALLOC_LIMIT) {
		gc_sweep(ctx);
		_stutter_gc_allocd = 0;
	}

	switch(obj->type) {
	case ST_CONS:
		/* evaluate s-expression */

		_stutter_lastcall_file = obj->file;
		_stutter_lastcall_line = obj->line;


		func = eval(ctx, obj->d.cons.car);
		if(func->type == ST_ERROR) return func;

		switch(func->type) {
			case ST_BUILTIN:
				/* builtin function call */

				/* optimise by replacing symbol in s-exp directly with builtin */
				if(obj->d.cons.car->type == ST_SYMBOL &&
					obj->d.cons.car->d.symbol.id == func->d.builtin.symbolid &&
					!func->d.builtin.nosubst) {

					obj->d.cons.car = func;
				}

				mark = gc_prot_mark();

				i = func->d.builtin.handler(ctx, obj->d.cons.cdr, func->d.builtin.data, tail);

				gc_prot_free(mark, i);
				assert(i != NULL);
				return i;

			case ST_FUNC:
				/* user function call */

				if(tail && !func->d.func.macro) {
					/* this is a tailcall!
					 * defer execution to parent funcexec */
					return create_defer(func, obj->d.cons.cdr, obj);
				}

				recursion_limit --;
				if(!recursion_limit) {
					recursion_limit = 0;
					return create_error_with_obj(SE_LIMIT, "Recursion limit exceeded", obj);
				}

				j = funcexec(ctx, obj, func, obj->d.cons.cdr, tail, 0);
				assert(j != NULL);

				recursion_limit ++;
				return j;

			case ST_GENERIC:
				/* generic function invocation */

				if(tail) {
					/* tail call, defer execution to parent funcexec */
					return create_defer(func, obj->d.cons.cdr, obj);
				}

				recursion_limit --;
				if(!recursion_limit) {
					recursion_limit = 0;
					return create_error_with_obj(SE_LIMIT, "Recursion limit exceeded", obj);
				}

				j = funcexec(ctx, obj, func, obj->d.cons.cdr, tail, 0);
				assert(j != NULL);

				recursion_limit ++;
				return j;

			default:
				return create_error_with_obj(SE_WFUNC, "Non-function in S-expression", obj);
		}
		break;

	case ST_SYMBOL:
		if(obj->d.symbol.colon)
			return obj; /* keyword symbol */

		/* look up symbol */
		var = varctx_get_s(ctx, obj->d.symbol.sdbm, obj->d.symbol.id);
		if(!var) {
			_stutter_lastcall_file = obj->file;
			_stutter_lastcall_line = obj->line;

			char buf[256];
			sprintf(buf, "Variable '%s' has no value", obj->d.symbol.name);
			obj = create_error(SE_UNDEF, buf);
			return obj;
		}

		assert(var->value != NULL);

		if(var->value->type == ST_FUNC && (var->value->d.func.macro & 4)) {
			_stutter_lastcall_file = obj->file;
			_stutter_lastcall_line = obj->line;

			j = funcexec(ctx, create_cons(var->value, create_nil()), var->value, create_nil(), 0, 1);
			if(j->type == ST_ERROR) return j;
			if(var->value->d.func.macro == 4) {
				freedata(obj);
				deepcopy(obj, j);
				j = obj;
			}

			j = eval_(ctx, j, 0);
			return j;
		}

		return var->value;
	default:
		/* directly return atom */
		assert(obj != NULL);
		return obj;
	}

	return create_error(SE_WTF, "Cannot evaluate object");
}


static s_obj *compile_intr(varctx *ctx, s_obj *expr, const int callee, compx * compx);

/* compile an argument list */
static s_obj *compile_arglist(varctx *ctx, s_obj *expr, compx *compx)
{
	s_obj *car, *cdr;
	switch(expr->type) {
	case ST_NIL:
		return expr;
	case ST_CONS:
		car = compile_intr(ctx, expr->d.cons.car, 0, compx);
		if(car->type == ST_ERROR) return car;
		cdr = compile_arglist(ctx, expr->d.cons.cdr, compx);
		if(cdr->type == ST_ERROR) return cdr;
		return create_cons(car, cdr);
	default:
		return create_error_with_obj(SE_COMPILE, "S-expression is a dotted list.", expr);
	}
}

/* compile a lambda */
static s_obj *compile_lambda(varctx *ctx, s_obj *expr, const int callee, compx *compx)
{
	/* TODO */
	return NULL;
}

/* compile an expression (internal)
 * callee: compiled thing is first of s-expr
 * tail: in tail position.
 */
static s_obj *compile_intr(varctx *ctx, s_obj *expr, const int callee, compx *compx)
{
	s_obj *car, *cdr;

	switch(expr->type) {
		case ST_CONS:
			/* S-expression */

			car = compile_intr(ctx, expr->d.cons.car, 1, compx);
			if(car->type == ST_ERROR) return car;

			if(car->type == ST_SYMBOL) {
				/* TODO: check car for specials */
				if(car->d.symbol.id == _stutter_lambda_id) {
					return compile_lambda(ctx, expr, callee, compx);
				}

				s_var *var = varctx_get_s(ctx, car->d.symbol.sdbm, car->d.symbol.id);
				if(var) {
					/* car of s-expr is a bound symbol. */
					s_obj *value = var->value;

					switch(value->type) {
					case ST_FUNC:
						if(value->d.func.macro & 1) {
							/* evaluate and replace (after symbol lookup part) */
							car = value;
						}
						break;
					case ST_BUILTIN:
						/* optimise by replacing symbol in s-exp directly with builtin */
						if(car->d.symbol.id == value->d.builtin.symbolid &&
								!value->d.builtin.nosubst) {
							car = value;
						}
						break;
					default:
						/* dunno. */
						break;
					}
				}
			}

			if(car->type == ST_FUNC && car->d.func.macro & 1) {
				/* macro. execute ~ */
				/* do it with inhibit_macro_expansion=1 atm */
				s_obj *res = funcexec(ctx, NULL, car, expr->d.cons.cdr, 0, 1);
#ifdef REPORT_EXPANSION
				printf("MACRO ");
				printobj(ctx, expr);
				printf(" => ");
				printobj(ctx, res);
				printf("\n");
#endif
				return compile_intr(ctx, res, callee, compx);
			}

			if(car->type == ST_BUILTIN && car->d.builtin.sform) {
				cdr = expr->d.cons.cdr;
			} else {
				cdr = compile_arglist(ctx, expr->d.cons.cdr, compx);
				if(cdr->type == ST_ERROR) return cdr;
			}

			return create_cons(car, cdr);
		case ST_SYMBOL:
			/* Symbol */
			if(expr->d.symbol.colon)
				return expr; /* keyword symbol */

			/* TODO: look up variable in compiler data structures. */
			return expr;
		default:
			return expr;
		}
}

/* compile an expression */
s_obj *compile(varctx *ctx, s_obj *expr)
{
#ifdef COMPILER
	compx *compx = compx_create();
	s_obj *res = compile_intr(ctx, expr, 0, compx);
#ifdef REPORT_EXPANSION
	printf("COMPILE ");
	printobj(ctx, expr);
	printf(" => ");
	printobj(ctx, res);
	printf("\n");
#endif

	compx_free(compx);
	return res;
#else
	return expr;
#endif
}

/* load and evaluate from fp */
s_obj *evalfp(varctx *ctx, FILE *fp)
{
	int c, ol;
	s_mark mark;
	s_obj *obj = NULL;
	s_obj *lobj = NULL;
	read_state *rs;

	mark = gc_prot_mark();

	ol = _stutter_parsed_line;

	rs = create_read_state();

	for(;;) {
		c = fgetc(fp);
		obj = read_ch(rs, c);
		if(obj && obj->type == ST_ERROR) {
			free_read_state(rs);
			gc_prot_free(mark, obj);
			return obj;
		}

		if(obj) {
			lobj = obj = eval(ctx, compile(ctx, obj));
			if(obj->type == ST_ERROR) {
				if(obj->d.error.code == SE_STOPREAD)
					obj = create_nil();
				free_read_state(rs);
				gc_prot_free(mark, obj);
				return obj;
			}
		}

		if(c == EOF) break;
	}

	free_read_state(rs);
	
	_stutter_parsed_line = ol;

	if(!lobj) lobj = create_nil();
	gc_prot_free(mark, lobj);
	return lobj;
}

/* load and evaluate file */
s_obj *evalfile(varctx *ctx, const char *name)
{
	FILE *fp;
	s_obj *r;
	char *of;
	char buf[1024];

	fp = fopen(name, "r");
	if(!fp) {
		sprintf(buf, "Can't open file %s", name);
		return create_error(SE_IO, buf);;
	}

	of = _stutter_file;
	mark_miscfree(_stutter_file = strdup(name));
	r = evalfp(ctx, fp);
	fclose(fp);
	_stutter_file = of;

	return r;
}


/* create a compilation context */
compx * compx_create()
{
	compx *cx = (compx *) xmalloc(sizeof(compx));
	cx->list = NULL;
	return cx;
}

/* destroy a compilation context */
void compx_free(compx *cx)
{
	while(cx->list) compx_popcsym(cx);
	free(cx);
}

/* push a bound symbol element on a compilation context */
void compx_pushcsym(compx *cx, int id, int stack)
{
	csym *c = (csym *) xmalloc(sizeof(csym));
	c->id = id;
	c->stack = stack;
	c->next = cx->list;
	cx->list = c;
}

/* pop bound symbol element from a compilation context */
void compx_popcsym(compx *cx)
{
	csym *next;

	if(cx->list) {
		next = cx->list->next;
		free(cx->list);
		cx->list = next;
	}
}

