#ifndef _ST_MEMORY_H
#define _ST_MEMORY_H

#include "types.h"

extern unsigned long int _stutter_gc_mark;
extern unsigned long int _stutter_gc_allocd; /* number of objects allocated since last GC run */
extern unsigned long int _stutter_gc_live; /* number of live objects left */
extern unsigned __int64 _stutter_varctx_id; /* first free varctx id */
extern unsigned long _stutter_symbol_id; /* first free symbol id */

/* generation sweep counters
 * these decrease until 1, then generation is sweeped
 * and counter is reset. */
extern int gc_gen2_count;
extern int gc_gen3_count;

/* perform GC each time after this many objects are allocated */
extern int GC_ALLOC_LIMIT;

/* verbose GC */
extern int GC_VERBOSE;

typedef int s_mark;

#define EXIT_MEMORY_FAIL 1

/* wrapper for calloc with error handling */
inline void *xcalloc(size_t nmemb, size_t size);
/* wrapper for malloc with error handling */
inline void *xmalloc(size_t size);
/* wrapper for realloc with error handling */
inline void *xrealloc(void *ptr, size_t size);

/* try to take an element of the recycle list */
s_obj * obj_respawn();

/* variable */
/* typedef struct s_var s_var; // forward-declared in types.h */

struct s_var {
	char *name; /* name */
	unsigned long sdbm; /* name sdbm */
	unsigned long id; /* symbol id */
	s_obj *value; /* value */
	s_var *next; /* next variable in this chain */
};

/* variable context */

/* forward declaration in types.h: */
// typedef struct varctx varctx;

struct varctx {
	varctx *parent; /* parent context */
	int ns; /* number of slots in the variable hash (power of two) */
	int nsm; /* ns - 1 (mask) */
	s_var **hash; /* variable hash */
	unsigned long id; /* varctx ID */
	int weak; /* 1 if weak varctx (local set doesn't create a new bind) */

	s_obj *ref; /* varctx object that references this ctx */
	unsigned long int mark; /* used by GC */
	int defer; /* use a deferred hash (for closures) */
};

#define SYMBOL_NAME_HASH_SIZE 8192
extern varctx * symbol_name_hash; /* global symbol name -> unique id hash */

extern varctx * class_hash; /* name -> class hash */

extern s_obj * class_precedence; /* precedence list of classes */

/* used list of protected varctx roots */

typedef struct s_protroot s_protroot;
struct s_protroot {
	varctx *root;
	s_protroot *next;
};

/* protect a varctx root (scan for variables in it) */
void gc_protect_root(varctx *ctx);
/* unprotect a varctx root (no longer scan for variables in it) */
void gc_unprotect_root(varctx *ctx);

/* add an object to the root set */
void gc_protect_obj(s_obj *obj);

/* remove an object from the root set */
void gc_unprotect_obj(s_obj *obj);

/* create variable context */
varctx *varctx_create(varctx *parent, const int size);

/* compute hash value for a string */
unsigned long sdbm(char *str);

/* get symbol name by symbol id */
char * get_symbolname(unsigned long sdbmv, unsigned long id);

/* get unique symbol id of a symbol */
unsigned long get_symbolid(char *name, unsigned long sdbmv);

/* utility function: get symbol id by string */
unsigned long symbolid(char *name);

/* store a variable inside a ctx */
void varctx_store(varctx *ctx, char *name, s_obj *val);
void varctx_store_sdbm(varctx *ctx, char *name, s_obj *val, unsigned long sdbmv);
void varctx_store_s(varctx *ctx, unsigned long sdbmv, unsigned long id, s_obj *val);

/* get variable from a ctx (local only)
 * returns NULL if variable doesn't exist */
s_var *varctx_retr(varctx *ctx, char *name);
s_var *varctx_retr_sdbm(varctx *ctx, char *name, unsigned long sdbmv);
s_var *varctx_retr_s(varctx *ctx, unsigned long sdbmv, unsigned long id);

/* get variable from a ctx (with parent fallback)
 * returns NULL if variable doesn't exist */
s_var *varctx_get(varctx *ctx, char *name);
s_var *varctx_get_sdbm(varctx *ctx, char *name, unsigned long sdbmv);
s_var *varctx_get_s(varctx *ctx, unsigned long sdbmv, unsigned long id);

/* set/store a variable inside a ctx
 * (updates if the var already exists) */
void varctx_set(varctx *ctx, char *name, s_obj *val);
void varctx_set_sdbm(varctx *ctx, char *name, s_obj *val, unsigned long sdbmv);
void varctx_set_s(varctx *ctx, unsigned long sdbmv, unsigned long id, s_obj *val);

/* dump contents of varctx to stderr */
s_obj * varctx_dump(varctx *ctx);

/* free varctx */
void varctx_free(varctx *ctx);

/* add object to protection list */
inline void gc_protlist_add(s_obj *obj);

/* record allocated object for GC */
void gc_record(s_obj *obj);

/* remove last allocated object from GC records (in case you want to free it yourself) */
void gc_unrecord();

/* checks whether the link should be followed
 * (not if the target object has the link in its weakptr list 
 * invokes gc_mark_obj as necessary */
inline void gc_follow(s_obj **obj);

/* free all data associated with object */
inline void gc_harvest(s_obj *obj);

/* mark variables reachable from object */
void gc_mark_obj(s_obj *obj);

/* mark variables in context */
void gc_mark(varctx *ctx);

/* do garbage collection. returns number of objects freed */
int gc_sweep(varctx *ctx);

/* get current mark in protection stack */
s_mark gc_prot_mark(void);

/* free up to mark in protection stack, except for object ret */
void gc_prot_free(s_mark mark, s_obj *ret);

/* mark misc. pointer for free() */
void mark_miscfree(void *ptr);

/* free pointers in miscfree list */
void free_miscfree();

/* initialize Stutter (called by resetparser if necessary) */
void stutter_init();

/* shut down Stutter */
void stutter_shutdown();

/* newhash */
unsigned long hash(unsigned char *k, unsigned long int length, unsigned long int initval);

#endif
