#include "stutter.h"
#include <sqlite3.h>
#include <stdio.h>

/* builtin: sqlite-open */
BUILTIN(sb_sqlite_open)
{
	sqlite3 *inst;

	ARG_STRING(obj);

	if(sqlite3_open(obj->d.string.value, &inst)) {
		obj = create_error(SE_SQLITE, sqlite3_errmsg(inst));
		sqlite3_close(inst);
		return obj;
	}

	obj = create_cookie(COOKIE_SQLITE, (void *) inst);
	return obj;
}

/* builtin: sqlite-close */
BUILTIN(sb_sqlite_close)
{
	ARG_OBJ(obj);
	if(obj->type != ST_COOKIE || obj->d.cookie.type != COOKIE_SQLITE)
		return create_error(SE_TYPE, "Invalid cookie given to sqlite-close");

	if(obj->d.cookie.ptr) {
		if(sqlite3_close((sqlite3*) obj->d.cookie.ptr)) {
			return create_error(SE_SQLITE, sqlite3_errmsg((sqlite3*) obj->d.cookie.ptr));
		} else obj->d.cookie.ptr = NULL;
	} else {
		return create_error(SE_PARM, "Sqlite handle already closed");
	}

	return create_nil();
}

/* sqlite-exec callback (calls user function with each row) */
static int sqlite_exec_lambda_callback(void *result, int argc, char **argv, char **azColName)
{
	s_obj *call = (s_obj *) ((void **) result)[0];
	varctx *ctx = (varctx *) ((void **) result)[1];
	s_obj *head, *tail;
	int i;
	s_obj *ret;
	s_mark mark;

	if(call->d.cons.cdr->d.cons.cdr->d.cons.car->type == ST_NIL) {
		/* first row, write column titles */

		head = NULL; tail = NULL;

		for(i=0; i<argc; i++) {
			if(!tail) {
				head = tail = create_cons(create_string(azColName[i]), create_nil());
			} else {
				tail->d.cons.cdr = create_cons(create_string(azColName[i]), create_nil());
				tail = tail->d.cons.cdr;
			}
		}

		call->d.cons.cdr->d.cons.cdr->d.cons.car = head ? quoteobj(head) : create_nil();
	}

	mark = gc_prot_mark();

	/* write row contents */

	head = NULL; tail = NULL;
	for(i=0; i<argc; i++) {
		if(!tail) {
			head = tail = create_cons(create_string(argv[i]), create_nil());
		} else {
			tail->d.cons.cdr = create_cons(create_string(argv[i]), create_nil());
			tail = tail->d.cons.cdr;
		}
	}

	call->d.cons.cdr->d.cons.car = head ? quoteobj(head) : create_nil();

	ret = eval(ctx, call);
	if(ret->type == ST_ERROR) {
		gc_prot_free(mark, ret);
		((void **) result)[1] = ret;
		return 1;
	}
	gc_prot_free(mark, NULL);
	return 0;
}

/* sqlite-exec callback (writes results into a list) */
static int sqlite_exec_callback(void *result, int argc, char **argv, char **azColName)
{
	s_obj ***resultptr = result;
	s_obj *head, *tail;
	int i;

	if((**resultptr)->type == ST_NIL) {
		/* first row, write column titles */

		head = NULL; tail = NULL;

		for(i=0; i<argc; i++) {
			if(!tail) {
				head = tail = create_cons(create_string(azColName[i]), create_nil());
			} else {
				tail->d.cons.cdr = create_cons(create_string(azColName[i]), create_nil());
				tail = tail->d.cons.cdr;
			}
		}

		**resultptr = create_cons(head ? head : create_nil(), create_nil());
	}

	/* write row contents */

	head = NULL; tail = NULL;
	for(i=0; i<argc; i++) {
		if(!tail) {
			head = tail = create_cons(create_string(argv[i]), create_nil());
		} else {
			tail->d.cons.cdr = create_cons(create_string(argv[i]), create_nil());
			tail = tail->d.cons.cdr;
		}
	}

	(**resultptr)->d.cons.cdr = create_cons(head ? head : create_nil(), create_nil());
	*resultptr = &((**resultptr)->d.cons.cdr);

	return 0;
}

/* builtin: sqlite-exec */
BUILTIN(sb_sqlite_exec)
{
	s_obj *fn;
	s_obj *result;
	void *fnresult[2];
	s_obj **resultptr;
	char *errmsg;
	int r;

	ARG_OBJ(obj);
	if(obj->type != ST_COOKIE || obj->d.cookie.type != COOKIE_SQLITE)
		return create_error(SE_TYPE, "Invalid cookie given to sqlite-exec");

	ARG_STRING(qry);

	if(nextarg(ctx, &fn, &parm)) {
		/* lambda method */

		if(fn->type == ST_ERROR) return fn;
		if(fn->type != ST_FUNC && fn->type != ST_BUILTIN)
			return create_error(SE_TYPE, "Third parameter for sqlite-exec has to be a function");

		fnresult[0] = (void *) create_cons(fn, create_cons(create_nil(), create_cons(create_nil(), create_nil())));
		fnresult[1] = (void *) ctx;

		r = sqlite3_exec((sqlite3*) obj->d.cookie.ptr, qry->d.string.value, sqlite_exec_lambda_callback, (void *) fnresult, &errmsg);
		if(r == SQLITE_ABORT) {
			return fnresult[1];
		}

		if(r) {
			obj = create_error(SE_SQLITE, errmsg);
			sqlite3_free(errmsg);
			return obj;
		}

		return create_nil();

	} else {
		/* list method */

		result = create_nil();
		resultptr = &result;

		if(sqlite3_exec((sqlite3*) obj->d.cookie.ptr, qry->d.string.value, sqlite_exec_callback, (void *) &resultptr, &errmsg)) {
			obj = create_error(SE_SQLITE, errmsg);
			sqlite3_free(errmsg);
			return obj;
		}

		return result;
	}
}

/* builtin: sqlite-prepare */
BUILTIN(sb_sqlite_prepare)
{
	s_obj *stmt;
	const char *pzTail;
	sqlite3_stmt *ppStmt;

	ARG_OBJ(obj);
	if(obj->type != ST_COOKIE || obj->d.cookie.type != COOKIE_SQLITE)
		return create_error(SE_TYPE, "Invalid cookie given to sqlite-prepare");

	ARG_STRING(qry);

	if(sqlite3_prepare((sqlite3*) obj->d.cookie.ptr, qry->d.string.value, -1, &ppStmt, &pzTail)) {
		obj = create_error(SE_SQLITE, sqlite3_errmsg((sqlite3*) obj->d.cookie.ptr));
		return obj;
	}

	if(ppStmt) {
		stmt = create_cookie(COOKIE_SQLITE_STMT, (void*) ppStmt);
		return create_cons(obj, stmt);
	} else {
		return create_nil();
	}
}

/* builtin: sqlite-finalize */
BUILTIN(sb_sqlite_finalize)
{
	ARG_OBJ(obj);
	if(obj->type != ST_CONS || obj->d.cons.car->type != ST_COOKIE ||
			obj->d.cons.car->d.cookie.type != COOKIE_SQLITE ||
			obj->d.cons.cdr->type != ST_COOKIE ||
			obj->d.cons.cdr->d.cookie.type != COOKIE_SQLITE_STMT)
		return create_error(SE_TYPE, "Invalid cookie given to sqlite-finalize");

	if(sqlite3_finalize((sqlite3_stmt*) obj->d.cons.cdr->d.cookie.ptr)) {
		return create_error(SE_SQLITE, sqlite3_errmsg((sqlite3*) obj->d.cons.car->d.cookie.ptr));
	} else {
		obj->d.cons.cdr = create_nil();
	}

	return create_nil();
}

/* builtin: sqlite-reset */
BUILTIN(sb_sqlite_reset)
{
	ARG_OBJ(obj);
	if(obj->type != ST_CONS || obj->d.cons.car->type != ST_COOKIE ||
			obj->d.cons.car->d.cookie.type != COOKIE_SQLITE ||
			obj->d.cons.cdr->type != ST_COOKIE ||
			obj->d.cons.cdr->d.cookie.type != COOKIE_SQLITE_STMT)
		return create_error(SE_TYPE, "Invalid cookie given to sqlite-reset");

	if(sqlite3_reset((sqlite3_stmt*) obj->d.cons.cdr->d.cookie.ptr)) {
		return create_error(SE_SQLITE, sqlite3_errmsg((sqlite3*) obj->d.cons.car->d.cookie.ptr));
	}

	return obj;
}

/* builtin: sqlite-bind-parameter-count */
BUILTIN(sb_sqlite_bind_parameter_count)
{
	int cnt;
	ARG_OBJ(obj);
	if(obj->type != ST_CONS || obj->d.cons.car->type != ST_COOKIE ||
			obj->d.cons.car->d.cookie.type != COOKIE_SQLITE ||
			obj->d.cons.cdr->type != ST_COOKIE ||
			obj->d.cons.cdr->d.cookie.type != COOKIE_SQLITE_STMT)
		return create_error(SE_TYPE, "Invalid cookie given to sqlite-bint-parameter-count");

	cnt = sqlite3_bind_parameter_count((sqlite3_stmt*) obj->d.cons.cdr->d.cookie.ptr);
	return create_int(cnt);
}

/* builtin: sqlite-bind */
BUILTIN(sb_sqlite_bind)
{
	s_obj *value;
	int idx;

	ARG_OBJ(obj);
	if(obj->type != ST_CONS || obj->d.cons.car->type != ST_COOKIE ||
			obj->d.cons.car->d.cookie.type != COOKIE_SQLITE ||
			obj->d.cons.cdr->type != ST_COOKIE ||
			obj->d.cons.cdr->d.cookie.type != COOKIE_SQLITE_STMT)
		return create_error(SE_TYPE, "Invalid cookie given to sqlite-bind");

	ARG_OBJ(cnt);
	if(cnt->type != ST_INT && cnt->type != ST_STRING)
		return create_error(SE_TYPE, "Second parameter for sqlite-bind must be an integer or a string");
	if(cnt->type == ST_STRING) {
		idx = sqlite3_bind_parameter_index((sqlite3_stmt*) obj->d.cons.cdr->d.cookie.ptr, cnt->d.string.value);
		if(idx == 0) return create_error(SE_PARM, "Named parameter wasn't found during sqlite-bind");
	} else {
		idx = cnt->d.integer.value;
	}

	if(!nextarg(ctx, &value, &parm))
		return create_error(SE_PNUM, "Insufficient parameters for sqlite-bind");
	switch(value->type) {
	case ST_ERROR:
		return value;
	case ST_INT:
		if(sqlite3_bind_int((sqlite3_stmt*) obj->d.cons.cdr->d.cookie.ptr, idx, value->d.integer.value)) {
			return create_error(SE_SQLITE, sqlite3_errmsg((sqlite3*) obj->d.cons.car->d.cookie.ptr));
		}
		break;
	case ST_NIL:
		if(sqlite3_bind_null((sqlite3_stmt*) obj->d.cons.cdr->d.cookie.ptr, idx)) {
			return create_error(SE_SQLITE, sqlite3_errmsg((sqlite3*) obj->d.cons.car->d.cookie.ptr));
		}
		break;
	case ST_NUM:
		if(sqlite3_bind_double((sqlite3_stmt*) obj->d.cons.cdr->d.cookie.ptr, idx, value->d.num.value)) {
			return create_error(SE_SQLITE, sqlite3_errmsg((sqlite3*) obj->d.cons.car->d.cookie.ptr));
		}
		break;
	case ST_STRING:
		if(sqlite3_bind_text((sqlite3_stmt*) obj->d.cons.cdr->d.cookie.ptr, idx, value->d.string.value, -1, SQLITE_TRANSIENT)) {
			return create_error(SE_SQLITE, sqlite3_errmsg((sqlite3*) obj->d.cons.car->d.cookie.ptr));
		}
		break;
	default:
		return create_error(SE_TYPE, "sqlite-bind can't bind given type");
	}

	return value;
}

/* builtin: sqlite-step */
BUILTIN(sb_sqlite_step)
{
	s_obj *value, *lhead, *ltail;
	int r, cnt, i;
	sqlite3_stmt *stmt;

	ARG_OBJ(obj);
	if(obj->type != ST_CONS || obj->d.cons.car->type != ST_COOKIE ||
			obj->d.cons.car->d.cookie.type != COOKIE_SQLITE ||
			obj->d.cons.cdr->type != ST_COOKIE ||
			obj->d.cons.cdr->d.cookie.type != COOKIE_SQLITE_STMT)
		return create_error(SE_TYPE, "Invalid cookie given to sqlite-step");

	stmt = (sqlite3_stmt*) obj->d.cons.cdr->d.cookie.ptr;
	r = sqlite3_step(stmt);
	switch(r) {
	case SQLITE_DONE:
		return create_nil();
	case SQLITE_ROW:
		cnt = sqlite3_column_count(stmt);
		lhead = NULL; ltail = NULL;

		for(i=0; i<cnt; i++) {
			switch(sqlite3_column_type(stmt, i)) {
			case SQLITE_INTEGER:
				value = create_int(sqlite3_column_int(stmt, i));
				break;
			case SQLITE_FLOAT:
				value = create_num(sqlite3_column_double(stmt, i));
				break;
			case SQLITE_TEXT:
				value = create_string((const char *) sqlite3_column_text(stmt, i));
				break;
			case SQLITE_NULL:
				value = create_nil();
				break;
			default:
				return create_error(SE_TYPE, "sqlite-step doesn't support blobs in results yet");
			}

			if(!lhead) {
				lhead = ltail = create_cons(value, create_nil());
			} else {
				ltail->d.cons.cdr = create_cons(value, create_nil());
				ltail = ltail->d.cons.cdr;
			}
		}

		return lhead ? lhead : create_nil();
	default:
		return create_error(SE_SQLITE, sqlite3_errmsg((sqlite3*) obj->d.cons.car->d.cookie.ptr));
	}
}

/* builtin: sqlite-columns */
BUILTIN(sb_sqlite_columns)
{
	s_obj *value, *lhead, *ltail;
	int cnt, i;
	sqlite3_stmt *stmt;

	ARG_OBJ(obj);
	if(obj->type != ST_CONS || obj->d.cons.car->type != ST_COOKIE ||
			obj->d.cons.car->d.cookie.type != COOKIE_SQLITE ||
			obj->d.cons.cdr->type != ST_COOKIE ||
			obj->d.cons.cdr->d.cookie.type != COOKIE_SQLITE_STMT)
		return create_error(SE_TYPE, "Invalid cookie given to sqlite-columns");

	stmt = (sqlite3_stmt*) obj->d.cons.cdr->d.cookie.ptr;
	cnt = sqlite3_column_count(stmt);
	lhead = NULL; ltail = NULL;

	for(i=0; i<cnt; i++) {
		value = create_string(sqlite3_column_name(stmt, i));
		if(!lhead) {
			lhead = ltail = create_cons(value, create_nil());
		} else {
			ltail->d.cons.cdr = create_cons(value, create_nil());
			ltail = ltail->d.cons.cdr;
		}
	}

	return lhead ? lhead : create_nil();
}

/**
 * intialize sqlite module
 */
void modinit_sqlite(varctx *ctx)
{
	varctx_set(ctx, "SQLITE-OPEN", create_builtin(sb_sqlite_open));
	varctx_set(ctx, "SQLITE-CLOSE", create_builtin(sb_sqlite_close));
	varctx_set(ctx, "SQLITE-EXEC", create_builtin(sb_sqlite_exec));
	varctx_set(ctx, "SQLITE-PREPARE", create_builtin(sb_sqlite_prepare));
	varctx_set(ctx, "SQLITE-FINALIZE", create_builtin(sb_sqlite_finalize));
	varctx_set(ctx, "SQLITE-BIND", create_builtin(sb_sqlite_bind));
	varctx_set(ctx, "SQLITE-RESET", create_builtin(sb_sqlite_reset));
	varctx_set(ctx, "SQLITE-BIND-PARAMETER-COUNT", create_builtin(sb_sqlite_bind_parameter_count));
	varctx_set(ctx, "SQLITE-STEP", create_builtin(sb_sqlite_step));
	varctx_set(ctx, "SQLITE-COLUMNS", create_builtin(sb_sqlite_columns));
}
