#include "stutter.h"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/* resolve &entity; to string */
static char * entresolve(char *ent)
{
	if(!strcmp(ent, "lt"))
		return strdup("<");
	if(!strcmp(ent, "gt"))
		return strdup(">");
	if(!strcmp(ent, "amp"))
		return strdup("&");
	if(!strcmp(ent, "apos"))
		return strdup("'");
	if(!strcmp(ent, "quot"))
		return strdup("\"");

	if(ent[0] == '#') {
		/* numeric character reference */

		char *start = ent + 1;
		char *endptr;
		int base = 10;
		int val;
		char buf[16];

		if(ent[1] == 'x') {
			start ++;
			base = 16;
		}

		/* fail in case of &#; &#x; */
		if(!*start) return NULL;

		val = strtol(start, &endptr, base);
		/* has to be a valid number */
		if(*endptr != 0) return NULL;

		/* no negatives kthx */
		if(val < 0) return NULL;

		/* 7 bit ascii */
		if(val < 0x80) {
			buf[0] = val;
			buf[1] = 0;
			return strdup(buf);
		}

		if(val < 0x800) {
			buf[0] = 128 + 64 + (val >> 6);
			buf[1] = 128 + (val & 63);
			buf[2] = 0;
			return strdup(buf);
		}

		if(val < 0x10000) {
			buf[0] = 128 + 64 + 32 + (val >> 12);
			buf[1] = 128 + ((val >> 6) & 63);
			buf[2] = 128 + (val & 63);
			buf[3] = 0;
			return strdup(buf);
		}

		buf[0] = 128 + 64 + 32 + 16 + (val >> 18);
		buf[1] = 128 + ((val >> 12) & 63);
		buf[2] = 128 + ((val >> 6) & 63);
		buf[3] = 128 + (val & 63);
		return strdup(buf);
	}

	return NULL;
}

/* xml reader states */
enum {
	XS_NORMAL,						/* between markup (character data) */
	XS_START,						/* after '<' */
	XS_ENTITY,						/* after '&' */
	XS_END,							/* after '</' */
	XS_BANG,						/* after '<!' */
	XS_START_TAG,					/* after '<'+character */
	XS_ATTR,						/* before an attribute */
	XS_ATTRNAME,					/* in attribute name */
	XS_ATTREQ,						/* between attribute name and value */
	XS_ATTR_APOS,					/* in attribute value, delimited by ' */
	XS_ATTR_QUOT,					/* in attribute value, delimited by " */
	XS_EMPTY,						/* between / and > of an empty entity tag */
	XS_BANG_CDATA,					/* after '<![' */
	XS_BANG_MINUS,					/* after '<!-' */
	XS_COMMENT,						/* in a comment */
	XS_COMMENT_MINUS,				/* in a comment, after '-' */
	XS_COMMENT_MINUS_MINUS,			/* in a comment, after '--' */
	XS_CDATA,						/* in a CDATA block */
	XS_CDATA_SB,					/* in a CDATA block, after ']' */
	XS_CDATA_SB_SB					/* in a CDATA block, after ']]' */
};


/* builtin: read-xml */
BUILTIN(sb_read_xml)
{
	ARG_FUNCTION(fnstart); /* callback for start tag */
	ARG_FUNCTION(fnend);   /* callback for end tag */
	ARG_FUNCTION(fnchar);  /* callback for character data */

	int state = XS_NORMAL;

	s_obj *ch;
	s_obj *ret;
	s_obj *tagstr;

	int begin_entity = 0;
	int entity_from = 0; /* after an entity ends, state goes to entity_from */ 
	char *p, *txt;

	char *tagname = NULL;

	/* current token data */
	char * token = NULL;
	int token_len = 0; /* significant bytes in token (strlen + 1) */
	int token_alloc = 0; /* bytes allocated for token */

	s_obj *attrlist = create_nil();
	s_obj *attrlist_tail = create_nil();

	s_mark mark = gc_prot_mark();

	/* append character to current token */
	void token_appch(int ch)
	{
		if(!token) {
			token = (char *) xmalloc(8);
			token_alloc = 8;
			token_len = 2;
			token[0] = ch;
			token[1] = 0;
			return;
		} else {
			if(token_len == token_alloc)
				token = (char *) xrealloc(token, token_alloc *= 2);
			token[token_len] = 0;
			token[(token_len++)-1] = ch;
		}
	}

	/* reset token */
	void token_reset()
	{
		free(token);
		token = NULL;
		token_alloc = 0;
		token_len = 0;
	}

	/* dispatch character data callback */
	s_obj * dispatch_fnchar(varctx *ctx)
	{
		if(token) {
			return eval(ctx, create_cons(fnchar, create_cons(create_string(token), create_nil())));
		}

		return create_t();
	}

	/* free all allocated data */
	void killdata()
	{
		token_reset();
		free(tagname);
		tagname = NULL;
	}

	/* append to attr list */
	void attrlist_append()
	{
		if(attrlist->type == ST_NIL) {
			attrlist_tail = attrlist = create_cons(create_cons(create_nil(), create_nil()), create_nil());
		} else {
			attrlist_tail->d.cons.cdr = create_cons(create_cons(create_nil(), create_nil()), create_nil());
			attrlist_tail = attrlist_tail->d.cons.cdr;
		}
	}

	/* main loop */
	while(1) {
		/* read next character from stream */
		ch = stream_stdin_read(ctx);
		if(ch->type == ST_ERROR) {
			free(token);
			return ch;
		}

		switch(state) {

		case XS_NORMAL:
			/* normal state (outside markup) */

			if(ch->type == ST_EOF) {
				ret = dispatch_fnchar(ctx);
				killdata();
				return create_eof();
			}

			switch(ch->d.ch.ch) {
			case '<':
				/* start tag */
				/* if we have character data, dispatch it */
				ret = dispatch_fnchar(ctx);
				token_reset();
				if(ret->type == ST_ERROR || ret->type == ST_NIL) {
					stream_stdin_unread(ctx, '<');
					killdata();
					return ret;
				}

				state = XS_START;
				continue;

			case '>':
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected '>' in text");

			case '&':
				/* start entity */
				token_appch('&');
				/* save where entity begins in text */
				begin_entity = token_len - 2;

				state = XS_ENTITY;
				entity_from = XS_NORMAL;
				continue;

			default:
				token_appch(ch->d.ch.ch);
				continue;
			}

		case XS_ENTITY:
			/* after & */

			if(ch->type == ST_EOF) {
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected EOF in entity");
			}

			if(ch->d.ch.ch == ';') {
				/* end of entity, try to resolve */
				txt = entresolve(token + begin_entity + 1);
				if(!txt) {
					ret = create_error_with_obj(SE_PARSE, "read-xml: can't resolve entity", create_string(token + begin_entity));
					killdata();
					return ret;
				}

				/* append resolved text starting at & */
				token_len = begin_entity + 1;
				for(p = txt; *p; p++)
					token_appch(*p);

				free(txt);
				state = entity_from;
				continue;
			} else {
				/* entity text */
				token_appch(ch->d.ch.ch);
				continue;
			}

		case XS_START:
			/* after '<' */

			if(ch->type == ST_EOF) {
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected EOF after <");
			}

			switch(ch->d.ch.ch) {
			case '/':
				/* end tag */
				state = XS_END;
				continue;

			case '!':
				/* beginning of comment or cdata */
				state = XS_BANG;
				continue;

			default:
				/* real tag, accept / ! no longer */
				token_appch(ch->d.ch.ch);
				state = XS_START_TAG;
				continue;
			}

		case XS_BANG:
			/* after '!' */

			if(ch->type == ST_EOF) {
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected EOF after <!");
			}

			switch(ch->d.ch.ch) {
			case '-':
				/* <!- */
				state = XS_BANG_MINUS;
				continue;

			case '[':
				state = XS_BANG_CDATA;
				continue;

			default:
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected stuff after <!");
			}

		case XS_BANG_CDATA:
			/* after '<![' */

			if(ch->type == ST_EOF) {
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected EOF after <![");
			}

			switch(ch->d.ch.ch) {
			case '[':
				state = XS_CDATA;
				continue;

			case 'C':
			case 'D':
			case 'A':
			case 'T':
				/* not entirely legal. */
				continue;

			default:
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected stuff after <![");
			}

		case XS_CDATA:
			/* inside CDATA */

			if(ch->type == ST_EOF) {
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected EOF in CDATA");
			}

			switch(ch->d.ch.ch) {
			case ']':
				state = XS_CDATA_SB;
				continue;

			default:
				token_appch(ch->d.ch.ch);
				continue;
			}

		case XS_CDATA_SB:
			/* inside CDATA, after ] */

			if(ch->type == ST_EOF) {
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected EOF in CDATA after ]");
			}

			switch(ch->d.ch.ch) {
			case ']':
				state = XS_CDATA_SB_SB;
				continue;

			default:
				token_appch(']');
				token_appch(ch->d.ch.ch);
				state = XS_CDATA;
				continue;
			}

		case XS_CDATA_SB_SB:
			/* inside CDATA, after ]] */

			if(ch->type == ST_EOF) {
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected EOF in CDATA after ]]");
			}

			switch(ch->d.ch.ch) {
			case '>':
				state = XS_NORMAL;
				continue;

			default:
				token_appch(']');
				token_appch(']');
				token_appch(ch->d.ch.ch);
				state = XS_CDATA;
				continue;
			}

		case XS_BANG_MINUS:
			/* after '<!-' */

			if(ch->type == ST_EOF) {
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected EOF after <!-");
			}

			switch(ch->d.ch.ch) {
			case '-':
				state = XS_COMMENT;
				continue;

			default:
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected stuff after <!-");
			}

		case XS_COMMENT:
			/* inside comment */

			if(ch->type == ST_EOF) {
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected EOF inside comment");
			}

			if(ch->d.ch.ch == '-')
				state = XS_COMMENT_MINUS;

			continue;

		case XS_COMMENT_MINUS:
			/* inside comment, after - */

			if(ch->type == ST_EOF) {
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected EOF after comment minus");
			}

			if(ch->d.ch.ch == '-')
				state = XS_COMMENT_MINUS_MINUS;
			else
				state = XS_COMMENT;

			continue;

		case XS_COMMENT_MINUS_MINUS:
			/* inside comment, after -- */

			if(ch->type == ST_EOF) {
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected EOF after comment minus minus");
			}

			if(ch->d.ch.ch == '>')
				state = XS_NORMAL;
			else
				state = XS_COMMENT;

			continue;

		case XS_START_TAG:
			/* after '<x' */

			if(ch->type == ST_EOF) {
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected EOF in tagname");
			}

			switch(ch->d.ch.ch) {
			case ' ':
			case '\n':
			case '\r':
			case '\t':
				/* tag ends, attributes begin */
				tagname = strdup(token);
				token_reset();
				state = XS_ATTR;
				continue;

			case '/':
				tagname = strdup(token);
				token_reset();
				state = XS_EMPTY;
				continue;

			case '?':
				/* skip these so <?xml ?> style declarations don't mess things up. fugly */
				continue;

			case '>':
				/* tag ends, no attributes */
				/* call (fnstart "tagname" nil) */
				ret = eval(ctx, create_cons(fnstart,
									create_cons(create_string(token),
										create_cons(create_nil(), create_nil()))));
				token_reset();
				if(ret->type == ST_ERROR || ret->type == ST_NIL) {
					killdata();
					return ret;
				}

				state = XS_NORMAL;
				continue;

			default:
				/* append to tag name */
				token_appch(ch->d.ch.ch);
				continue;
			}

		case XS_EMPTY:
			/* between / and final > */

			if(ch->type == ST_EOF) {
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected EOF between / and >");
			}

			switch(ch->d.ch.ch) {
			case ' ':
			case '\n':
			case '\r':
			case '\t':
				continue;

			case '>':
				/* tag ends */
				/* call (fnstart "tagname" 'attrlist) */

				if(tagname[0] != '?') {
					tagstr = create_string(tagname);
					free(tagname);
					tagname = NULL;

					ret = eval(ctx, create_cons(fnstart,
										create_cons(tagstr,
											create_cons(quoteobj(attrlist), create_nil()))));
					if(ret->type == ST_ERROR || ret->type == ST_NIL) {
						killdata();
						return ret;
					}

					/* call (fnend "tagname") */
					ret = eval(ctx, create_cons(fnend,
										create_cons(tagstr, create_nil())));
					if(ret->type == ST_ERROR || ret->type == ST_NIL) {
						killdata();
						return ret;
					}
				} else {
					/* skip <?xml style stuff. */
					free(tagname);
					tagname = NULL;
				}

				attrlist = create_nil();
				attrlist_tail = create_nil();

				gc_prot_free(mark, NULL);
				state = XS_NORMAL;
				continue;

			default:
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected stuff between / and >");
			}

		case XS_ATTR:
			/* after '<abc ' */

			if(ch->type == ST_EOF) {
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected EOF before attr name");
			}

			/* skip space until attribute name starts. */
			switch(ch->d.ch.ch) {
			case ' ':
			case '\n':
			case '\r':
			case '\t':
				continue;

			case '/':
				/* empty element tag */
				state = XS_EMPTY;
				continue;

			case '?':
				/* skip these so <?xml ?> style declarations don't mess things up. fugly */
				continue;

			case '>':
				/* tag ends */
				/* call (fnstart "tagname" 'attrlist) */

				if(tagname[0] != '?') {
					ret = eval(ctx, create_cons(fnstart,
										create_cons(create_string(tagname),
											create_cons(quoteobj(attrlist), create_nil()))));
					free(tagname);
					tagname = NULL;
					if(ret->type == ST_ERROR || ret->type == ST_NIL) {
						killdata();
						return ret;
					}
				} else {
					/* skip <?xml style stuff. */
					free(tagname);
					tagname = NULL;
				}

				attrlist = create_nil();
				attrlist_tail = create_nil();

				gc_prot_free(mark, NULL);
				state = XS_NORMAL;
				continue;

			default:
				/* attrname starts */
				attrlist_append();
				token_appch(ch->d.ch.ch);
				state = XS_ATTRNAME;
				continue;
			}

		case XS_ATTRNAME:
			/* in attribute name */

			if(ch->type == ST_EOF) {
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected EOF in attr name");
			}

			switch(ch->d.ch.ch) {
			case ' ':
			case '\n':
			case '\r':
			case '\t':
			case '=':
				/* end of attrname */
				attrlist_tail->d.cons.car->d.cons.car = create_string(token);
				token_reset();
				state = XS_ATTREQ;
				continue;

			default:
				/* append to attrname */
				token_appch(ch->d.ch.ch);
				continue;
			}

		case XS_ATTREQ:
			/* skip space and = until attribute value */

			if(ch->type == ST_EOF) {
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected EOF near attr =");
			}

			switch(ch->d.ch.ch) {
			case ' ':
			case '\n':
			case '\r':
			case '\t':
			case '=':
				continue;
			
			case '>':
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected > instead of attr value");

			case '\'':
				state = XS_ATTR_APOS;
				continue;

			case '"':
				state = XS_ATTR_QUOT;
				continue;
			}

		case XS_ATTR_QUOT:
			/* attribute value between " */

			if(ch->type == ST_EOF) {
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected EOF in attr value");
			}

			switch(ch->d.ch.ch) {
			case '"':
				/* end of value */
				attrlist_tail->d.cons.car->d.cons.cdr = create_string(token ? token : "");
				token_reset();
				state = XS_ATTR;
				continue;

			case '&':
				/* start entity */
				token_appch('&');
				/* save where entity begins in text */
				begin_entity = token_len - 2;

				state = XS_ENTITY;
				entity_from = XS_ATTR_QUOT;
				continue;

			default:
				/* append to value */
				token_appch(ch->d.ch.ch);
				continue;
			}

		case XS_ATTR_APOS:
			/* attribute value between ' */

			if(ch->type == ST_EOF) {
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected EOF in attr value");
			}

			switch(ch->d.ch.ch) {
			case '\'':
				/* end of value */
				attrlist_tail->d.cons.car->d.cons.cdr = create_string(token ? token : "");
				token_reset();
				state = XS_ATTR;
				continue;

			case '&':
				/* start entity */
				token_appch('&');
				/* save where entity begins in text */
				begin_entity = token_len - 2;

				state = XS_ENTITY;
				entity_from = XS_ATTR_APOS;
				continue;

			default:
				/* append to value */
				token_appch(ch->d.ch.ch);
				continue;
			}

		case XS_END:
			/* </ tag */

			if(ch->type == ST_EOF) {
				killdata();
				return create_error(SE_PARSE, "read-xml: unexpected EOF in end tag name");
			}

			/* after '</' */
			switch(ch->d.ch.ch) {
			case ' ':
			case '\n':
			case '\r':
			case '\t':
				/* space after tagname, skip */
				continue;

			case '>':
				/* end tag ends */

				if(!token) {
					/* </> */
					killdata();
					return create_error(SE_PARSE, "read-xml: empty end tag");
				}

				/* call (fnend "tagname") */
				ret = eval(ctx, create_cons(fnend,
									create_cons(create_string(token), create_nil())));
				token_reset();
				if(ret->type == ST_ERROR || ret->type == ST_NIL) {
					killdata();
					return ret;
				}

				state = XS_NORMAL;
				continue;

			default:
				/* append character to tag name */
				token_appch(ch->d.ch.ch);
				continue;
			}
		}

	}

}

/* builtin: xmlencode */
s_obj *sb_xmlencode (varctx *ctx, s_obj *parm, void *data, const int tail)
{
	s_obj *arg;
	unsigned char *c, *dest, *result_body;
	int estimated_length;

	if (!nextarg(ctx, &arg, &parm))
		return create_error(SE_PNUM, "XMLENCODE requires a parameter");
	if (arg->type == ST_ERROR)
		return arg;
	if (arg->type != ST_STRING)
		return create_error(SE_TYPE, "XMLENCODE's parameter must be a string");

	estimated_length = 1;
	for (c = (unsigned char *) arg->d.string.value; *c; c++)
		if (*c <= 0x1F || *c >= 0x7F || *c == '&' || *c == '<' || *c == '>' || *c == '"' || *c == '\'') {
			if (*c <= 9)
				estimated_length += 4;
			else if (*c <= 99)
				estimated_length += 5;
			else
				estimated_length += 6;
		} else {
			estimated_length++;
		}
	result_body = xmalloc(estimated_length);
	dest = result_body;
	for (c = (unsigned char *) arg->d.string.value; *c; c++)
		if (!isspace(*c) && (*c <= 0x1F || *c >= 0x7F || *c == '&' || *c == '<' || *c == '>' || *c == '"' || *c == '\'')) {
			dest += sprintf((char *) dest, "&#%i;", *c);
		} else {
			*dest++ = *c;
		}
		*dest = '\0';
	return create_string_allocd((char *) result_body);
}

/* builtin: utf8-xmlencode */
s_obj *sb_utf8_xmlencode (varctx *ctx, s_obj *parm, void *data, const int tail)
{
	s_obj *arg;
	unsigned char *c, *dest, *result_body;
	unsigned long int code = 0; // only initialised to avoid a spurious warning
	int estimated_length;
	int pending;

	if (!nextarg(ctx, &arg, &parm))
		return create_error(SE_PNUM, "UTF8-XMLENCODE requires a parameter");
	if (arg->type == ST_ERROR)
		return arg;
	if (arg->type != ST_STRING)
		return create_error(SE_TYPE, "UTF8-XMLENCODE's parameter must be a string");

	estimated_length = 1;
	pending = 0;
	for (c = (unsigned char *) arg->d.string.value; *c; c++) {
		if (*c <= 0x7E) {
			if (pending != 0)
				goto BAD_UTF8;
			code = *c;
		} else if (*c <= 0xBF) {
			if (pending == 0)
				goto BAD_UTF8;
			pending--;
			code <<= 6;
			code |= *c & 0x3F;
		} else if (*c <= 0xDF) {
			if (pending != 0)
				goto BAD_UTF8;
			pending = 1;
			code = *c & 0x1F;
		} else if (*c <= 0xEF) {
			if (pending != 0)
				goto BAD_UTF8;
			pending = 2;
			code = *c & 0x0F;
		} else if (*c <= 0xF7) {
			if (pending != 0)
				goto BAD_UTF8;
			pending = 3;
			code = *c & 0x07;
		} else if (*c <= 0xFB) {
			if (pending != 0)
				goto BAD_UTF8;
			pending = 4;
			code = *c & 0x03;
		} else if (*c <= 0xFD) {
			if (pending != 0)
				goto BAD_UTF8;
			pending = 5;
			code = *c & 0x01;
		} else {
			goto BAD_UTF8;
		}
		if (pending == 0) {
			if (code >= 0x20 && code <= 0x7E && code != '&' && code != '<' && code != '>' && code != '"' && code != '\'')
				estimated_length += 1;
			else if (code <= 9)
				estimated_length += 4;
			else if (code <= 99)
				estimated_length += 5;
			else if (code <= 999)
				estimated_length += 6;
			else if (code <= 9999)
				estimated_length += 7;
			else if (code <= 99999)
				estimated_length += 8;
			else if (code <= 999999)
				estimated_length += 9;
			else if (code <= 9999999)
				estimated_length += 10;
			else if (code <= 99999999)
				estimated_length += 11;
			else if (code <= 999999999)
				estimated_length += 12;
			else
				estimated_length += 13;
		}
	}
	if (pending != 0) {
BAD_UTF8:
		return create_error(SE_TYPE, "UTF8-XMLENCODE's parameter must be a UTF-8 encoded string");
	}
	result_body = xmalloc(estimated_length);
	dest = result_body;
	pending = 0;
	for (c = (unsigned char *) arg->d.string.value; *c; c++) {
		if (*c <= 0x7E) {
			if (pending != 0)
				goto BAD_UTF8;
			code = *c;
		} else if (*c <= 0xBF) {
			if (pending == 0)
				goto BAD_UTF8;
			pending--;
			code <<= 6;
			code |= *c & 0x3F;
		} else if (*c <= 0xDF) {
			if (pending != 0)
				goto BAD_UTF8;
			pending = 1;
			code = *c & 0x1F;
		} else if (*c <= 0xEF) {
			if (pending != 0)
				goto BAD_UTF8;
			pending = 2;
			code = *c & 0x0F;
		} else if (*c <= 0xF7) {
			if (pending != 0)
				goto BAD_UTF8;
			pending = 3;
			code = *c & 0x07;
		} else if (*c <= 0xFB) {
			if (pending != 0)
				goto BAD_UTF8;
			pending = 4;
			code = *c & 0x03;
		} else if (*c <= 0xFD) {
			if (pending != 0)
				goto BAD_UTF8;
			pending = 5;
			code = *c & 0x01;
		} else {
			goto BAD_UTF8;
		}
		if (pending == 0) {
			if (isspace(code) || (code >= 0x20 && code <= 0x7E && code != '&' && code != '<' && code != '>' && code != '"' && code != '\''))
				*dest++ = code;
			else if(code == '<')
				dest += sprintf((char *) dest, "&lt;");
			else if(code == '>')
				dest += sprintf((char *) dest, "&gt;");
			else if(code == '&')
				dest += sprintf((char *) dest, "&amp;");
			else if(code == '"')
				dest += sprintf((char *) dest, "&quot;");
			else {
				dest += sprintf((char *) dest, "&#%lu;", code);
			}
		}
	}
	if (pending != 0)
		goto BAD_UTF8;
	*dest = '\0';

	return create_string_allocd((char *) result_body);
}

/* builtin: urlencode */
s_obj *sb_urlencode (varctx *ctx, s_obj *parm, void *data, const int tail)
{
	s_obj *arg;
	unsigned char *c, *dest, *result_body;
	int estimated_length;

	if (!nextarg(ctx, &arg, &parm))
		return create_error(SE_PNUM, "URLENCODE requires a parameter");
	if (arg->type == ST_ERROR)
		return arg;
	if (arg->type != ST_STRING)
		return create_error(SE_TYPE, "URLENCODE's parameter must be a string");

	estimated_length = 1;
	for (c = (unsigned char *) arg->d.string.value; *c; c++)
		if ((*c >= '0' && *c <= '9') || (*c >= 'A' && *c <= 'Z') || (*c >= 'a' && *c <= 'z') || *c == '-' || *c == '.' || *c == '/') {
			estimated_length++;
		} else {
			estimated_length += 3;
		}
	result_body = xmalloc(estimated_length);
	dest = result_body;
	for (c = (unsigned char *) arg->d.string.value; *c; c++)
		if ((*c >= '0' && *c <= '9') || (*c >= 'A' && *c <= 'Z') || (*c >= 'a' && *c <= 'z') || *c == '-' || *c == '.' || *c == '/') {
			*dest++ = *c;
		} else {
			dest += sprintf((char *) dest, "%%%02x", *c);
		}
		*dest = '\0';
	return create_string_allocd((char *) result_body);
}

/* builtin: urldecode */
s_obj *sb_urldecode (varctx *ctx, s_obj *parm, void *data, const int tail)
{
	s_obj *arg;
	char *c, *dest, *result_body;

	if (!nextarg(ctx, &arg, &parm))
		return create_error(SE_PNUM, "URLDECODE requires a parameter");
	if (arg->type == ST_ERROR)
		return arg;
	if (arg->type != ST_STRING)
		return create_error(SE_TYPE, "URLDECODE's parameter must be a string");

	dest = result_body = xmalloc(strlen(arg->d.string.value) + 1);
	for(c = arg->d.string.value; *c; c++) {
		if (*c == '%') {
			char n1, n2;

			// First nibble
			c++;
			n1 = *c;
			if (n1 == '\0') {
				// Prematurely terminated code.
				*dest++ = '%';
				break;
			}
			if (n1 >= '0' && n1 <= '9') {
				n1 &= 0x0F;
			} else if ((n1 | 0x20) >= 'a' && (n1 | 0x20) <= 'f') {
				n1 &= 0x0F;
				n1 += 'A' - '9' + 1;
			} else {
				// Invalid code.
				*dest++ = '%';
				c--;
				continue;
			}

			// Second nibble
			c++;
			n2 = *c;
			if (n2 == '\0') {
				// Prematurely terminated code.
				*dest++ = '%';
				c -= 2;
				continue;
			}
			if (n2 >= '0' && n2 <= '9') {
				n2 &= 0x0F;
			} else if ((n2 | 0x20) >= 'a' && (n2 | 0x20) <= 'f') {
				n2 &= 0x0F;
				n2 += 'A' - '9' + 1;
			} else {
				// Invalid code.
				*dest++ = '%';
				c -= 2;
				continue;
			}
			*dest++ = (n1 << 4) | n2;
		} else if (*c == '+') {
			*dest++ = ' ';
		} else {
			*dest++ = *c;
		}
	}
	*dest++ = 0;

	return create_string_allocd(result_body);
}


/* initialize xml module */
void modinit_xml(varctx *ctx)
{
	varctx_set(ctx, "READ-XML", create_builtin(sb_read_xml));
	varctx_set(ctx, "XMLENCODE", create_builtin(sb_xmlencode));
	varctx_set(ctx, "UTF8-XMLENCODE", create_builtin(sb_utf8_xmlencode));
	varctx_set(ctx, "URLENCODE", create_builtin(sb_urlencode));
	varctx_set(ctx, "URLDECODE", create_builtin(sb_urldecode));
}
