#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "builtins.h"
#include "eval.h"
#include "types.h"
#include "read.h"

/* create read_state */
read_state * create_read_state(void)
{
	read_state *s = xcalloc(1, sizeof(read_state));
	read_reset(s);
	s->line = 1;
	s->col = 0;
	return s;
}

/* free current token data */
void resettoken(read_state *s)
{
	if(s->token) {
		free(s->token);
		s->token = NULL;
		s->buflen = s->strlen = 0;
	}
}

/* destroy read_state */
void free_read_state(read_state *s)
{
	resettoken(s);
	free(s);
}

/* reset parser */
void read_reset(read_state *s)
{
	resettoken(s);
	s->state = RS_NORM;
	s->stack = create_nil();
	s->out = create_nil();
	s->act = create_nil();
	s->ahead = 0;
}

/* append character to current token */
static void token_appch(read_state *s, const char ch)
{
	if(!s->token) {
		s->token = (char *) xmalloc(8);
		s->buflen = 8;
		s->strlen = 2;
		s->token[0] = ch;
		s->token[1] = 0;
		return;
	} else {
		if(s->strlen == s->buflen)
			s->token = (char *) xrealloc(s->token, s->buflen *= 2);

		s->token[s->strlen] = 0;
		s->token[(s->strlen++)-1] = ch;
	}
}

/* append object to current list */
int read_append(read_state *s, s_obj *obj)
{
	s_obj *prv;
	int code;

	if(s->act->type == ST_NIL) {
		/* list is currently empty */

		s->out = s->act = create_cons(obj, create_nil());
		return 0;
	} else if(s->act->type == ST_CONS && s->act->d.cons.cdr->type == ST_NIL) {

		s->act->d.cons.cdr = create_cons(obj, create_nil());
		if(obj->type == ST_ERROR) { /* appending list special */
			s->act = s->act->d.cons.cdr;
			return 0;
		}

		/* handling list specials */
		while(s->act->d.cons.car->type == ST_ERROR) {
			prv = s->act->d.cons.car->d.error.obj;
			code = s->act->d.cons.car->d.error.code;

			/* list special: . */
			if(code == SE_WTF) {
				if(prv->type != ST_CONS) return 1;
				if(s->act->d.cons.cdr->type != ST_CONS) return 2;
				prv->d.cons.cdr = s->act->d.cons.cdr->d.cons.car;
				s->act = prv;
				continue;
			}

			/* list special: ' */
			if(code == SE_UNDEF) {
				if(s->act->d.cons.cdr->type != ST_CONS) return 3;
				if(prv->type == ST_NIL) {
					s->act = s->out = create_cons(quoteobj(s->act->d.cons.cdr->d.cons.car), create_nil());
					return 0;
				} else if(prv->type == ST_CONS) {
					prv->d.cons.cdr = create_cons(quoteobj(s->act->d.cons.cdr->d.cons.car), create_nil());
					s->act = prv;
					continue;
				} else return 4;
			}

			/* list special: ` */
			if(code == SE_PARM) {
				if(s->act->d.cons.cdr->type != ST_CONS) return 5;
				if(prv->type == ST_NIL) {
					s->act = s->out = create_cons(create_cons(create_symbol("backquote"), create_cons(s->act->d.cons.cdr->d.cons.car, create_nil())), create_nil());
					return 0;
				} else if(prv->type == ST_CONS) {
					prv->d.cons.cdr = create_cons(create_cons(create_symbol("backquote"), create_cons(s->act->d.cons.cdr->d.cons.car, create_nil())), create_nil());
					s->act = prv;
					continue;
				} else return 6;
			}

			/* list special: , */
			if(code == SE_WFUNC) {
				if(s->act->d.cons.cdr->type != ST_CONS) return 7;
				if(prv->type == ST_NIL) {
					s->act = s->out = create_cons(create_cons(create_symbol("UNQUOTE"), create_cons(s->act->d.cons.cdr->d.cons.car, create_nil())), create_nil());
					return 0;
				} else if(prv->type == ST_CONS) {
					prv->d.cons.cdr = create_cons(create_cons(create_symbol("UNQUOTE"), create_cons(s->act->d.cons.cdr->d.cons.car, create_nil())), create_nil());
					s->act = prv;
					continue;
				} else return 8;
			}

			/* list special: ,@ */
			if(code == SE_TYPE) {
				if(s->act->d.cons.cdr->type != ST_CONS) return 9;
				if(prv->type == ST_NIL) {
					s->act = s->out = create_cons(create_cons(create_symbol("UNQUOTE-SPLICING"), create_cons(s->act->d.cons.cdr->d.cons.car, create_nil())), create_nil());
					return 0;
				} else if(prv->type == ST_CONS) {
					prv->d.cons.cdr = create_cons(create_cons(create_symbol("UNQUOTE-SPLICING"), create_cons(s->act->d.cons.cdr->d.cons.car, create_nil())), create_nil());
					s->act = prv;
					continue;
				} else return 10;
			}
		}

		s->act = s->act->d.cons.cdr;

		return 0;
	} else return 11;
}

/* test for symbol constituents */
int issymbol(int ch) {
	return isalnum(ch) || (ispunct(ch) && ch != ':' && ch != ';' && ch != '#' && ch != '`' && ch != ',' && ch != '.'
			&& ch != '\'' && ch != '(' && ch != ')');
}

/* read a character */
s_obj *read_ch(read_state *s, int ch)
{
	s_obj *lst;
	char *c;
	int code;

	if(s->ahead == 1) {
		/* this happens when the current character is to be reconsidered
		 * in a different state (thus the location should stay the same) */
		s->ahead = 0;
	} else {
		if(ch == '\n') {
			_stutter_parsed_line = ++ s->line;
			s->col = 1;
		} else s->col ++;
	}

	switch(s->state) {

		/* default state */
		case RS_NORM:

			if(ch == '(') {
				s->stack = create_cons(create_cons(s->out, s->act), s->stack);
				s->out = s->act = create_nil();
			} else if(ch == ')') {
				if(s->stack->type != ST_CONS)
					return create_error(SE_PARSE, "Unexpected ')'");
				lst = s->out;
				if(s->act->type == ST_CONS) {
					if(s->act->d.cons.car->type == ST_ERROR)
						return create_error_with_obj(SE_PARSE, "Unexpected ')' following list special",
								create_int(s->act->d.cons.car->d.error.code));
				}

				s->out = s->stack->d.cons.car->d.cons.car;
				s->act = s->stack->d.cons.car->d.cons.cdr;
				s->stack = s->stack->d.cons.cdr;
				if((code = read_append(s, lst)))
					return create_error_with_obj(SE_PARSE, "Dot confusion", create_int(code));

				if(s->stack->type == ST_NIL) {
					lst = s->out->d.cons.car;
					s->out = s->act = create_nil();
					return lst;
				}
			} else if(isspace(ch)) {
			} else if(ch == '#') {
				if(s->col == 1) s->state = RS_COMMENT; else s->state = RS_HASH;
				return NULL;
			} else if(ch == ';') {
				s->state = RS_COMMENT;
				return NULL;
			} else if(ch == '\"') {
				s->state = RS_STRING;
				return NULL;
			} else if(ch == EOF) {
				if(s->stack->type != ST_NIL)
					return create_error(SE_PARSE, "Unexpected EOF");
				return NULL;
			} else if(isdigit(ch)) {
				s->state = RS_INTEGER;
				token_appch(s, ch);
				return NULL;
			} else if(ch == '-') {
				s->state = RS_MINUS;
				token_appch(s, ch);
				return NULL;
			} else if(issymbol(ch)) {
				s->state = RS_SYMBOL;
				token_appch(s, ch);
				return NULL;
			} else if(ch == ':') {
				s->state = RS_COLON;
				return NULL;
			} else if(ch == '.') {
				s->state = RS_DOT;
				return NULL;
			} else if(ch == '\'') {
				if((code = read_append(s, create_error_with_obj(SE_UNDEF, NULL, s->act))))
					return create_error_with_obj(SE_PARSE, "Dot confusion", create_int(code));
				return NULL;
			} else if(ch == '`') {
				if((code = read_append(s, create_error_with_obj(SE_PARM, NULL, s->act))))
					return create_error_with_obj(SE_PARSE, "Dot confusion", create_int(code));
				return NULL;
			} else if(ch == ',') {
				s->state = RS_COMMA;
				return NULL;
			} else {
				return create_error(SE_PARSE, "Unexpected character");
			}

			break;

		/* after a - */
		case RS_MINUS:

			if(isdigit(ch)) {
				token_appch(s, ch);
				s->state = RS_INTEGER;
				return NULL;
			} else if(issymbol(ch)) {
				token_appch(s, ch);
				s->state = RS_SYMBOL;
				return NULL;
			} else if(ch == '.') {
				token_appch(s, ch);
				s->state = RS_NUM;
				return NULL;
			} else {
				if((code = read_append(s, create_symbol("-"))))
					return create_error_with_obj(SE_PARSE, "Dot confusion", create_int(code));
				resettoken(s);
				s->state = RS_NORM;

				s->ahead = 1;
				lst = read_ch(s, ch);
				if(lst) return lst;

				if(s->stack->type == ST_NIL) {
					lst = s->out->d.cons.car;
					s->out = s->act = create_nil();
					return lst;
				}

				return NULL;
			}

			break;

		/* reading an integer */
		case RS_INTEGER:

			if(isdigit(ch)) {
				token_appch(s, ch);
			} else if(ch == '.') {
				token_appch(s, ch);
				s->state = RS_NUM;
				return NULL;
			} else if(issymbol(ch)) {
				token_appch(s, ch);
				s->state = RS_SYMBOL;
				return NULL;
			} else {
				if((code = read_append(s, create_int(atoi(s->token)))))
					return create_error_with_obj(SE_PARSE, "Dot confusion", create_int(code));
				resettoken(s);
				s->state = RS_NORM;

				s->ahead = 1;
				lst = read_ch(s, ch);
				if(lst) return lst;

				if(s->stack->type == ST_NIL) {
					lst = s->out->d.cons.car;
					s->out = s->act = create_nil();
					return lst;
				}

				return NULL;
			}

			break;

		/* reading a float */
		case RS_NUM:

			if(isdigit(ch)) {
				token_appch(s, ch);
			} else {
				if((code = read_append(s, create_num(atof(s->token)))))
					return create_error_with_obj(SE_PARSE, "Dot confusion", create_int(code));
				resettoken(s);
				s->state = RS_NORM;

				s->ahead = 1;
				lst = read_ch(s, ch);
				if(lst) return lst;

				if(s->stack->type == ST_NIL) {
					lst = s->out->d.cons.car;
					s->out = s->act = create_nil();
					return lst;
				}

				return NULL;
			}

			break;

		/* after a : */
		case RS_COLON:
			if(issymbol(ch)) {
				token_appch(s, ch);
			} else {
				if(!s->token)
					return create_error(SE_PARSE, "Invalid ':'");
				for(c = s->token; *c; c++)
					if(!isdigit(*c)) break;
				if(!*c)
					return create_error(SE_PARSE, "Invalid ':'");
				lst = create_symbol(s->token);
				lst->d.symbol.colon = 1;
				if((code = read_append(s, lst)))
					return create_error_with_obj(SE_PARSE, "Dot confusion", create_int(code));
				resettoken(s);
				s->state = RS_NORM;

				s->ahead = 1;
				lst = read_ch(s, ch);
				if(lst) return lst;

				if(s->stack->type == ST_NIL) {
					lst = s->out->d.cons.car;
					s->out = s->act = create_nil();
					return lst;
				}

				return NULL;
			}

			break;

		/* reading a symbol */
		case RS_SYMBOL:

			if(issymbol(ch)) {
				token_appch(s, ch);
			} else {
				if((code = read_append(s, create_symbol(s->token))))
					return create_error_with_obj(SE_PARSE, "Dot confusion", create_int(code));
				resettoken(s);
				s->state = RS_NORM;

				s->ahead = 1;
				lst = read_ch(s, ch);
				if(lst) return lst;

				if(s->stack->type == ST_NIL) {
					lst = s->out->d.cons.car;
					s->out = s->act = create_nil();
					return lst;
				}

				return NULL;
			}

			break;

		/* after #\ */
		case RS_CHAR_CONST:
			if((code = read_append(s, create_char(ch))))
				return create_error_with_obj(SE_PARSE, "Dot confusion", create_int(code));
			s->state = RS_NORM;

			if(s->stack->type == ST_NIL) {
				lst = s->out->d.cons.car;
				s->out = s->act = create_nil();
				return lst;
			}

			return NULL;

		/* after a # */
		case RS_HASH:
			if(ch == '\\') {
				s->state = RS_CHAR_CONST;
				return NULL;
			} else return create_error(SE_PARSE, "Invalid stuff after #");

		/* filename-reseting comment (;%#) */
		case RS_COMMENT_FILE:
			if(ch != '\n' && isspace(ch) && !s->token) break;
			if(ch == ' ' || ch == ';' || ch == '\n') {
				if(s->token) {
					mark_miscfree(_stutter_file = strdup(s->token));
					s->line = 0;
					resettoken(s);
				}
				s->state = (ch == '\n' ? RS_NORM : RS_COMMENT_BODY);
				return NULL;
			} else token_appch(s, ch);
			break;

		/* after a ;% */
		case RS_COMMENT_1:
			if(ch == '#') {
				s->state = RS_COMMENT_FILE;
				return NULL;
			}

			if(ch == '\n') {
				s->state = RS_NORM;
				return NULL;
			}

			s->state = RS_COMMENT_BODY;
			return NULL;

		/* inside a comment */
		case RS_COMMENT_BODY:
			if(ch == '\n') {
				s->state = RS_NORM;
				return NULL;
			}
			break;

		/* beginning a comment */
		case RS_COMMENT:
			if(ch == ';') break;
			if(ch == '\n') {
				s->state = RS_NORM;
				return NULL;
			}
			if(isspace(ch)) break;
			if(ch == '%') {
				s->state = RS_COMMENT_1;
				return NULL;
			}
			s->state = RS_COMMENT_BODY;
			return NULL;

		/* reading a string */
		case RS_STRING:

			if(ch == '\\') {
				s->state = RS_STRING_ESCAPE;
				return NULL;
			} else if(ch == '\"') {
				if((code = read_append(s, create_string(s->token ? s->token : ""))))
					return create_error_with_obj(SE_PARSE, "Dot confusion", create_int(code));
				resettoken(s);
				s->state = RS_NORM;

				if(s->stack->type == ST_NIL) {
					lst = s->out->d.cons.car;
					s->out = s->act = create_nil();
					return lst;
				}

				return NULL;
			} else if(ch == EOF) {
				return create_error(SE_PARSE, "Unexpected EOF");
			}

			token_appch(s, ch);
			break;

		/* after a \ while reading a string */
		case RS_STRING_ESCAPE:

			if(ch == EOF)
				return create_error(SE_PARSE, "Unexpected EOF");

			switch(ch) {
				case 'a': ch = '\a'; break;
				case 'b': ch = '\b'; break;
				case 'f': ch = '\f'; break;
				case 'n': ch = '\n'; break;
				case 'r': ch = '\r'; break;
				case 't': ch = '\t'; break;
				case 'v': ch = '\v'; break;
				case 'x':
					/* supporting \xHH style character input */
					s->state = RS_STRING_ESCAPE_HEX;
					return NULL;
				default: ;
			}

			token_appch(s, ch);
			s->state = RS_STRING;
			return NULL;

		/* \xHH character input, first H */
		case RS_STRING_ESCAPE_HEX:

			if(ch == EOF)
				return create_error(SE_PARSE, "Unexpected EOF");

			if(ch >= '0' && ch <= '9')
				s->ch = (ch - '0') << 4;
			else if(ch >= 'a' && ch <= 'f')
				s->ch = (ch - 'a' + 10) << 4;
			else if(ch >= 'A' && ch <= 'F')
				s->ch = (ch - 'A' + 10) << 4;
			else
				return create_error(SE_PARSE, "Invalid first H in \\xHH escaped character");

			s->state = RS_STRING_ESCAPE_HEX_2;
			return NULL;

		/* \xHH character input, second H */
		case RS_STRING_ESCAPE_HEX_2:

			if(ch == EOF)
				return create_error(SE_PARSE, "Unexpected EOF");

			if(ch >= '0' && ch <= '9')
				s->ch += (ch - '0');
			else if(ch >= 'a' && ch <= 'f')
				s->ch += (ch - 'a' + 10);
			else if(ch >= 'A' && ch <= 'F')
				s->ch += (ch - 'A' + 10);
			else
				return create_error(SE_PARSE, "Invalid second H in \\xHH escaped character");

			token_appch(s, s->ch);
			s->state = RS_STRING;
			return NULL;

		/* after a . */
		case RS_DOT:

			if(isdigit(ch)) {
				token_appch(s, '.');
				token_appch(s, ch);
				s->state = RS_NUM;
				return NULL;
			} else {
				s->state = RS_NORM;
				if((code = read_append(s, create_error_with_obj(SE_WTF, NULL, s->act))))
					return create_error_with_obj(SE_PARSE, "Dot confusion", create_int(code));
	
				s->ahead = 1;
				return read_ch(s, ch);
			}

			break;

		/* after a , */
		case RS_COMMA:

			if(ch == '@') {
				s->state = RS_NORM;
				if((code = read_append(s, create_error_with_obj(SE_TYPE, NULL, s->act))))
					return create_error_with_obj(SE_PARSE, "Dot confusion", create_int(code));
				return NULL;
			} else if(ch == EOF) {
				return create_error(SE_PARSE, "Unexpected EOF");
			} else {
				s->state = RS_NORM;
				if((code = read_append(s, create_error_with_obj(SE_WFUNC, NULL, s->act))))
					return create_error_with_obj(SE_PARSE, "Dot confusion", create_int(code));
				s->ahead = 1;
				return read_ch(s, ch);
			}

			break;

		default:
			return create_error(SE_PARSE, "Undefined state");

	}

	return NULL;
}

/* create object from string */
s_obj *tokenstr2obj(char *str, const int type)
{
	switch(type) {
		case ST_SYMBOL:
			return create_symbol(str);
		case ST_INT:
			if(!strcmp(str, "-"))
				return create_symbol("-");
			else
				return create_int(atoi(str));
		case ST_NUM:
			return create_num(atof(str));
		default:
			return create_error(SE_TYPE, "tokenstr2obj: invalid type");
	}
}

