#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <stdarg.h>
#include <errno.h>
#include "types.h"
#include "memory.h"
#include "eval.h"
#include "builtins.h"

s_obj *cached_nil = NULL, *cached_t = NULL, *cached_eof = NULL;

#ifndef DICT_THRESHOLD
#define DICT_THRESHOLD 192
#endif

#ifndef DEFAULT_DICT_SIZE
#define DEFAULT_DICT_SIZE 8
#endif

/* size of varctx for initarg, slotname lookups in classes */
#ifndef CLASS_NAMES_SIZE
#define	CLASS_NAMES_SIZE 32
#endif

#ifndef DEFAULT_DISPATCH_SIZE
#define DEFAULT_DISPATCH_SIZE 8
#endif

/* fill threshold of method dispatch table (/256) */
#ifndef DISPATCH_THRESHOLD
#define DISPATCH_THRESHOLD 128
#endif

int default_dict_size = DEFAULT_DICT_SIZE;
int dict_threshold = DICT_THRESHOLD;

unsigned __int64 _stutter_next_serial = 1;

int next_typeid = 0;

typeinfo *typedb = NULL;
int typedb_size;

/******************************************** OBJECT CREATION ***/

/* creates and returns a new s_obj, zeroed */
s_obj *alloc_s_obj(void)
{
	s_obj *o;

	o = obj_respawn();
	if(!o) o = (s_obj *) xcalloc(1, sizeof(s_obj));

	o->file = _stutter_file;
	o->line = _stutter_parsed_line;
	o->serial = _stutter_next_serial ++;

	gc_record(o);

	return o;
}

/* creates a new s_obj without updating the records of the GC */
s_obj *alloc_ungc_obj(void)
{
	s_obj *o;
	o = (s_obj *) xcalloc(1, sizeof(s_obj));
	o->file = _stutter_file;
	o->line = _stutter_parsed_line;

	return o;
}

/* create a numeric atom */
s_obj *create_num(const double val)
{
	s_obj *o = alloc_s_obj();
	o->type = ST_NUM;
	o->d.num.value = val;
	return o;
}

/* create an integer atom */
s_obj *create_int(const int val)
{
#ifdef DISABLE_INTCACHE
	s_obj *o = alloc_s_obj();
	o->type = ST_INT;
	o->d.integer.value = val;
	return o;
#else
	s_obj *o;
	if(val >= REUSABLE_INT_LIMIT || val < 0) {
		o = alloc_s_obj();
		o->type = ST_INT;
		o->d.integer.value = val;
		return o;
	} else {
		return reusable_ints[val];
	}
#endif
}

/* create a character atom */
s_obj *create_char(const unsigned char ch)
{
#ifdef DISABLE_CHARCACHE
	s_obj *o = alloc_s_obj();
	o->type = ST_CHAR;
	o->d.ch.ch = val;
	return o;
#else
	return reusable_chars[ch];
#endif
}

/* create a symbolic atom */
s_obj *create_symbol(char *val)
{
	char *c;
	s_obj *o = alloc_s_obj();
	o->type = ST_SYMBOL;
	o->d.symbol.name = strdup(val);
	for(c=o->d.symbol.name;*c;c++) *c = toupper(*c);
	o->d.symbol.sdbm = sdbm(o->d.symbol.name);
	o->d.symbol.id = get_symbolid(o->d.symbol.name, o->d.symbol.sdbm);
	return o;
}

/* create a string */
s_obj *create_string(const char *val)
{
	s_obj *o = alloc_s_obj();
	o->type = ST_STRING;
	if(!(o->d.string.value = strdup(val))) {
		perror("create_string");
		exit(EXIT_MEMORY_FAIL);
	}
	return o;
}

/* create a string from n bytes */
s_obj *create_string_counted(const char *val, const int len)
{
	s_obj *o = alloc_s_obj();
	o->type = ST_STRING;
	o->d.string.value = (char *) xmalloc(len + 1);
	memcpy(o->d.string.value, val, len);
	o->d.string.value[len] = 0;
	return o;
}

/* create a string (doesn't strdup val) */
s_obj *create_string_allocd(char *val)
{
	s_obj *o = alloc_s_obj();
	o->type = ST_STRING;
	o->d.string.value = val;
	return o;
}

/* create a CONS */
s_obj *create_cons(s_obj *car, s_obj *cdr)
{
	s_obj *o = alloc_s_obj();
	o->type = ST_CONS;
	o->d.cons.car = car;
	o->d.cons.cdr = cdr;
	return o;
}

/* set last element of a list */
void list_setlast(s_obj *lst, s_obj *obj)
{
	while(lst->d.cons.cdr->type == ST_CONS)
		lst = lst->d.cons.cdr;
	lst->d.cons.cdr = obj;
}

/* create an error object */
s_obj *create_error(const int code, const char *message)
{
	s_obj *o = alloc_s_obj();
	o->type = ST_ERROR;
	o->d.error.code = code;
	if(message)
		o->d.error.msg = strdup(message);
	o->d.error.file = _stutter_lastcall_file;
	o->d.error.line = _stutter_lastcall_line;
	return o;
}

/* create an error object (with ->obj set) */
s_obj *create_error_with_obj(const int code, const char *message, s_obj *obj)
{
	s_obj *o = create_error(code, message);
	o->d.error.obj = obj;
	return o;
}

/* create a builtin object, with user data */
s_obj *create_builtin_with_data(builtin_handler handler, void *data)
{
	s_obj *o = alloc_s_obj();
	o->type = ST_BUILTIN;
	o->d.builtin.handler = handler;
	o->d.builtin.data = data;
	return o;
}

/* create a builtin object, with user data */
s_obj *create_sform_with_data(builtin_handler handler, void *data)
{
	s_obj *o = alloc_s_obj();
	o->type = ST_BUILTIN;
	o->d.builtin.handler = handler;
	o->d.builtin.data = data;
    o->d.builtin.sform = 1;
	return o;
}

/* create a builtin object */
s_obj *create_builtin(builtin_handler handler)
{
	s_obj *o = alloc_s_obj();
	o->type = ST_BUILTIN;
	o->d.builtin.handler = handler;
	return o;
}

/* create a builtin object */
s_obj *create_sform(builtin_handler handler)
{
	s_obj *o = alloc_s_obj();
	o->type = ST_BUILTIN;
	o->d.builtin.handler = handler;
    o->d.builtin.sform = 1;
	return o;
}

/* create an EOF object */
s_obj *create_eof(void)
{
	s_obj *o;

	if(!cached_eof) {
		o = (s_obj *) xcalloc(1, sizeof(s_obj));
		o->file = _stutter_file;
		o->line = _stutter_parsed_line;
		o->type = ST_EOF;
		cached_eof = o;
		mark_miscfree(o);
		return o;
	} else return cached_eof;
}

/* create a T object */
s_obj *create_t(void)
{
	s_obj *o;
	if(!cached_t) {
		o = (s_obj *) xcalloc(1, sizeof(s_obj));
		o->file = _stutter_file;
		o->line = _stutter_parsed_line;
		o->type = ST_T;
		cached_t = o;
		mark_miscfree(o);
		return o;
	} else return cached_t;
}

/* create a NIL object */
s_obj *create_nil(void)
{
	s_obj *o;
	if(!cached_nil) {
		o = (s_obj *) xcalloc(1, sizeof(s_obj));
		o->file = _stutter_file;
		o->line = _stutter_parsed_line;
		o->type = ST_NIL;
		cached_nil = o;
		mark_miscfree(o);
		return o;
	} else return cached_nil;
}

s_obj *create_new_nil(void)
{
	s_obj *o = alloc_s_obj();
	o->type = ST_NIL;
	return o;
}

s_obj *create_new_t(void)
{
	s_obj *o = alloc_s_obj();
	o->type = ST_T;
	return o;}

/* create a function */
s_obj *create_func(s_obj *prm, s_obj *exp)
{
	s_obj *o = alloc_s_obj();
	o->type = ST_FUNC;
	o->d.func.parm = prm;
	o->d.func.exp = exp;
	return o;
}

/* create a call deferral */
s_obj *create_defer(s_obj *func, s_obj *args, s_obj *obj)
{
	s_obj *o = alloc_s_obj();
	o->type = ST_DEFER;
	o->d.defer.func = func;
	o->d.defer.args = args;
	o->d.defer.obj = obj;
	return o;
}

/* create an array */
s_obj *create_array(const int dim, const int *size)
{
	int i, a = 1;
	s_obj *o = alloc_s_obj();
	o->type = ST_ARRAY;
	o->d.array.dim = dim;

	for(i=0;i<dim; i++)
		a *= (o->d.array.size[i] = size[i]);
	o->d.array.data = (s_obj **) malloc(a * sizeof(s_obj *));
	if(!o->d.array.data)
		return create_error(SE_MEM, "Insufficient memory for array allocation");
	for(i=0; i<a; i++)
		o->d.array.data[i] = create_nil();
	return o;
}

/* create a dictionary */
s_obj *create_dict()
{
	s_obj *o = alloc_s_obj();
	o->type = ST_DICT;
	o->d.dict.size = default_dict_size;
	o->d.dict.mask = o->d.dict.size - 1;
	o->d.dict.threshold = (o->d.dict.size * dict_threshold) >> 8;
	o->d.dict.data = (dictentry *) xcalloc(o->d.dict.size, sizeof(dictentry));
	return o;
}

/* create a stream */
s_obj *create_stream(const int type)
{
	s_obj *o = alloc_s_obj();
	o->type = ST_STREAM;
	o->d.stream.type = type;
	o->d.stream.unread = 256;
	o->d.stream.line = 1;
	return o;
}

/* create encapsulated varctx */
s_obj *create_obj_varctx(varctx *ctx)
{
	s_obj *o = alloc_s_obj();
	o->type = ST_VARCTX;
	o->d.varctx.ctx = ctx;
	return o;
}

/* create cookie */
s_obj *create_cookie(const int type, void *ptr)
{
	s_obj *o = alloc_s_obj();
	o->type = ST_COOKIE;
	o->d.cookie.type = type;
	o->d.cookie.ptr = ptr;
	return o;
}

/* create class */
s_obj *create_class()
{
	s_obj *o = alloc_s_obj();
	o->type = ST_CLASS;
	o->d.clos_class.initarg = varctx_create(NULL, CLASS_NAMES_SIZE);
	o->d.clos_class.slotname = varctx_create(NULL, CLASS_NAMES_SIZE);
	o->d.clos_class.parents = create_nil();
	o->d.clos_class.children = create_nil();
	o->d.clos_class.slots = create_nil();
	o->d.clos_class.prec = create_nil();
	o->d.clos_class.ancestry = create_nil();

	return o;
}

/* create slot */
s_obj *create_slot()
{
	s_obj *o = alloc_s_obj();
	o->type = ST_SLOT;
	o->d.clos_slot.initform = create_nil();
	return o;
}

/* create object */
s_obj *create_object(s_obj *type)
{
	s_obj *o = alloc_s_obj();
	o->type = ST_OBJECT;
	o->d.clos_object.type = type;
	o->d.clos_object.data_size = type->d.clos_class.instdata;
	o->d.clos_object.data = (s_obj **) xcalloc(o->d.clos_object.data_size, sizeof(s_obj *));
	return o;
}

/* create generic */
s_obj *create_generic()
{
	s_obj *o = alloc_s_obj();
	o->type = ST_GENERIC;
	o->d.clos_generic.dispatch = (dispatch_entry *) xcalloc(DEFAULT_DISPATCH_SIZE, sizeof(dispatch_entry));
	o->d.clos_generic.dispatch_size = DEFAULT_DISPATCH_SIZE;
	o->d.clos_generic.dispatch_fill = 0;
	o->d.clos_generic.dispatch_limit = (DEFAULT_DISPATCH_SIZE * DISPATCH_THRESHOLD) >> 8;
	o->d.clos_generic.m_primary = create_nil();
	o->d.clos_generic.m_before = create_nil();
	o->d.clos_generic.m_after = create_nil();
	o->d.clos_generic.m_around = create_nil();
	return o;
}

/* create method */
s_obj *create_method()
{
	s_obj *o = alloc_s_obj();
	o->type = ST_METHOD;
	return o;
}

/* create dispatch state */
s_obj *create_dispatch()
{
	s_obj *o = alloc_s_obj();
	o->type = ST_DISPATCH;
	o->d.clos_dispatch.argeval = (s_obj **) xcalloc(FUNC_MAXPARMS + 1, sizeof(s_obj *));
	return o;
}

/* create a varref */
s_obj *create_varref(int offset)
{
	s_obj *o = alloc_s_obj();
	o->type = ST_VARREF;
	o->d.varref.offset = offset;
	return o;
}

/*********************************** STREAMS ***/

/* print array contents */
s_obj * printarray(varctx *ctx, const s_obj *ar, s_obj **data, int dim)
{
	int i, ds;
	s_obj *o;
	ds = 1;
	for(i=0;i<dim;i++)
		ds *= ar->d.array.size[i];
	o = stream_stdout_write(ctx, '<');
	if(o->type == ST_ERROR) return o;

	if(dim < ar->d.array.dim-1) {
		for(i=0;i<ar->d.array.size[dim];i++) {
			if(i) {
				o = stream_stdout_write(ctx, ' ');
				if(o->type == ST_ERROR) return o;
			}
			o = printarray(ctx, ar, data, dim+1);
			if(o->type == ST_ERROR) return o;
			data += ds;
		}
	} else {
		for(i=0;i<ar->d.array.size[dim];i++) {
			if(i) {
				o = stream_stdout_write(ctx, ' ');
				if(o->type == ST_ERROR) return o;
			}
			o = printobj(ctx, *data);
			if(o->type == ST_ERROR) return o;
			data += ds;
		}
	}

	o = stream_stdout_write(ctx, '>');
	if(o->type == ST_ERROR) return o;

	return create_nil();
}

/* write a character to a stream */
s_obj *stream_write_char(varctx *ctx, s_obj *stream, int ch)
{
	if(stream->type != ST_STREAM)
		return create_error(SE_TYPE, "Type mismatch in stream_write_char");

	if(ch == '\n') stream->d.stream.line ++;

	switch(stream->d.stream.type) {
		case STREAM_FP:
			if(!stream->d.stream.s.fp)
				return create_error(SE_IO, "I/O error (stream closed)");
			if(fputc(ch, stream->d.stream.s.fp) == EOF)
				return create_error(SE_IO, "I/O error during stream_write_char");
			break;
		case STREAM_FUNC:
			return funcexec(ctx, NULL, stream->d.stream.s.func, create_cons(create_char(ch), create_nil()), 0, 0);
		case STREAM_C:
			if(!stream->d.stream.s.c.writer)
				return create_error(SE_IO, "External stream not writable");
			if(stream->d.stream.s.c.writer(ch, &stream->d.stream.s.c.data, stream->d.stream.s.c.type))
				return create_error_with_obj(SE_IO, "I/O error in external writer function", create_string(strerror(errno)));
			break;
		default:
			;
	}

	return create_nil();
}

/* printf to the standard output stream */
s_obj *stream_printf(varctx *ctx, const char *format, ...)
{
	va_list ap;
	char buf[128];
	char *dst;
	int r, l, f = 0;
	char *c;
	s_obj *o;

	va_start(ap, format);
	r = vsnprintf(buf, 128, format, ap);
	if(r >= 0 && r < 128) {
		va_end(ap);
		dst = buf;
	} else {
		va_end(ap);
		f = 1; // denote |dst| needs to be |free|d
		l = 256; // initial buffer size
		while(1) {
			dst = (char *) xmalloc(l);
			va_start(ap, format);
			r = vsnprintf(dst, l, format, ap);
			va_end(ap);
			if(r >= 0 && r < l)
				break;
			free(dst);
			l *= 2;
		}
	}

	o = NULL;
	for(c = dst; *c; c++) {
		o = stream_stdout_write(ctx, *c);
		if(o->type == ST_ERROR) {
			if(f) free(dst);
			return o;
		}
	}

	if(f) free(dst);

	return o ? o : create_nil();
}

/* unread a character into a stream */
s_obj *stream_unread_char(s_obj *stream, const unsigned char ch)
{
	if(stream->type != ST_STREAM)
		return create_error(SE_TYPE, "Type mismatch in stream_unread_char");
	if(stream->d.stream.unread != 256)
		return create_error(SE_IO, "Can't unread two characters without reading one in the middle");

	stream->d.stream.unread = ch;
	return stream;
}

/* read a character form a stream */
s_obj *stream_read_char(varctx *ctx, s_obj *stream)
{
	s_obj *obj;
	int ch;

	if(stream->type != ST_STREAM)
		return create_error(SE_TYPE, "Type mismatch in stream_read_char");

	if(stream->d.stream.unread != 256) {
		/* there is an unread character, let's give it back. */
		obj = create_char(stream->d.stream.unread);
		stream->d.stream.unread = 256;
		return obj;
	}

	switch(stream->d.stream.type) {
		case STREAM_FP:
			if(!stream->d.stream.s.fp)
				return create_error(SE_IO, "I/O error (stream closed)");
			ch = fgetc(stream->d.stream.s.fp);
			if(ch == '\n') _stutter_parsed_line = stream->d.stream.line ++;
			if(ch == EOF)
				return create_eof();
			return create_char(ch);
		case STREAM_FUNC:
			obj = funcexec(ctx, NULL, stream->d.stream.s.func, create_nil(), 0, 0);
			switch(obj->type) {
				case ST_STRING:
					if(obj->d.string.value[0] == '\n') _stutter_parsed_line = stream->d.stream.line ++;
					return create_char(obj->d.string.value[0]);
				case ST_CHAR:
					if(obj->d.ch.ch == '\n') _stutter_parsed_line = stream->d.stream.line ++;
				case ST_ERROR:
				case ST_EOF:
					return obj;
				default:
					return create_error(SE_TYPE, "Type mismatch in stream_read_char");
			}
		case STREAM_C:
			if(!stream->d.stream.s.c.reader)
				return create_error(SE_IO, "External stream not readable");
			ch = stream->d.stream.s.c.reader(&stream->d.stream.s.c.data, stream->d.stream.s.c.type);
			if(ch == '\n') _stutter_parsed_line = stream->d.stream.line ++;
			if(ch == EOF)
				return create_eof();
			return create_char(ch);
		default:
			return create_error(SE_TYPE, "Invalid stream type in stream_read_char");
	}

	return create_nil();
}

/* read a character from *input-stream* */
s_obj *stream_stdin_read(varctx *ctx)
{
	static unsigned long input_stream_sdbm = 0;
	static unsigned long input_stream_id = 0;
	s_var *v;

	if(!input_stream_id) {
		input_stream_sdbm = sdbm("*INPUT-STREAM*");
		input_stream_id = get_symbolid("*INPUT-STREAM*", input_stream_sdbm);
	}

	v = varctx_get_s(ctx, input_stream_sdbm, input_stream_id);
	if(!v)
		return create_error(SE_WTF, "*INPUT-STREAM* undefined");

	return stream_read_char(ctx, v->value);
}

/* unread a character into *input-stream* */
s_obj *stream_stdin_unread(varctx *ctx, const unsigned char ch)
{
	static unsigned long input_stream_sdbm = 0;
	static unsigned long input_stream_id = 0;
	s_var *v;

	if(!input_stream_id) {
		input_stream_sdbm = sdbm("*INPUT-STREAM*");
		input_stream_id = get_symbolid("*INPUT-STREAM*", input_stream_sdbm);
	}

	v = varctx_get_s(ctx, input_stream_sdbm, input_stream_id);
	if(!v)
		return create_error(SE_WTF, "*INPUT-STREAM* undefined");

	return stream_unread_char(v->value, ch);
}

/* write a character to *output-stream* */
s_obj *stream_stdout_write(varctx *ctx, int ch)
{
	static unsigned long output_stream_sdbm = 0;
	static unsigned long output_stream_id = 0;
	s_var *v;

	if(!output_stream_id) {
		output_stream_sdbm = sdbm("*OUTPUT-STREAM*");
		output_stream_id = get_symbolid("*OUTPUT-STREAM*", output_stream_sdbm);
	}

	v = varctx_get_s(ctx, output_stream_sdbm, output_stream_id);
	if(!v)
		return create_error(SE_WTF, "*OUTPUT-STREAM* undefined");

	return stream_write_char(ctx, v->value, ch);
}

/* print description/value of an object */
s_obj * printobj(varctx *ctx, const s_obj *obj)
{
	s_obj *o, *oi;
	char *c;
	int i;
	char buf[32];

	if(!obj) {
		o = stream_printf(ctx, "NULL");
		return o;
	}

	switch(obj->type) {
		/* string */
		case ST_STRING:
			o = stream_stdout_write(ctx, '"');
			if(o->type == ST_ERROR) return o;
			for(c = obj->d.string.value; *c; c++) {
				if(*c == '\\' || *c == '\"') {
					o = stream_stdout_write(ctx, '\\');
					if(o->type == ST_ERROR) return o;
				}
				o = stream_stdout_write(ctx, *c);
				if(o->type == ST_ERROR) return o;
			}
			o = stream_stdout_write(ctx, '"');
			return o;

		/* double */
		case ST_NUM:
			sprintf(buf, "%.12f", obj->d.num.value);
			c = buf + strlen(buf) - 1;
			for(;;) {
				if(*c == '.') {
					*++c = '0';
					break;
				} else if(*c == '0') {
					*c-- = 0;
				} else break;
			}
			return stream_printf(ctx, buf);

		/* integer */
		case ST_INT:
			return stream_printf(ctx, "%d", obj->d.integer.value);

		/* character */
		case ST_CHAR:
			return stream_printf(ctx, "#\\%c", obj->d.ch.ch);

		/* symbol */
		case ST_SYMBOL:
			if(obj->d.symbol.colon) {
				o = stream_stdout_write(ctx, ':');
				if(o->type == ST_ERROR) return o;
			}
			return stream_printf(ctx, "%s", obj->d.symbol.name);

		/* array */
		case ST_ARRAY:
			o = stream_printf(ctx, "#A");
			if(o->type == ST_ERROR) return o;
			return printarray(ctx, obj, obj->d.array.data, 0);

		/* dictionary */
		case ST_DICT:
			o = stream_printf(ctx, "#D<");
			if(o->type == ST_ERROR) return o;
			for(i=0; i<obj->d.dict.size; i++)
				if(obj->d.dict.data[i].name) {
					o = stream_stdout_write(ctx, ' ');
					if(o->type == ST_ERROR) return o;
					o = printobj(ctx, obj->d.dict.data[i].name);
					if(o->type == ST_ERROR) return o;
					o = stream_stdout_write(ctx, ':');
					if(o->type == ST_ERROR) return o;
					o = printobj(ctx, obj->d.dict.data[i].value);
					if(o->type == ST_ERROR) return o;
				}
			return stream_printf(ctx, " >");

		/* NIL */
		case ST_NIL:
			return stream_printf(ctx, "nil");

		/* T */
		case ST_T:
			return stream_printf(ctx, "T");

		/* error */
		case ST_ERROR:
			if(obj->d.error.code <= SE_LAST)
				o = stream_printf(ctx, "ERROR %s: %s at %s:%d", errornames[obj->d.error.code], obj->d.error.msg, obj->d.error.file, obj->d.error.line);
			else
				o = stream_printf(ctx, "ERROR %d: %s at %s:%d", obj->d.error.code, obj->d.error.msg, obj->d.error.file, obj->d.error.line);
			if(o->type == ST_ERROR) return o;

			if(obj->d.error.obj) {
				o = stream_printf(ctx, " [");
				if(o->type == ST_ERROR) return o;
				o = printobj(ctx, obj->d.error.obj);
				if(o->type == ST_ERROR) return o;
				o = stream_printf(ctx, "]");
				if(o->type == ST_ERROR) return o;
			}

			for(oi = obj->d.error.trace; oi && oi->type == ST_CONS; oi=oi->d.cons.cdr) {
				if(oi->d.cons.car->type == ST_ERROR) {
					o = stream_printf(ctx, "\n  [");
					if(o->type == ST_ERROR) return o;
					if(oi->d.cons.car->d.error.obj &&
							oi->d.cons.car->d.error.obj->type == ST_FUNC &&
							oi->d.cons.car->d.error.obj->d.func.symbol) {	
						o = stream_printf(ctx, "%s ", oi->d.cons.car->d.error.obj->d.func.symbol);
						if(o->type == ST_ERROR) return o;
					}

					o = stream_printf(ctx, "%s:%d]",
							oi->d.cons.car->d.error.file,
							oi->d.cons.car->d.error.line);
					if(o->type == ST_ERROR) return o;
				}
			}
			break;

		/* builtin function */
		case ST_BUILTIN:
			if(obj->d.builtin.symbol)
				return stream_printf(ctx, "#b:%s", obj->d.builtin.symbol);
			else
				return stream_printf(ctx, "#b:%p", obj->d.builtin.handler);

		/* stream */
		case ST_STREAM:
			return stream_printf(ctx, "#STREAM");

		/* function */
		case ST_FUNC:
			if(obj->d.func.symbol)
				o = stream_printf(ctx, "#%c:%s", obj->d.func.macro ? 'm' : 'l', obj->d.func.symbol);
			else
				o = stream_printf(ctx, obj->d.func.macro ? "macro" : "lambda");
			if(o->type == ST_ERROR) return o;
			if(obj->d.func.parm->type == ST_CONS)
				o = printobj(ctx, obj->d.func.parm);
			else
				o = stream_printf(ctx, "()");
			if(o->type == ST_ERROR) return o;
			if(obj->d.func.docstr) {
				o = stream_stdout_write(ctx, ' ');
				if(o->type == ST_ERROR) return o;
				o = printobj(ctx, obj->d.func.docstr);
			}
			return o;

		/* deferred execution (doesn't happen) */
		case ST_DEFER:
			return stream_printf(ctx, "#DEFER");

		/* encapsulated varctx */
		case ST_VARCTX:
			return stream_printf(ctx, "#VARCTX:%p", obj->d.varctx.ctx);

		/* cookie */
		case ST_COOKIE:
			return stream_printf(ctx, "#COOKIE:%d", obj->d.cookie.type);

		/**** CLOS ****/

		/* class */
		case ST_CLASS:
			return stream_printf(ctx, "#CLASS");

		/* slot */
		case ST_SLOT:
			return stream_printf(ctx, "#SLOT");

		/* instance */
		case ST_OBJECT:
			return stream_printf(ctx, "#OBJECT<%s>", obj->d.clos_object.type->d.clos_class.name->d.symbol.name);

		/* method */
		case ST_METHOD:
			o = stream_printf(ctx, "\n  #METHOD<%p>[code=%p, args={", obj, obj->d.clos_method.code);
			if(o->type == ST_ERROR) return o;

			for(i=0; i<obj->d.clos_method.argn; i++) {
				o = stream_stdout_write(ctx, ' ');
				if(o->type == ST_ERROR) return o;

				o = printobj(ctx, obj->d.clos_method.args[i].name);
				if(o->type == ST_ERROR) return o;
				o = stream_printf(ctx, ",%d,", obj->d.clos_method.args[i].type);
				if(o->type == ST_ERROR) return o;
				o = printobj(ctx, obj->d.clos_method.args[i].value);
				if(o->type == ST_ERROR) return o;
			}

			o = stream_printf(ctx, " }; miscargs=");
			if(o->type == ST_ERROR) return o;

			o = printobj(ctx, obj->d.clos_method.miscargs);
			if(o->type == ST_ERROR) return o;

			return stream_printf(ctx, "]");

		/* dispatch info */
		case ST_DISPATCH:
			return stream_printf(ctx, "#DISPATCH");

		/* generic */
		case ST_GENERIC:
			o = stream_printf(ctx, "#GENERIC[\nprimary=");
			if(o->type == ST_ERROR) return o;

			o = printobj(ctx, obj->d.clos_generic.m_primary);
			if(o->type == ST_ERROR) return o;

			o = stream_printf(ctx, "\nbefore=");
			if(o->type == ST_ERROR) return o;

			o = printobj(ctx, obj->d.clos_generic.m_before);
			if(o->type == ST_ERROR) return o;

			o = stream_printf(ctx, "\nafter=");
			if(o->type == ST_ERROR) return o;

			o = printobj(ctx, obj->d.clos_generic.m_after);
			if(o->type == ST_ERROR) return o;

			o = stream_printf(ctx, "\naround=");
			if(o->type == ST_ERROR) return o;

			o = printobj(ctx, obj->d.clos_generic.m_around);
			if(o->type == ST_ERROR) return o;

			return stream_printf(ctx, "]");

		/**** /CLOS ****/

		/* EOF */
		case ST_EOF:
			return stream_printf(ctx, "EOF");

		/* cons */
		case ST_CONS:
			/* try to print 'A instead of (b:something A) */
			if(obj->d.cons.car->type == ST_BUILTIN &&
					obj->d.cons.car->d.builtin.printprefix != 0 &&
					obj->d.cons.cdr->type == ST_CONS &&
					obj->d.cons.cdr->d.cons.cdr->type == ST_NIL) {
				o = stream_stdout_write(ctx, obj->d.cons.car->d.builtin.printprefix);
				if(o->type == ST_ERROR) return o;
				return printobj(ctx, obj->d.cons.cdr->d.cons.car);
			} else {
				/* just a standard cons.. */
				o = stream_stdout_write(ctx, '(');
				if(o->type == ST_ERROR) return o;
				o = printobj(ctx, obj->d.cons.car);
				if(o->type == ST_ERROR) return o;

				if(obj->d.cons.cdr != NULL && obj->d.cons.cdr->type != ST_NIL) {
					if(obj->d.cons.cdr->type != ST_CONS) {
						o = stream_printf(ctx, " . ");
						if(o->type == ST_ERROR) return o;
						o = printobj(ctx, obj->d.cons.cdr);
						if(o->type == ST_ERROR) return o;
					} else {
						for(oi = obj->d.cons.cdr; oi && oi->type != ST_NIL;) {
							o = stream_stdout_write(ctx, ' ');
							if(o->type == ST_ERROR) return o;
							o = printobj(ctx, oi->d.cons.car);
							if(o->type == ST_ERROR) return o;

							oi = oi->d.cons.cdr;
							if(oi && oi->type != ST_CONS && oi->type != ST_NIL) {
								o = stream_printf(ctx, " . ");
								if(o->type == ST_ERROR) return o;
								o = printobj(ctx, oi);
								if(o->type == ST_ERROR) return o;
								break;
							}
						}
					}
				}

				return stream_stdout_write(ctx, ')');
			}
			break;

		/* dunno (doesn't happen) */
		default:
			return stream_printf(ctx, "?");
	}

	return create_nil();
}

/******************************************** UTILS ***/

/* returns new type id, registers random seed */
int gettypeid() {
	int id;

	if(!typedb) {
		typedb = (typeinfo *) xcalloc(128, sizeof(int));
		typedb_size = 128;
	} else {
		if(next_typeid == typedb_size) {
			typedb_size *= 2;
			typedb = (typeinfo *) xrealloc(typedb, typedb_size * sizeof(int));
		}
	}

	typedb[next_typeid].typeseed = hash((unsigned char *) &next_typeid, sizeof(int), 0);
	typedb[next_typeid].cl = NULL;
	id = next_typeid ++;

	return id;
}

/* get type id for symbol. -1 if no such thing */
int symbol_to_typeid(s_obj *sym)
{
	assert(sym->type == ST_SYMBOL);

	/* check if a class exists by that name */
	s_var *var = varctx_get_s(class_hash, sym->d.symbol.sdbm, sym->d.symbol.id);
	if(var) return var->value->d.clos_class.typeid; /* yep, return type id */

	/* check for atomic types
	 * typeid should be equal to the value of the type constant in the enum */

	if(!strcmp(sym->d.symbol.name, "NIL"))
		return ST_NIL;
	if(!strcmp(sym->d.symbol.name, "INT"))
		return ST_INT;
	if(!strcmp(sym->d.symbol.name, "NUM"))
		return ST_NUM;
	if(!strcmp(sym->d.symbol.name, "CHAR"))
		return ST_CHAR;
	if(!strcmp(sym->d.symbol.name, "T"))
		return ST_T;
	if(!strcmp(sym->d.symbol.name, "STRING"))
		return ST_STRING;
	if(!strcmp(sym->d.symbol.name, "COOKIE"))
		return ST_COOKIE;
	if(!strcmp(sym->d.symbol.name, "EOF"))
		return ST_EOF;
	if(!strcmp(sym->d.symbol.name, "BUILTIN"))
		return ST_BUILTIN;
	if(!strcmp(sym->d.symbol.name, "FUNC"))
		return ST_FUNC;
	if(!strcmp(sym->d.symbol.name, "ERROR"))
		return ST_ERROR;
	if(!strcmp(sym->d.symbol.name, "DEFER"))
		return ST_DEFER;
	if(!strcmp(sym->d.symbol.name, "ARRAY"))
		return ST_ARRAY;
	if(!strcmp(sym->d.symbol.name, "VARCTX"))
		return ST_VARCTX;
	if(!strcmp(sym->d.symbol.name, "DICT"))
		return ST_DICT;
	if(!strcmp(sym->d.symbol.name, "STREAM"))
		return ST_STREAM;
	if(!strcmp(sym->d.symbol.name, "CLASS"))
		return ST_CLASS;
	if(!strcmp(sym->d.symbol.name, "SLOT"))
		return ST_SLOT;
	if(!strcmp(sym->d.symbol.name, "OBJECT"))
		return ST_OBJECT;
	if(!strcmp(sym->d.symbol.name, "GENERIC"))
		return ST_GENERIC;
	if(!strcmp(sym->d.symbol.name, "METHOD"))
		return ST_METHOD;

	return -1;
}

/* return type id of object */
int object_typeid(s_obj *obj)
{
	/* if it's a CLOS object, return type id of class */
	if(obj->type == ST_OBJECT)
		return obj->d.clos_object.type->d.clos_class.typeid;

	/* otherwise, return stutter type */
	return obj->type;
}

/* get hash value for an object */
unsigned long hashfunc(const s_obj *obj) {
	switch(obj->type) {
		case ST_SYMBOL:
			return obj->d.symbol.sdbm;
		case ST_STRING:
			return sdbm(obj->d.string.value);
		case ST_INT:
			return hash((unsigned char *) &obj->d.integer.value, sizeof(obj->d.integer.value), 0);
		case ST_NUM:
			return hash((unsigned char *) &obj->d.num.value, sizeof(obj->d.num.value), 0);
		default:
			return (unsigned long) obj;
	}
}

/* copy object with inner data */
void deepcopy(s_obj *tgt, s_obj *obj) {
	int i, ds;

	memcpy(tgt, obj, sizeof(s_obj));
//	tgt->ref = NULL;

	switch(obj->type) {
		case ST_STRING:
			tgt->d.string.value = strdup(obj->d.string.value);
			break;
		case ST_SYMBOL:
			tgt->d.symbol.name = strdup(obj->d.symbol.name);
			break;
		case ST_ERROR:
			tgt->d.error.msg = strdup(obj->d.error.msg);
			break;
		case ST_ARRAY:
			ds = 1;
			for(i=0;i<obj->d.array.dim;i++)
				ds *= obj->d.array.size[i];
			tgt->d.array.data = (s_obj **) malloc(ds * sizeof(s_obj *));
			if(!tgt->d.array.data)
				tgt = create_error(SE_MEM, "Insufficient memory for deep copy");
			else {
				memcpy(tgt->d.array.data, obj->d.array.data, ds * sizeof(s_obj *));
			}
			break;
		case ST_DICT:
			tgt->d.dict.data = (dictentry *) xcalloc(obj->d.dict.size, sizeof(dictentry));
			memcpy(tgt->d.dict.data, obj->d.dict.data, obj->d.dict.size * sizeof(dictentry));
			break;

		case ST_OBJECT:
			tgt->d.clos_object.data = (s_obj **) xcalloc(obj->d.clos_object.data_size, sizeof(s_obj *));
			memcpy(tgt->d.clos_object.data, obj->d.clos_object.data, obj->d.clos_object.data_size * sizeof(s_obj *));
			break;

		default:
			;
	}
}

/* remove weakptr references */
inline void weak_unref(s_obj **o)
{
	weakptr *ptr, *n;
	if(!*o) return;
	for(ptr = (*o)->weak; ptr; ) {
		if(ptr->ptr == o) {
			n = ptr->next;
			free(ptr);
			ptr = n;
			continue;
		}

		ptr = ptr->next;
	}
}

/* free inner data of object */
void freedata(s_obj *obj) {
	int i;

	switch(obj->type) {
		case ST_DICT:
			free(obj->d.dict.data);
			break;
		case ST_ARRAY:
			free(obj->d.array.data);
			break;
		case ST_CLASS:
			varctx_free(obj->d.clos_class.initarg);
			varctx_free(obj->d.clos_class.slotname);
			free(obj->d.clos_class.allslots);
			free(obj->d.clos_class.data);
			break;
		case ST_OBJECT:
			free(obj->d.clos_object.data);
			break;
		case ST_METHOD:
			free(obj->d.clos_method.args);
			break;
		case ST_DISPATCH:
			free(obj->d.clos_dispatch.argeval);
			break;
		case ST_GENERIC:
			for(i=0; i<obj->d.clos_generic.dispatch_size; i++)
				free(obj->d.clos_generic.dispatch[i].signature);

			free(obj->d.clos_generic.dispatch);
			break;
		case ST_SYMBOL:
			free(obj->d.symbol.name);
			break;
		case ST_STRING:
			free(obj->d.string.value);
			break;
		case ST_VARCTX:
			varctx_free(obj->d.varctx.ctx);
			break;
		case ST_ERROR:
			if(obj->d.error.msg)
				free(obj->d.error.msg);
			break;
		default:
			;
	}
}

/* returns true if two objects are isoform (all fields are equal) */
int isoform(const s_obj *a, const s_obj *b)
{
	int i, s;
	if(a == NULL && b != NULL) return 0;
	if(a != NULL && b == NULL) return 0;
	if(a == NULL && b == NULL) return 1;

	if(a->type != b->type) return 0;
	switch(a->type) {
		case ST_INT:
			return a->d.integer.value == b->d.integer.value;
		case ST_NUM:
			return a->d.num.value == b->d.num.value;
		case ST_STRING:
			return !strcmp(a->d.string.value, b->d.string.value);
		case ST_CHAR:
			return a->d.ch.ch == b->d.ch.ch;
		case ST_SYMBOL:
			return a->d.symbol.id == b->d.symbol.id;
		case ST_NIL:
		case ST_T:
		case ST_EOF:
			return 1;
		case ST_CONS:
			return isoform(a->d.cons.car, b->d.cons.car) &&
				isoform(a->d.cons.cdr, b->d.cons.cdr);
		case ST_ARRAY:
			if(a->d.array.dim != b->d.array.dim) return 0;
			if(memcmp(a->d.array.size, b->d.array.size, a->d.array.dim * sizeof(int))) return 0;
			for(s=1,i=0;i<a->d.array.dim;i++)
				s *= a->d.array.size[i];
			for(i=0;i<s;i++)
				if(!isoform(a->d.array.data[i], b->d.array.data[i])) return 0;
			return 1;
		case ST_OBJECT:
			if(a->d.clos_object.type != b->d.clos_object.type) return 0;
			for(i=0; i<a->d.clos_object.data_size; i++) {
				if(!isoform(a->d.clos_object.data[i], b->d.clos_object.data[i])) return 0;
			}

			return 1;
		default:
			return a == b;
	}
}

/* initialize reusable object caches */
void objcache_init()
{
	int i;

#ifndef DISABLE_INTCACHE
	for(i=0; i<REUSABLE_INT_LIMIT; i++) {
		reusable_ints[i] = alloc_ungc_obj();
		reusable_ints[i]->type = ST_INT;
		reusable_ints[i]->d.integer.value = i;

		mark_miscfree((void *) reusable_ints[i]);
	}
#endif

#ifndef DISABLE_CHARCACHE
	for(i=0; i<256; i++) {
		reusable_chars[i] = alloc_ungc_obj();

		reusable_chars[i]->type = ST_CHAR;
		reusable_chars[i]->d.ch.ch = i;

		mark_miscfree((void *) reusable_chars[i]);
	}
#endif
}

/* add a weak reference place to object */
void weak_reference(s_obj *host, unsigned __int64 serial, s_obj **ptr)
{
	weakptr *wp;

//	fprintf(stderr, "weak_reference host=%p serial=%d ptr=%p obj=%p\n", host, (int)serial, ptr, *ptr);

	for(wp = (*ptr)->weak; wp; wp=wp->next)
		if(wp->ptr == ptr) {
			wp->host = host;
			wp->serial = serial;
			return;
		}

	wp = (weakptr *) xmalloc(sizeof(weakptr));
	wp->ptr = ptr;
	wp->host = host;
	wp->serial = serial;
	wp->next = (*ptr)->weak;
	(*ptr)->weak = wp;
}

/* check weakptrs, clean up invalid entries
 * returns number of entries removed */
int weak_check(s_obj *obj)
{
	weakptr *wp, *n, *pr;

	int c = 0;

	pr = NULL;

	for(wp = obj->weak; wp; ) {
		if(wp->host->serial != wp->serial ||
				(* wp->ptr) != obj) {

			if(pr) pr->next = wp->next; else obj->weak = wp->next;

			n = wp;
			wp = wp->next;
			free(n);
			c++;
			continue;
		}

		pr = wp;
		wp = wp->next;
	}

	return c;
}


/****************************************************** CLOS ***/

/* add tree branch to the ancestry of a class */
void build_class_ancestry(s_obj *root, s_obj *cl)
{
	s_obj *o, *iter, *last;

	/* add self to family tree, in order */
	if(cl->d.clos_class.ancestry->type == ST_NIL) {
		cl->d.clos_class.ancestry = create_cons(root, create_nil());
	} else {
		last = NULL;
		iter = cl->d.clos_class.ancestry;
		while(iter->type == ST_CONS && iter->d.cons.car->d.clos_class.prec_score < root->d.clos_class.prec_score) {
			last = iter;
			iter = iter->d.cons.cdr;
		}

		/* if prec_score is equal (same class), don't add again */
		if(iter->type == ST_CONS && iter->d.cons.car->d.clos_class.prec_score == root->d.clos_class.prec_score) {
			return;
		}

		if(last) {
			last->d.cons.cdr = create_cons(root, last->d.cons.cdr);
		} else {
			cl->d.clos_class.ancestry = create_cons(root, cl->d.clos_class.ancestry);
		}
	}

	/* recurse into parents */
	iter = root->d.clos_class.parents;
	while(nextparm(&o, &iter)) {
		build_class_ancestry(o, cl);
	}
}

/* add slots from class1 to slot assignments of class2
 * slots with the same name don't get overwritten
 * updates class2->instdata and data_size */
void slots_merge(s_obj *class1, s_obj *class2)
{
	s_obj *slot, *iter;
	int i;
	slotasg *asg;

	iter = class1->d.clos_class.slots;
	while(nextparm(&slot, &iter)) {
		/* scan class2's slot assignments for current slot name */
		for(i=0; i<class2->d.clos_class.allslotsnum; i++) {
			if(class2->d.clos_class.allslots[i].slot->d.clos_slot.name->d.symbol.id == slot->d.clos_slot.name->d.symbol.id) break;
		}

		/* no such slot found */
		if(i == class2->d.clos_class.allslotsnum) {
			class2->d.clos_class.allslots = (slotasg *) xrealloc(class2->d.clos_class.allslots, (class2->d.clos_class.allslotsnum + 1) * sizeof(slotasg));

			asg = class2->d.clos_class.allslots + (class2->d.clos_class.allslotsnum ++);

			asg->slot = slot;

			if(slot->d.clos_slot.alloc == CL_ALLOC_INSTANCE) {
				asg->stor = NULL;
				asg->num = (class2->d.clos_class.instdata ++);
			} else {
				asg->stor = class2;
				asg->num = (class2->d.clos_class.data_size ++);
			}
		}
	}
}

/* calculate and update internal data of class */
s_obj *class_calculate(s_obj *cl)
{
	s_obj *o, *iter, *iter2;
	int i;
	slotasg *asg;

	/* add self to parents' children */
	iter = cl->d.clos_class.parents;
	while(nextparm(&o, &iter)) {
		iter2 = o->d.clos_class.children;
		if(iter2->type == ST_NIL) {
			/* we are the first child */
			o->d.clos_class.children = create_cons(cl, create_nil());
		} else {
			/* append to list */
			while(iter2->d.cons.cdr->type == ST_CONS)
				iter2 = iter2->d.cons.cdr;

			iter2->d.cons.cdr = create_cons(cl, create_nil());
		}
	}

	/* rebuild class precedence data */
	build_precedence();

	/* build sorted family tree */
	iter = cl->d.clos_class.parents;
	while(nextparm(&o, &iter)) {
		build_class_ancestry(o, cl);
	}

	/* calculate slot assignments */
	slots_merge(cl, cl);
	iter = cl->d.clos_class.ancestry;
	while(nextparm(&o, &iter)) {
		slots_merge(o, cl);
	}

	/* fill in slotname->assignment varctxes */
	for(i=0, asg = cl->d.clos_class.allslots; i<cl->d.clos_class.allslotsnum; i++, asg++) {
		varctx_set_s(cl->d.clos_class.slotname, asg->slot->d.clos_slot.name->d.symbol.sdbm, asg->slot->d.clos_slot.name->d.symbol.id, create_int(i));

		if(asg->slot->d.clos_slot.initarg) {
			varctx_set_s(cl->d.clos_class.initarg, asg->slot->d.clos_slot.name->d.symbol.sdbm, asg->slot->d.clos_slot.name->d.symbol.id, create_int(i));
		}
	}

	/* allocate space for static data */
	cl->d.clos_class.data = (s_obj **) xcalloc(cl->d.clos_class.data_size, sizeof(s_obj *));

	return cl;
}

/* walk a branch of class precedence */
void build_precedence_branch(s_obj *root, s_obj **tail)
{
	s_obj *o, *sib, *iter;
	s_obj *iter2;
	int branch;

	/* add root to precedence */
	if(*tail) {
		(*tail)->d.cons.cdr = create_cons(root, create_nil());
		*tail = (*tail)->d.cons.cdr;
	} else {
		class_precedence = *tail = create_cons(root, create_nil());
	}
	root->d.clos_class.prec = *tail;

	/* walk parents */
	iter = root->d.clos_class.parents;
	while(nextparm(&o, &iter)) {

		/* check whether there are unexplored siblings */
		iter2 = o->d.clos_class.children;
		branch = 1;
		while(nextparm(&sib, &iter2)) {
			if(!sib->d.clos_class.prec) {
				/* prec unset, sibling is unvisited */
				branch = 0;
				break;
			}
		}

		/* we are the last visited children, go on to parent. */
		if(branch) build_precedence_branch(o, tail);
	}
}

/* calculate class precedence */
void build_precedence()
{
	int i;
	s_var *var;
	s_obj *tail;

	s_obj *o, *iter;

	class_precedence = NULL;
	tail = NULL;

	/* set prec pointer of all classes to NULL. (unvisited) */
	for(i=0; i<class_hash->ns; i++) {
		for(var = class_hash->hash[i]; var; var = var->next) {
			if(var->value) {
				var->value->d.clos_class.prec = NULL;
			}
		}
	}

	/* iterate over class_hash
	 * the classes that don't have children. */
	for(i=0; i<class_hash->ns; i++) {
		for(var = class_hash->hash[i]; var; var = var->next) {
			if(var->value) {
				if(var->value->d.clos_class.children->type == ST_NIL) {
					build_precedence_branch(var->value, &tail);
				}
			}
		}
	}

	/* set precedence scores */
	iter = class_precedence; i = 0;
	while(nextparm(&o, &iter)) {
		o->d.clos_class.prec_score = i++;
	}
}

/* calculate and update data for method */
s_obj *method_calculate(varctx *ctx, s_obj *name, s_obj *method)
{
	s_obj *generic;

	assert(name->type == ST_SYMBOL);

	/* check whether a generic already exists by this name */
	s_var *var = varctx_get_s(ctx, name->d.symbol.sdbm, name->d.symbol.id);
	if(var) {
		/* there's a binding for this name */
		if(var->value->type != ST_GENERIC)
			return create_error(SE_TYPE, "method_calculate: name is bound but not a generic");
		generic = var->value;

		/* clear dispatch table */
		generic->d.clos_generic.dispatch_fill = 0;
		int i;
		for(i = 0; i<generic->d.clos_generic.dispatch_size; i++) {
			generic->d.clos_generic.dispatch[i].method = NULL;
			free(generic->d.clos_generic.dispatch[i].signature);
			generic->d.clos_generic.dispatch[i].signature = NULL;
		}

	} else {
		/* name unbound, create generic */
		generic = create_generic();

		/* bind name in global ctx */
		while(ctx->parent) ctx = ctx->parent;
		varctx_set_s(ctx, name->d.symbol.sdbm, name->d.symbol.id, generic);
	}

	s_obj **mlist;
	int order = 1;

	/* determine which method list should this be inserted to */
	switch(method->d.clos_method.place) {
	case CL_METHOD_BEFORE:
		mlist = &generic->d.clos_generic.m_before;
		break;
	case CL_METHOD_AFTER:
		mlist = &generic->d.clos_generic.m_after;
		order = -1; /* :after list is sorted the other way */
		break;
	case CL_METHOD_AROUND:
		mlist = &generic->d.clos_generic.m_around;
		break;
	default:
	case CL_METHOD_PRIMARY:
		mlist = &generic->d.clos_generic.m_primary;
		break;
	}

	/* is target list empty? */
	if((*mlist)->type == ST_NIL) {
		/* if so, put this as first element and exit. */

		*mlist = create_cons(method, create_nil());
		update_generic_nexts(generic);
		return create_t();
	}

	if(order*compare_specificity(method, (*mlist)->d.cons.car) > 0) {
		/* more specific than first element in list */

		*mlist = create_cons(method, *mlist);
		update_generic_nexts(generic);
		return create_t();
	}

	/* skip *mlist until element before list end
	 * or until next element is less specific */
	while((*mlist)->d.cons.cdr->type == ST_CONS && order*compare_specificity(method, (*mlist)->d.cons.cdr->d.cons.car) < 0) {
		mlist = &(*mlist)->d.cons.cdr;
	}

	if((*mlist)->d.cons.cdr->type == ST_NIL) {
		/* append after list. */

		(*mlist)->d.cons.cdr = create_cons(method, create_nil());
		update_generic_nexts(generic);
		return create_t();
	}

	if(compare_specificity(method, (*mlist)->d.cons.cdr->d.cons.car) == 0) {
		/* equally specific to next element
		 * replace it. */

		(*mlist)->d.cons.cdr->d.cons.car = method;
		update_generic_nexts(generic);
		return create_t();
	}

	/* next element is less specific, insert before. */

	(*mlist)->d.cons.cdr = create_cons(method, (*mlist)->d.cons.cdr);
	update_generic_nexts(generic);
	return create_t();
}

/* tries to find a valid next method in lst */
s_obj *find_next_method(s_obj *method, s_obj *lst)
{
	assert(method->type == ST_METHOD);

	s_obj *item, *iter;
	iter = lst;
	while(nextparm(&item, &iter)) {
		if(is_next_method(method, item))
			return item;
	}

	return NULL;
}

/* update next pointers for all methods in a generic */
void update_generic_nexts(s_obj *generic)
{
	assert(generic->type == ST_GENERIC);

	s_obj *item, *iter;

	/* primary methods */
	iter = generic->d.clos_generic.m_primary;
	while(nextparm(&item, &iter)) {
		if(iter->type == ST_CONS)
			item->d.clos_method.next = find_next_method(item, iter);
		else
			item->d.clos_method.next = NULL;
	}

	/* before methods are now unoptimized */
	iter = generic->d.clos_generic.m_before;
	while(nextparm(&item, &iter)) {
		if(iter->type == ST_CONS)
			item->d.clos_method.next = iter->d.cons.car;
		else
			item->d.clos_method.next = NULL;
	}

	/* after methods are now unoptimized */
	iter = generic->d.clos_generic.m_after;
	while(nextparm(&item, &iter)) {
		if(iter->type == ST_CONS)
			item->d.clos_method.next = iter->d.cons.car;
		else
			item->d.clos_method.next = NULL;
	}

	/* around methods */
	iter = generic->d.clos_generic.m_around;
	while(nextparm(&item, &iter)) {
		if(iter->type == ST_CONS) {
			item->d.clos_method.next = find_next_method(item, iter);
			if(item->d.clos_method.next == NULL)
				item->d.clos_method.next = find_next_method(item, generic->d.clos_generic.m_primary);
		} else {
			/* last around method is followed by first primary method */
			if(generic->d.clos_generic.m_primary->type == ST_CONS) {
				item->d.clos_method.next = generic->d.clos_generic.m_primary->d.cons.car;
			} else {
				item->d.clos_method.next = NULL;
			}
		}
	}
}

/* compare specificity of two methods
 * returns
 *  <0   if a is less specific than b
 *  0    if a has the same signature as b
 *  >0   if a is more specific than b
 */
int compare_specificity(const s_obj *a, const s_obj *b)
{
	assert(a->type == ST_METHOD);
	assert(b->type == ST_METHOD);

	int arga = -1, argb = -1;

	for(;;) {
		arga ++;
		argb ++;

		if(arga == a->d.clos_method.argn) {
			/* "A" out of arguments. */

			/* if "B" is too, they're considered equal */
			if(argb == b->d.clos_method.argn)
				return 0;

			/* otherwise, "A" is less specific. */
			return -1;
		}

		/* if "B" is out of arguments, "A" is more specific. */
		if(argb == b->d.clos_method.argn)
			return 1;

		method_arg *aa = a->d.clos_method.args + arga;
		method_arg *ab = b->d.clos_method.args + argb;

		if(aa->value) {
			/* "A" has value filter */

			/* if "B" hasn't, "A" is more specific */
			if(!ab->value)
				return 1;

			/* if "A" value == "B" value, continue to next args */
			if(isoform(aa->value, ab->value))
				continue;

			/* they both have filters, for different values
			 * order them arbitrarily */
			return -1;
		}

		if(ab->value) {
			/* "B" has value filter, where "A" hasn't
			 * "B" is more specific. */
			return -1;
		}

		if(aa->type > -1) {
			/* "A" has a type specified */

			/* if "B" has the same type specified, continue. */
			if(aa->type == ab->type)
				continue;

			/* if "B" has no type, "A" is more specific. */
			if(ab->type == -1)
				return 1;

			if(aa->type > TYPE_LAST_MARK) {
				/* "A" has a class type */

				/* if "B" has an atomic type, "A" is more specific. */
				if(ab->type < TYPE_LAST_MARK)
					return 1;

				/* both have class types, compare class specificity. */
				if(typedb[aa->type].cl->d.clos_class.prec_score < typedb[ab->type].cl->d.clos_class.prec_score) {
					/* if "A" is before "B" in the class precedence list,
					 * "A" is more specific. */
					return 1;
				} else {
					/* if "A" is after "B", "A" is less specific. */
					return -1;
				}

			} else {
				/* "A" has an atomic type */

				/* "B" is considered more specific if it has a class type
				 * (> TYPE_LAST_MARK) or has a larger type ID. */

				return ab->type > aa->type ? -1 : 1;
			}
		}

		if(ab->type > -1) {
			/* "B" has a type specified, but "A" hasn't.
			 * "B" is more specific */

			return -1;
		}

		/* both args unspecified, continue. */
	}
}

/* returns true if A class is a generalisation of, or equal to, B class */
int is_generalisation(const s_obj *a, const s_obj *b)
{
	assert(a->type == ST_CLASS);
	assert(b->type == ST_CLASS);

	if(a == b) return 1;

	s_obj *item, *iter;
	iter = b->d.clos_class.ancestry;
	while(nextparm(&item, &iter)) {
		if(item == a) return 1;
	}

	return 0;
}

/* returns true if B method is a possible next-method of A method. */
int is_next_method(const s_obj *a, const s_obj *b)
{
	assert(a->type == ST_METHOD);
	assert(b->type == ST_METHOD);

	int arga = -1, argb = -1;

	for(;;) {
		arga ++;
		argb ++;

		if(arga == a->d.clos_method.argn) {
			/* "A" out of arguments. */

			/* if "B" is too, then B is a next-method */
			if(argb == b->d.clos_method.argn)
				return 1;

			/* otherwise, it's not. */
			return 0;
		}

		if(argb == b->d.clos_method.argn) {
			/* "B" is out of arguments.
			 * "A" has more, so "B" is not a next-method for "A". */
			return 0;
		}

		method_arg *aa = a->d.clos_method.args + arga;
		method_arg *ab = b->d.clos_method.args + argb;

		if(aa->value) {
			/* "A" has a value filter */

			if(ab->value) {
				/* "B" has as well. */

				if(isoform(aa->value, ab->value)) {
					/* the two values are equal, continue */
					continue;
				}

				/* the two values aren't equal. fail */
				return 0;
			}

			/* "B" has no value filter. */

			/* if "B" has no type specialization, then it's ok */
			if(ab->type == -1)
				continue;

			/* it has type specialization. it has to be the same type
			 * or a generalisation */
			if(aa->value->type == ab->type)
				continue;

			/* fail if "A" value filter isn't an instance */
			if(aa->value->type != ST_OBJECT) return 0;

			/* fail if "B" type is atomic */
			if(ab->type < TYPE_LAST_MARK) return 0;

			if(is_generalisation(typedb[ab->type].cl, aa->value->d.clos_object.type))
				continue;

			return 0;
		}

		if(aa->type != -1) {
			/* "A" has a type specialization */

			if(ab->value) {
				/* if "B" has a value filter, it has to be the same type
				 * or a generalisation. */

				if(ab->value->type == aa->type)
					continue;

				if(ab->value->type != ST_OBJECT) return 0;
				if(aa->type < TYPE_LAST_MARK) return 0;
				if(is_generalisation(ab->value->d.clos_object.type, typedb[aa->type].cl))
					continue;

				return 0;
			}

			/* if "B" hasn't a type specialization, continue */
			if(ab->type == -1)
				continue;

			/* otherwise, it has to be the same type or a generalisation */
			if(aa->type == ab->type)
				continue;

			/* fail if either type is atomic */
			if(aa->type < TYPE_LAST_MARK || ab->type < TYPE_LAST_MARK) return 0;

			if(is_generalisation(typedb[ab->type].cl, typedb[aa->type].cl))
				continue;

			return 0;
		}

		/* both arguments are unspecialized, continue */
	}
}

/* checks if a method is applicable to an argument array.
 * (last element NULL)
 * returns first argument in the array unaccounted for
 * (this doesn't care about misc arguments), or NULL.
 */
s_obj ** is_applicable(const s_obj *method, s_obj **args)
{
	assert(method->type == ST_METHOD);

	int arg = 0;
	s_obj **iter;

	if(method->d.clos_method.argn == 0)
		return args;

	for(iter = args; *iter; iter++) {
		method_arg *ma = method->d.clos_method.args + arg;

		/* eql filter */
		if(ma->value) {
			if(isoform(ma->value, *iter)) {
				arg++;
				if(arg == method->d.clos_method.argn) return iter;
				continue;
			} else {
				return NULL;
			}
		}

		/* type specialization */
		if(ma->type != -1) {
			if(ma->type < TYPE_LAST_MARK) {
				/* atomic type */
				if((*iter)->type == ma->type) {
					arg++;
					if(arg == method->d.clos_method.argn) return iter;
					continue;
				} else {
					return NULL;
				}
			}

			/* class type, arg has to be an instance */
			if((*iter)->type != ST_OBJECT)
				return NULL;

			if(!is_generalisation(typedb[ma->type].cl, (*iter)->d.clos_object.type))
				return NULL;
		}

		arg++;
		if(arg == method->d.clos_method.argn) {
			/* at this place, we don't care of misc arguments. */
			return iter;
		}
	}

	/* more specific arguments than elements in the argument list. */
	return NULL;
}

/* expand/zero dispatch table */
void expand_dispatch(s_obj *generic)
{
	int i;
	for(i=0; i<generic->d.clos_generic.dispatch_size; i++)
		free(generic->d.clos_generic.dispatch[i].signature);

	generic->d.clos_generic.dispatch_fill = 0;
	generic->d.clos_generic.dispatch_size *= 2;
	generic->d.clos_generic.dispatch_limit = (generic->d.clos_generic.dispatch_size * DISPATCH_THRESHOLD) >> 8;
	free(generic->d.clos_generic.dispatch);
	generic->d.clos_generic.dispatch = (dispatch_entry *) xcalloc(generic->d.clos_generic.dispatch_size, sizeof(dispatch_entry));
}

/* do method dispatch
 * argeval should be an array of s_obj * (last one NULL) of pre-evaluated arguments.
 * returns the ST_METHOD the call should use, or NULL
 * if ignore_around is 1, around methods are ignored
 */
s_obj * method_dispatch(s_obj *generic, s_obj **argeval, const int ignore_around)
{
	assert(generic->type == ST_GENERIC);

	int hashval = 0;
	int fail;

	/* compute hash value */
	s_obj **iter;
	for(iter = argeval; *iter; iter++)
		hashval += typedb[(*iter)->type].typeseed;

	dispatch_entry *table = generic->d.clos_generic.dispatch;
	int size = generic->d.clos_generic.dispatch_size;
	int row = hashval & (size-1);

	/* find row in dispatch table */
	for(;;) {
		if(table[row].method == NULL) break;	/* no hit */
		if(table[row].hash == hashval && table[row].ignore_around == ignore_around) {
			/* hash values match, check signature */

			dispatch_signature *ds = table[row].signature - 1;
			s_obj **arg = argeval - 1;
			fail = 0;
			for(;;) {
				ds++;
				arg++;

				/* if out of signature, it's a match */
				if(ds->type == -1) {
					break;
				}

				/* if out of arguments, no match. */
				if(*arg == NULL) {
					fail = 1;
					break;
				}

				/* if specified, check for value match */
//				if(ds->value) {
//					if(!isoform(ds->value, *arg)) {
//						fail = 1;
//						break;
//					}

					/* no need to check type in this case */
//					continue;
//				}

				if(ds->type < TYPE_LAST_MARK) {
					/* atomic type */
					if(ds->type != (*arg)->type) {
						fail = 1;
						break;
					}
				} else {
					/* class type, arg has to be an instance */
					if((*arg)->type != ST_OBJECT) {
						fail = 1;
						break;
					}

					if((*arg)->d.clos_object.type->d.clos_class.typeid != ds->type) {
						fail = 1;
						break;
					}
				}
			}

			if(!fail) {
				/* found a match from dispatch table */
				return table[row].method;
			}
		}

		row = (row + 1) & (size - 1);
	}

	/* no result, calculate method by hand */

	s_obj *item, *mlist;

	if(!ignore_around) {
		/* look for most specific applicable around method. */
		mlist = generic->d.clos_generic.m_around;
		while(nextparm(&item, &mlist)) {
			s_obj **misc = is_applicable(item, argeval);

			if(misc) {
				if(size == generic->d.clos_generic.dispatch_limit) {
					expand_dispatch(generic);
					write_dispatch(generic, hashval & (size*2 - 1), hashval, argeval, item, ignore_around);
				} else {
					write_dispatch(generic, row, hashval, argeval, item, ignore_around);
				}

				return item;
			}
		}
	}

	/* no around method found, look in primary methods */
	mlist = generic->d.clos_generic.m_primary;
	while(nextparm(&item, &mlist)) {
		s_obj **misc = is_applicable(item, argeval);

		if(misc) {
			if(size == generic->d.clos_generic.dispatch_limit) {
				expand_dispatch(generic);
				write_dispatch(generic, hashval & (size*2 - 1), hashval, argeval, item, ignore_around);
			} else {
				write_dispatch(generic, row, hashval, argeval, item, ignore_around);
			}

			return item;
		}
	}

	/* nothing found. */
	return NULL;
}

/* write row into dispatch table
 * finds first empty row starting from <row>
 * returns that row, after filling it in.
 */
int write_dispatch(s_obj *generic, int row, int hashval, s_obj **argeval, s_obj *method, const int ignore_around)
{
	assert(generic->type == ST_GENERIC);

	dispatch_entry *table = generic->d.clos_generic.dispatch;
	int size = generic->d.clos_generic.dispatch_size;
	int arglen;
	for(arglen = 0; argeval[arglen]; arglen ++);

	while(table[row].method) {
		row = (row + 1) & (size - 1);
	}

	table[row].hash = hashval;
	table[row].method = method;

	table[row].signature = (dispatch_signature *) xcalloc(arglen + 1, sizeof(dispatch_signature));
	table[row].ignore_around = ignore_around;

	int i;
	for(i=0; i<arglen; i++) {
		if(argeval[i]->type == ST_OBJECT)
			table[row].signature[i].type = argeval[i]->d.clos_object.type->d.clos_class.typeid;
		else
			table[row].signature[i].type = argeval[i]->type;
	}

	/* append a -1 type after last element. */
	table[row].signature[i].type = -1;

	return row;
}


/*************************************** DICTIONARY ***/


/* rehash dictionary */
void dictrehash(s_obj *dict)
{
	int i;
	int ns = dict->d.dict.size * 2;
	int nm = ns - 1;
	dictentry *ndata = (dictentry *) xcalloc(ns, sizeof(dictentry));
	dictentry *odata = dict->d.dict.data;

	dict->d.dict.fill = 0;
	for(i=0; i<dict->d.dict.size; i++) {
		if(odata[i].name != NULL && odata[i].name != (s_obj *) 1)
			dict->d.dict.fill += dictput_(ndata, nm, odata[i].hash, odata[i].name, odata[i].value);
	}
	free(odata);
	dict->d.dict.data = ndata;
	dict->d.dict.size = ns;
	dict->d.dict.mask = nm;
	dict->d.dict.threshold = (ns * dict_threshold) >> 8;
}

/* put value into dictionary (helper function) */
int dictput_(dictentry *data, unsigned int mask, unsigned long keyhash, s_obj *key, s_obj *value)
{
	unsigned int slot = keyhash & mask;
	unsigned int s = slot;
	for(;;) {
		if(!data[s].name || ((ptrint) data[s].name) == 1) {
			data[s].name = key;
			data[s].value = value;
			data[s].hash = keyhash;
			return 1;
		}
		if(data[s].hash == keyhash && isoform(data[s].name, key)) {
			data[s].value = value;
			return 0;
		}

		s = (s + 1) & mask;
		assert(s != slot);
	}
}

/* put value into dictionary */
void dictput(s_obj *dict, unsigned long keyhash, s_obj *key, s_obj *value)
{
	if(dict->d.dict.fill == dict->d.dict.threshold)
		dictrehash(dict);
	dict->d.dict.fill += dictput_(dict->d.dict.data, dict->d.dict.mask, keyhash, key, value);
}

/* get value from dictionary */
s_obj *dictget(const s_obj *dict, const s_obj *key)
{
	unsigned long int keyhash = hashfunc(key);
	unsigned int slot = keyhash & dict->d.dict.mask;
	unsigned int s = slot;
	dictentry *data = dict->d.dict.data;
	unsigned int mask = dict->d.dict.mask;
	for(;;) {
		if(!data[s].name) return NULL;
		if(((ptrint) data[s].name) != 1 && data[s].hash == keyhash && isoform(data[s].name, key))
			return data[s].value;

		s = (s + 1) & mask;
		if(s == slot) return NULL;
	}
}

/* delete item from dictionary */
int dictdel(s_obj *dict, const s_obj *key)
{
	unsigned long int keyhash = hashfunc(key);
	unsigned int slot = keyhash & dict->d.dict.mask;
	unsigned int s = slot;
	dictentry *data = dict->d.dict.data;
	unsigned int mask = dict->d.dict.mask;
	for(;;) {
		if(!data[s].name) return 0;
		if((ptrint) data[s].name != 1 && data[s].hash == keyhash && isoform(data[s].name, key)) {
			data[s].name = (s_obj *) 1;
			return 1;
		}

		s = (s + 1) & mask;
		if(s == slot) return 0;
	}
}

