#ifndef _ST_TYPES_H
#define _ST_TYPES_H

#include <stdio.h>

#ifndef ARRAY_MAXDIMS
#define ARRAY_MAXDIMS 4
#endif

#define REUSABLE_INT_LIMIT 256 /* cache first N integer objects (unless DISABLE_INTCACHE is defined) */

#ifndef _MSC_VER
#define __int64 long long
#endif

#ifdef AMD64
typedef long ptrint;
#else
typedef int ptrint;
#endif

/* object serial number for next allocation */
extern unsigned __int64 _stutter_next_serial;

extern int default_dict_size; /* default size for a dictionary */
extern int dict_threshold;    /* dictionary rehash threshold (/256) */

extern int next_typeid; /* next type ID */


/* primitive object types */
typedef enum {
	ST_NIL, /* array depends on this being zero */
	ST_INT,
	ST_NUM,
	ST_CHAR,
	ST_T,
	ST_STRING,

	ST_COOKIE,
	ST_EOF,

	TYPE_ATOM_MARK, /* no types before this have references to other objects */

	ST_BUILTIN,
	ST_FUNC,
	ST_ERROR,
	ST_DEFER,
	ST_ARRAY,
	ST_VARCTX,
	ST_DICT,
	ST_STREAM,
	ST_READTABLE,

	ST_CLASS,
	ST_SLOT,
	ST_OBJECT,
	ST_GENERIC,
	ST_METHOD,
	ST_DISPATCH,

	TYPE_NOEVAL_MARK, /* all types before this evaluate to themselves */

	ST_CONS,
	ST_SYMBOL,
	ST_VARREF,

	TYPE_LAST_MARK /* first non-used type ID */
} s_type;

/* cookie types */
typedef enum {
	COOKIE_SQLITE,
	COOKIE_SQLITE_STMT,
	COOKIE_DISPATCH
} cookie_type;

/* forward declaration of generic object type */
typedef struct s_obj s_obj;

typedef struct {
	int typeseed;	/* seed for a type */
	s_obj *cl;		/* class object, if applicable */
} typeinfo;


extern typeinfo *typedb; /* type ID -> type info */
extern int typedb_size;

/* CONS object */
typedef struct {
	s_obj *car;
	s_obj *cdr;
} st_cons;

/* Number object */
typedef struct {
	double value;
} st_num;

/* Integer object */
typedef struct {
	int value;
} st_int;

typedef struct s_var s_var;

/* Symbol object */
typedef struct {
	char *name;
	unsigned long sdbm;
	unsigned long id;
	int colon;
} st_symbol;

/* Stack variable reference */
typedef struct {
	int offset; /* added to BP. in bytes (so cell*sizeof(void*)) */
} st_varref;

/* String object */
typedef struct {
	char *value;
} st_string;

/* Array object */
typedef struct {
	int dim; /* number of dimensions */
	int size[ARRAY_MAXDIMS]; /* size of each dimension */
	s_obj **data; /* data */
} st_array;

/* Character object */
typedef struct {
	unsigned char ch;
} st_char;

/* Entry in a dictionary */
typedef struct {
	unsigned long hash;
	s_obj *name;
	s_obj *value;
} dictentry;

/* Dictionary object */
typedef struct {
	unsigned int size; /* dictionary size (power of 2) */
	unsigned int mask; /* size - 1 */
	unsigned int fill; /* number of slots occupied */
	unsigned int threshold; /* occupation threshold (surpassing which a rehash occurs) */
	dictentry *data;
} st_dict;

/* Stream types */
enum {
	STREAM_FP,
	STREAM_FUNC,
	STREAM_C
};

/* Stream object */
typedef struct {
	int type;
	int unread;

	/* parsing position */
	int line;

	union {
		FILE *fp;    /* STREAM_FP */
		s_obj *func; /* STREAM_FUNC */
		struct {     /* STREAM_C */
			void *data;
			int type;
			int (* reader) (void **data, const int type);
			int (* writer) (int c, void **data, const int type);
		} c;
	} s;
} st_stream;

struct varctx; /* defined in memory.h */
typedef struct varctx varctx;

/* prototype for builtin handler */
typedef s_obj* (* builtin_handler) (struct varctx *, s_obj *, void *, const int tail);

/* Builtin function object */
typedef struct {
	builtin_handler handler; /* function handler */
	void *data; /* user data passed to handler */
	char *symbol; /* first name this builtin got bound to.. */
	unsigned long symbolid; /* first symbol ID this builtin got bound to */
	char printprefix; /* when printing, use {prefix}(list) instead of (builtin list) */
	int nosubst; /* substitution defuse flag */
    int sform; /* don't compile arguments */
} st_builtin;

/* Function object */
typedef struct {
	s_obj *parm; /* list of function parameters */
	s_obj *exp; /* function body: list of expressions */
	s_obj *docstr; /* docstring */
	char *symbol; /* first name this func got bound to.. */
	int macro;	/* 1 for macro, 2 for r-macro
				 * 4 for symbolic-macro, 5 for r-symbolic-macro */
	s_obj *closure; /* if this is a closure, reference to a ST_VARCTX object */
	int trace;
} st_func;

/* Function call deferral object */
typedef struct {
	s_obj *func; /* user function object */
	s_obj *args; /* list of arguments */
	s_obj *obj; /* s-expression of function call */
} st_defer;

/* Error object */
typedef struct {
	int code; /* error code */
	s_obj *obj; /* thrown object */
	char *msg; /* error message */
	char *file; /* location */
	s_obj *trace; /* error trace */
	int line;
} st_error;

/* Cookie object */
typedef struct {
	int type; /* cookie type */
	void *ptr;
} st_cookie;

/* Encapsulated varctx */
typedef struct {
	varctx *ctx;
} st_varctx;

/****************************** CLOS TYPES ***/

/* slot assignment */
typedef struct {
	s_obj *stor;	/* ST_CLASS on static, or NULL */
	s_obj *slot;
	int num;
} slotasg;

/* CLOS class */
typedef struct {
	s_obj *name;	/* symbol */
	int typeid;	/* type ID */

	s_obj *parents;	/* list of parents (ST_CLASS) */
	s_obj *slots;	/* list of own slots (ST_SLOT) */

	s_obj *children;	/* list of children (ST_CLASS) */

	s_obj *prec;	/* cons in precedence list */
	int prec_score;	/* position in precedence list */

	s_obj *ancestry;	/* ancestors sorted by precedence */

	slotasg *allslots;	/* list of all slots */
	int allslotsnum;	/* number of all slots */

	varctx *initarg;	/* map initarg->slot_asg (idx) */
	varctx *slotname;	/* map slotname->slot_asg (idx) */

	s_obj **data;	/* storage of static data */
	int data_size;

	int instdata;	/* data size for instances */
} st_class;

/* CLOS object */
typedef struct {
	s_obj *type;	/* ST_CLASS */

	s_obj **data;	/* data storage */
	int data_size;
} st_object;

/* slot allication type */
typedef enum {
	CL_ALLOC_CLASS,
	CL_ALLOC_INSTANCE
} slot_allocation;

/* CLOS slot */
typedef struct {
	s_obj * name;	/* name: symbol */
	s_obj * accessor;	/* name of accessor, or NULL */
	s_obj * writer;	/* name of writer, or NULL */
	s_obj * reader;	/* name of reader, or NULL */
	s_obj * initarg;	/* name of initarg, or NULL */

	slot_allocation alloc;	/* type of slot allocation */
	s_obj *initform;	/* initializer expression */
} st_slot;

/* signature entry in dispatch table */
typedef struct {
	int type;		/* type id used in call, -1 after last */
	s_obj *value;	/* instance used in call, or NULL */
} dispatch_signature;

/* entry in method dispatch table */
typedef struct {
	int hash;	
	s_obj *method;	/* method mapped to call, or NULL */
	dispatch_signature *signature;	/* signature of call */
	int ignore_around;
} dispatch_entry;

/* CLOS generic function */
typedef struct {
	s_obj * m_primary;  /* list of primary methods, most specific to least */
	s_obj * m_before;   /* :before methods, most specific to least */
	s_obj * m_after;	/* :after methods, least specific to most */
	s_obj * m_around;   /* :around methods, most specific to least */

	dispatch_entry *dispatch;	/* method dispatch table */
	int dispatch_size;			/* entries in dispatch table */
	int dispatch_fill;			/* filled entries */
	int dispatch_limit;			/* limit of fill before expansion */
} st_generic;

/* possible places for a method */
typedef enum {
	CL_METHOD_BEFORE,
	CL_METHOD_PRIMARY,
	CL_METHOD_AFTER,
	CL_METHOD_AROUND
} method_place;

/* method argument definition */
typedef struct {
	s_obj *name;	/* argument name (symbol) */
	int type;		/* type ID, or -1 */
	s_obj *value;	/* value filter (eql ...) or NULL */
} method_arg;

/* CLOS method */
typedef struct {
	method_place place;		/* method type (before, after, etc) */
	method_arg *args;		/* array of specific arguments */
	int argn;				/* number of specific arguments */

	s_obj *miscargs;		/* list of misc arguments */
	s_obj *code;			/* method code */
	s_obj *next;			/* next specific method, or NULL */
} st_method;

/* Method dispatch state */
typedef struct {
	s_obj **argeval;		/* array of evaluated arguments (FUNC_MAXPARMS + 1 long) */
	s_obj *generic;			/* generic */
	s_obj *method;			/* selected method */
} st_dispatch;

/*********************** INTERNAL TYPES ***/

/* type-dependent data is contained in this union */
union s_obj_data {
	st_cons cons;
	st_num num;
	st_symbol symbol;
	st_varref varref;
	st_error error;
	st_builtin builtin;
	st_func func;
	st_string string;
	st_int integer;
	st_defer defer;
	st_array array;
	st_varctx varctx;
	st_dict dict;
	st_stream stream;
	st_cookie cookie;
	st_char ch;

	st_class clos_class;
	st_slot clos_slot;
	st_object clos_object;
	st_generic clos_generic;
	st_method clos_method;
	st_dispatch clos_dispatch;
};

typedef struct weakptr weakptr;
struct weakptr {
	s_obj *host;			 /* link host */
	unsigned __int64 serial; /* serial of link host */
	s_obj **ptr;			 /* link location */
	weakptr *next;
};

/* generic object of the language */
struct s_obj {
	unsigned __int64 serial; /* serial number */

	s_type type; /* object type */
	char *file; /* point of definition */
	short line;
	union s_obj_data d; /* type-dependent data contained in the object */
	weakptr *weak;  /* list of weak pointers pointing here */

	unsigned long int mark; /* used by the GC */
};


/* reusable object caches */

#ifndef DISABLE_INTCACHE
s_obj *reusable_ints[REUSABLE_INT_LIMIT];
#endif

#ifndef DISABLE_CHARCACHE
s_obj *reusable_chars[256];
#endif


/* initialize reusable object caches */
void objcache_init();


/* creates and returns a new s_obj, zeroed */
s_obj *alloc_s_obj(void);

/* creates a new s_obj without updating the records of the GC */
s_obj *alloc_ungc_obj(void);

/* create a numeric atom */
s_obj *create_num(const double val);

/* create an integer atom */
s_obj *create_int(const int val);

/* create a character atom */
s_obj *create_char(const unsigned char ch);

/* create an EOF atom */
s_obj *create_eof();

/* create a symbolic atom */
s_obj *create_symbol(char *val);

/* create a string */
s_obj *create_string(const char *val);

/* create a string from n bytes */
s_obj *create_string_counted(const char *val, const int len);

/* create a string (doesn't strdup val) */
s_obj *create_string_allocd(char *val);

/* create a CONS */
s_obj *create_cons(s_obj *car, s_obj *cdr);

/* set last element of a list */
void list_setlast(s_obj *lst, s_obj *obj);

/* create an error object */
s_obj *create_error(const int code, const char *message);
s_obj *create_error_with_obj(const int code, const char *message, s_obj *obj);

/* create a builtin object */
s_obj *create_builtin(builtin_handler handler);

/* create a builtin object, with user data */
s_obj *create_builtin_with_data(builtin_handler handler, void *data);

/* create a builtin object (with sform flag) */
s_obj *create_sform(builtin_handler handler);

/* create a builtin object (with sform flag), with user data */
s_obj *create_sform_with_data(builtin_handler handler, void *data);

/* return cached T object */
s_obj *create_t(void);

/* return cached nil object */
s_obj *create_nil(void);

/* create a T object */
s_obj *create_new_t(void);

/* create a NIL object */
s_obj *create_new_nil(void);

/* create a function */
s_obj *create_func(s_obj *prm, s_obj *exp);

/* create a call deferral */
s_obj *create_defer(s_obj *func, s_obj *args, s_obj *obj);

/* create an array */
s_obj *create_array(const int dim, const int *size);

/* create a dictionary */
s_obj *create_dict();

/* create a stream */
s_obj *create_stream(const int type);

/* create encapsulated varctx */
s_obj *create_obj_varctx(varctx *ctx);

/* create a cookie */
s_obj *create_cookie(const int type, void *ptr);

/* create a class */
s_obj *create_class();

/* create a slot */
s_obj *create_slot();

/* create a object */
s_obj *create_object(s_obj *type);

/* create a generic */
s_obj *create_generic();

/* create a method */
s_obj *create_method();

/* create a dispatch state */
s_obj *create_dispatch();

/* create a varref */
s_obj *create_varref(int offset);

/* add a weak reference place to object */
void weak_reference(s_obj *host, unsigned __int64 serial, s_obj **ptr);

/* remove weakptr references */
inline void weak_unref(s_obj **o);

/* check weakptrs, clean up invalid entries
 * returns number of entries removed */
int weak_check(s_obj *obj);



/****** generic object functions ******/


/* get next type id, create random hash seed */
int gettypeid();

/* get type id for symbol. -1 if no such thing */
int symbol_to_typeid(s_obj *sym);

/* return type id of object */
int object_typeid(s_obj *obj);

/* print description/value of an object */
s_obj * printobj(varctx *ctx, const s_obj *obj);

/* free inner data of object */
void freedata(s_obj *obj);

/* deep copy object */
void deepcopy(s_obj *tgt, s_obj *obj);

/* returns true if two objects are isoform (all fields are equal) */
int isoform(const s_obj *a, const s_obj *b);



/****** CLOS support *****/


/* calculate and update internal data of class */
s_obj *class_calculate(s_obj *cl);

/* calculate and update data for method */
s_obj *method_calculate(varctx *ctx, s_obj *name, s_obj *method);

/* update next pointers for all methods in a generic */
void update_generic_nexts(s_obj *generic);

/* compare specificity of two methods
 * returns
 *  <0   if a is less specific than b
 *  0    if a has the same signature as b
 *  >0   if a is more specific than b
 */
int compare_specificity(const s_obj *a, const s_obj *b);

/* calculate class precedence */
void build_precedence();

/* returns true if A class is a generalisation of, or equal to, B class */
int is_generalisation(const s_obj *a, const s_obj *b);

/* returns true if B method is a possible next-method of A method. */
int is_next_method(const s_obj *a, const s_obj *b);

/* expand/zero dispatch table */
void expand_dispatch(s_obj *generic);

/* checks if a method is applicable to an argument array.
 * (last element NULL)
 * returns first argument in the array unaccounted for
 */
s_obj ** is_applicable(const s_obj *method, s_obj **args);

/* do method dispatch
 * argeval should be an array of s_obj * (last one NULL) of pre-evaluated arguments.
 * returns the ST_METHOD the call should use, or NULL
 * if ignore_around is 1, around methods are ignored
 */
s_obj * method_dispatch(s_obj *generic, s_obj **argeval, const int ignore_around);

/* write row into dispatch table
 * finds first empty row starting from <row>
 * returns that row, after filling it in.
 */
int write_dispatch(s_obj *generic, int row, int hashval, s_obj **argeval, s_obj *method, const int ignore_around);


/****** dictionaries ******/




/* get hash value for an object */
unsigned long hashfunc(const s_obj *obj);

/* put value into dictionary (helper function) */
int dictput_(dictentry *data, unsigned int mask, unsigned long keyhash, s_obj *key, s_obj *value);

/* put value into dictionary */
void dictput(s_obj *dict, unsigned long keyhash, s_obj *key, s_obj *value);

/* rehash dictionary */
void dictrehash(s_obj *dict);

/* get value from dictionary */
s_obj *dictget(const s_obj *dict, const s_obj *key);

/* delete item from dictionary */
int dictdel(s_obj *dict, const s_obj *key);



/****** streams ******/




/* write a character to a stream */
s_obj *stream_write_char(varctx *ctx, s_obj *stream, int ch);

/* unread a character into a stream */
s_obj *stream_unread_char(s_obj *stream, const unsigned char ch);

/* read a character form a stream */
s_obj *stream_read_char(varctx *ctx, s_obj *stream);

/* read a character from *input-stream* */
s_obj *stream_stdin_read(varctx *ctx);

/* unread a character into *input-stream* */
s_obj *stream_stdin_unread(varctx *ctx, const unsigned char ch);

/* write a character to *output-stream* */
s_obj *stream_stdout_write(varctx *ctx, int ch);

/* printf to the standard output stream */
s_obj *stream_printf(varctx *ctx, const char *format, ...);


#endif
